// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.logic;


import java.util.Collection;
import java.util.Iterator;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.inst.SVInstantiations;


/** Meta operator expands the big operators. It is the heart
 * of the meta operator story. It is inspired by the now
 * obsolete MetaJoin operator
 *
 * @author Hubert Schmid
 * @author Stanislas Nanchen
 */

public class MetaBigOperator extends AsmMetaOperator {

    /** Short refernce to TermFactory. */
    private static final TermFactory tf = TermFactory.DEFAULT;

    /** The MetaBigOperator operator has to know all dynamic function
     * symbols.  This list should be initialized on problem loading,
     * because all dynamic function symbols are declared in the
     * problem file.  The elements of this collection are instances of
     * type 'Function'. */
    private Collection functions;


    /** creates a new MetaJoin operator (with uninitialized list of
     * dynamic function symbols). */
    public MetaBigOperator() {
        super("#ASM_BIGOPERATOR", Sort.FORMULA,
              new Sort[] { Sort.FORMULA });
    }

    /** Set the list of dynamic function symbols that are used by this
     * meta operator. This method can only be called once.  The list
     * of dynamic function symbols is typically known on problem
     * loading. The elements of the collection have to be instances of
     * type 'Function'. */
    public void setDynamicFunctions(Collection c) {
        if (c == null) {
            throw new NullPointerException();
        } else if (functions != null) {
            throw new IllegalStateException("Dynamic functions on ASM_JOIN already set.");
        }
        functions = c;
    }

     /** This method replaces a big operator by a suitable formula */
    public Term calculate(Term term, SVInstantiations svInst) {
	BigOperator big;
	Term formula;
	Collection localFunctions;
	Term result;

	if (term.op() instanceof BigOperator) {
	    big = (BigOperator) term.op();
	    localFunctions = big.filter(this.functions);
	    if (localFunctions == null) {
		throw new IllegalStateException("Dynamic functions are not initialized.");
	    }
	    if (localFunctions.size() == 0) {
		result = tf.createJunctorTerm(Op.TRUE);
	    } else {
		result = null;
		formula = term.sub(0);
		Iterator it = localFunctions.iterator();
		while (it.hasNext()) {
		    Function f = (Function) it.next();
		    
		    /* We create first n(arity) new logical variable and terms
		     * corresponding to them; and the f(x1, ..., xn) term
		     */
		    LogicVariable[] xs = new LogicVariable[f.arity()];
		    Term[] terms = new Term[f.arity()];
		    Term fterm;
		    VariableFactory factory = new VariableFactory();
		    for(int j = 0; j < xs.length; j++) {
			xs[j] = factory.createNewVariable(term, "x", f.argSort(j));
			terms[j] = tf.createVariableTerm(xs[j]);
		    }
		    fterm = tf.createFunctionTerm(f, terms);
		    
		    /* then we need to transform the formula of the big operator
		     * i.e. transform recursively the %f into:
		     * - a series of all quantors or ex quantors.
		     * - the fterm.
		     */
		    formula = transform_and_subst(formula, big.getVariableName(), xs, fterm);
		    
		    /* Now we can collect the results.
		     */
		    if (result == null) {
			result = formula;
		    } else {
			result = big.join(result, formula);
		    }
		}
	    }
	} else {
	    result = term;
	}
	return result;
    }

    private Term transform_and_subst(Term formula,
				     Name var,
				     LogicVariable[] xs,
				     Term fterm) {
	Term[] subterms = new Term[formula.op().arity()];
	Operator op = formula.op();
	Term result;

	if (op.name().equals(var)) {
	    /* We have found our NArityLogicVariable as term.
	     */
	    result = fterm;
	} else {
	    /* We are not at the end of the tree and must recursively
	     * go down.
	     */
	    for(int i=0; i<subterms.length; i++) {
		subterms[i] = transform_and_subst(formula.sub(i), var, xs, fterm);
	    }
	    
	    ArrayOfQuantifiableVariable vars = formula.varsBoundHere(0);
	    if ((op == Op.ALL || op == Op.EX) &&
		vars.size() == 0 &&
		vars.getQuantifiableVariable(0) == var) {
		/* if we have our found a NArityLogicVariable var, we must
		 * make the remplacement.
		 */
		result = subterms[0];
		for (int i = xs.length - 1; i >= 0; i--) {
		    result = tf.createQuantifierTerm((Quantifier) op, xs[i], result);
		}
	    } else if ((op instanceof BigOperator) &&
		       ((BigOperator) op).getVariableName().equals(var)) {
		/* we check if we should subsitute further, VERY UGLY */
		result = formula;
	    } else {
		/* we can simply reconstitute */
		result = tf.createTerm(op,
				       subterms,
				       op.arity()==0?Term.EMPTY_VAR_LIST:formula.varsBoundHere(0),
				       formula.javaBlock(),
				       formula.ruleVar());
	    }
	}
	
	return result;
    }
}
