// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

header {

    package de.uka.ilkd.asmkey.parser.antlr;

    import de.uka.ilkd.asmkey.parser.ast.*;
    import de.uka.ilkd.key.parser.Location;
    import de.uka.ilkd.asmkey.logic.AsmOperator;
    import de.uka.ilkd.asmkey.logic.dfinfo.*;

    import java.util.*;

}


class AsmKeyParser extends Parser;


options {
	importVocab = AsmKeyLexer;
	k = 2;
	defaultErrorHandler = false;
}


{

    /* --- static methods --- */

    private static final Location getLocation(AstDeclaration[] decls) {
        if (decls.length == 0) {
            return null;
        } else {
            return decls[0].getLocation();
        }
    }

    private static final Location getLocation(AstTerm[] terms) {
        if (terms.length == 0) {
            return null;
        } else {
            return terms[0].getLocation();
        }
    }

    /** Return a new empty list. */
    private static List createList() {
        return new ArrayList();
    }

    private static Location createLocation(Token token) {
        return new Location(null, token.getLine(), token.getColumn());
    }

    /** return a new array 't # terms'. */
    private static AstTerm[] prependTerm(AstTerm t, AstTerm[] terms) {
        AstTerm[] a = new AstTerm[terms.length + 1];
        System.arraycopy(terms, 0, a, 1, terms.length);
        a[0] = t;
        return a;
    }


    /* --- static fields --- */

    private static final AstTerm[] EMPTY_TERMS = new AstTerm[]{ };

    private static final Identifier[] EMPTY_IDS = new Identifier[]{ };

    private static final AstCondition[] EMPTY_CONDITIONS = new AstCondition[]{ };

    private static final AstParameter[] EMPTY_PARAMETERS = new AstParameter[]{ };

    
    /* --- constructors --- */

    /** Create a parser with the given input and the given ast factory. */
    public AsmKeyParser(TokenStream lexer, AstFactory af) {
        this(lexer);
        this.af = af;
    }


    /* --- fields --- */

    /** The used AstFactory. */
    private AstFactory af = null;


    /* --- inner class --- */

    /** This class stores either a term or a sequent. It is needed
     * because some productions can't distinguish between terms and 
     * sequents without infinte lookahead.  */
    private static final class TermSequent {
        
        final AstTerm term;
        
        final AstSequent sequent;


        TermSequent(AstTerm term) {
            this(term, null);
        }

        TermSequent(AstSequent sequent) {
            this(null, sequent);
        }

        TermSequent(AstTerm term, AstSequent sequent) {
            this.term = term;
            this.sequent = sequent;
        }

    }
    
}


top_formula returns [AstTerm t]
    :
        t=term EOF
    ;

top_term returns [AstTerm t]
    :
        t=term EOF
    ;

top_problem returns [AstProblem p]
    :
        p=problem EOF
    ;

top_decls returns [AstDeclaration[] decls]
    :
        decls=decls EOF
    ;

top_package_decls returns [AstDeclaration[] decls]
    :
        decls=package_decls EOF
    ;

top_package_lemmas returns [AstDeclaration[] decls]
    :
        decls=package_lemmas EOF
    ;

top_asm_rule returns [AstAsmRule r]
    :
        r=asm_rule EOF
    ;

top_importdecls returns [AstImportFile[] impfiles]
    :
        impfiles=problem_importdecls EOF
    ;

top_importlemmas returns [AstImportFile[] impfiles]
    :
        impfiles=problem_importlemmas EOF
    ;

protected
ident returns [Identifier s]
	:
		id:IDENT
        {
			s = new Identifier(createLocation(id), id.getText());
		}
    |
        /* '+' is also a valid identifier, because it is used as
         * function symbol. This is a very ugly. */
        plus:PLUS
        {
            s = new Identifier(createLocation(plus), "+");
        }
	;

protected
ident_asm_rule returns [Identifier s]
    :
        (
            (s=ident)
        |
            (
                asm:ASM RULE
                {
                    s = new Identifier(createLocation(asm), "asm-rule");
                }
            )
        )
    ;
        

protected
identtimed returns [Identifier s]
    :
        id:IDENTTIMED
        {
            s = new Identifier(createLocation(id), id.getText());
        }
    ;

protected
identnarity returns [Identifier s]
    :
        id:IDENTNARITY
        {
            s = new Identifier(createLocation(id), id.getText());
        }
    ;

protected
id_list returns [Identifier[] ids]
{
    Identifier id;
    List list = createList();
}
    :
        id=ident
        {
            list.add(id);
        }
        (
            COMMA id=ident
            {
                list.add(id);
            }
        )*
        {
            ids = (Identifier[]) list.toArray(new Identifier[list.size()]);
        }
    ;

protected
id_asm_rule_list returns [Identifier[] ids]
{
    Identifier id;
    List list = createList();
}
    :
       id=ident_asm_rule
        {
            list.add(id);
        }
        (
            COMMA id=ident_asm_rule
            {
                list.add(id);
            }
        )*
        {
            ids = (Identifier[]) list.toArray(new Identifier[list.size()]);
        } 
    ;

/**
 * In the special but important case of Taclets, we don't yet know
 * whether we are going to have a term or a formula, and it is hard
 * to decide a priori what we are looking at.  So the `term'
 * non-terminal will recognize a term or a formula.  The `formula'
 * reads a formula/term and throws an error if it wasn't a formula.
 * This gives a rather late error message. */


protected
formula returns [AstTerm t]
    :
        t=term
    ;


protected
term returns [AstTerm t]
    :
        t=term20
    ;


protected
term20 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term30
        (
            eqv:EQV t2=term20
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(eqv), Operator.EQV),
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term30 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term40
        (
            imp:IMP t2=term30
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(imp), Operator.IMP), 
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term40 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term50
        (
            (
                or:OR t2=term40
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(or), Operator.OR), 
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                orelse:ORELSE t2=term40
                {
                    t1 = af.createTerm(t1.getLocation(),
                        new AstOperator(createLocation(orelse), Operator.ORELSE),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;


protected
term50 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term60
        (
            (
                and:AND t2=term50
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(and), Operator.AND), 
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                andalso:ANDALSO t2=term50
                {
                    t1 = af.createTerm(t1.getLocation(),
                        new AstOperator(createLocation(andalso), Operator.ANDALSO),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;


protected
term60 returns [AstTerm t]
    :
        t=term70
    |
        t=bindingterm
    |
        t=dl_term
    |
        t=big_term
    |
        t=function_term
    |
        not:NOT t=term60
        {
            t = af.createTerm(createLocation(not), 
                new AstOperator(createLocation(not), Operator.NOT), 
                new AstTerm[]{ t });
        }
    ;


protected
term70 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term80
        (
            equals:EQUALS t2=term80
            {
                t1 = af.createTerm(t1.getLocation(), 
                                   new AstOperator(createLocation(equals), Operator.EQUALS), 
                                   new AstTerm[]{ t1, t2 });
            }
        )?
/*        |
            (
                timed:TIMED t2=term80
                {
                    t1 = af.createTerm(createLocation(timed),
                        new AstOperator(createLocation(timed), Operator.TIMED),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )? */
    ;


protected
term80 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term90
        (
            (
                asterix:ASTERIX t2=term80
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(asterix), Operator.MULT),
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                slash:SLASH t2=term80
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(slash), Operator.DIV),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;


protected
term90 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term95
        (
            (
                plus:PLUS t2=term90
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(plus), Operator.ADD),
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                minus:MINUS t2=term90
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(minus), Operator.SUB),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;

protected
term95 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term100
        (
            timed:DLBRACK t2=simpleterm DRBRACK
            {
                t1 = af.createTerm(createLocation(timed),
                                   new AstOperator(createLocation(timed), Operator.TIMED),
                                   new AstTerm[]{ t1, t2 });
            }
        )?
    ;

protected
term100 returns [AstTerm t]
{
    Location location;
    Identifier abbr;
}
    :
        t=funcpredvarterm
    |
        t=set_term
    |
        t=sequence_term
    |
        ASM LPAREN t=asm_rule RPAREN
    |
        LPAREN t=term RPAREN
    |
        true_:TRUE
        {
            t = af.createTerm(createLocation(true_), 
                new AstOperator(createLocation(true_), Operator.TRUE),
                null);
        }
    |
        false_:FALSE
        {
            t = af.createTerm(createLocation(false_), 
                new AstOperator(createLocation(false_), Operator.FALSE), 
                null);
        }
    |
        at:AT abbr=ident
        {
            t = af.createAbbreviationTerm(createLocation(at), abbr);
        }
    ;

protected
simpleterm returns [AstTerm t]
{
    Location location;
    Identifier abbr;
}
    :
        t=varterm
    |
        LPAREN t=term RPAREN
    |
        true_:TRUE
        {
            t = af.createTerm(createLocation(true_), 
                new AstOperator(createLocation(true_), Operator.TRUE),
                null);
        }
    |
        false_:FALSE
        {
            t = af.createTerm(createLocation(false_), 
                new AstOperator(createLocation(false_), Operator.FALSE), 
                null);
        }
    |
        at:AT abbr=ident
        {
            t = af.createAbbreviationTerm(createLocation(at), abbr);
        }
    ;


protected
bindingterm returns [AstTerm t]
{
    AstQuantifier q;
    Identifier var;
    AstTerm t1, t2;
    Identifier sort = null;
}
    :
        (
            q=quantifier (var=ident | var=identnarity) ( COLON sort=ident )? DOT t1=term60
            {
                t = af.createTerm(q.getLocation(), q, var, sort, t1);
            }
        |
            LBRACE var=ident ( COLON sort=ident )? t1=term RBRACE t2=term60
            {
                t = af.createSubstitutionTerm(var.getLocation(), var, sort, t1, t2);
            }
        )
    ;


protected
quantifier returns [AstQuantifier q]
    :
        all:FORALL
        {
            q = new AstQuantifier(createLocation(all), Quantifier.ALL);
        }
    |
        ex:EXISTS
        {
            q = new AstQuantifier(createLocation(ex), Quantifier.EX);
        }
    ;


protected
dl_term returns [AstTerm t]
{
    AstAsmRule r, s;
    AstTerm t1, t2, t3=null, f;
}
    :
        lbrack:LBRACK r=asm_rule RBRACK  (TIMED t3=simpleterm)? t=term60
        {
            t = af.createDlBox(createLocation(lbrack), r, t, t3);
        }
    |
        lt:LT r=asm_rule GT (TIMED t3=simpleterm)?  t=term60
        {
            t = af.createDlDiamond(createLocation(lt), r, t, t3);
        }
    |
        def:ASMDEF (TIMED t3=simpleterm )? LPAREN r=asm_rule RPAREN
        {
            t = af.createDlDef(createLocation(def), r, t3);
        }
    |
        upd:ASMUPD (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA t1=term COMMA t2=term RPAREN
        {
            t = af.createDlUpd(createLocation(upd), t1, t2, r, t3);
        }
    |
        con:ASMCON (TIMED t3=simpleterm)? LPAREN r=asm_rule RPAREN
        {
            t = af.createDlCon(createLocation(con), r, t3);
        }
    |
        inv:ASMINV (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA t=term RPAREN
        {
            t = af.createDlInv(createLocation(inv), t, r, t3);
        }
    |
        joinable:ASMJOINABLE  (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA s=asm_rule RPAREN
        {
            t = af.createDlJoinable(createLocation(joinable), r, s, t3);
        }
    |
        equi:ASMEQUI (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA s=asm_rule RPAREN
        {
            t = af.createDlEqui(createLocation(equi), r, s, t3);
        }
    |
        equiS:ASMEQUIS (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA s=asm_rule RPAREN
        {
            t = af.createDlEquiS(createLocation(equiS), r, s, t3);
        }
    |
        equiA:ASMEQUIA (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA s=asm_rule RPAREN
        {
            t = af.createDlEquiA(createLocation(equiA), r, s, t3);
        }
    |
        accT:ASMACCT (TIMED t3=simpleterm)? LPAREN t2=term COMMA t1=term RPAREN
        {
            t = af.createDlAccT(createLocation(accT), t1, t2, t3);
        }
    |
        accF:ASMACCF (TIMED t3=simpleterm)? LPAREN f=formula COMMA t=term RPAREN
        {
            t = af.createDlAccF(createLocation(accF), t ,f, t3);
        }
    |
        acc:ASMACC (TIMED t3=simpleterm)? LPAREN r=asm_rule COMMA t=term RPAREN
        {
            t = af.createDlAcc(createLocation(acc), t, r, t3);
        }
    ;

protected
function_term returns [AstTerm term]
{
    Identifier id;
    AstTerm func, form;
}
    :
        function:FUNCTION LPAREN
        id=ident COMMA
        func=term COMMA
        form=formula
        RPAREN
        {
            term=af.createFunctionTerm(createLocation(function),id,func,form);
        }
    ;

protected
big_term returns [AstTerm term]
{
    boolean and_flag;
    int type_flag = AstFactory.ALL_TYPE;
    int prop_flag = AstFactory.ALL_PROPERTY;
    Identifier id;
    Identifier[] rules = null;
    AstTerm form;
}
    :
        ( BIGAND { and_flag = true; } | BIGOR { and_flag = false; })
        big:LPAREN
        id = ident
        (
            RPAREN
            {
                term = af.createTerm(id.getLocation(), id, null);
                term = af.createBigTerm(createLocation(big), term, and_flag);
            }
        |
            (
                ( STATIC
                    {type_flag = AstFactory.STATIC_TYPE;}
                | DYNAMIC
                    (
                        ( UPDATE {prop_flag = AstFactory.UPDATE_PROPERTY;}
                        | ACCESS {prop_flag = AstFactory.ACCESS_PROPERTY;}
                        )
                        LPAREN
                        rules = big_term_rules
                        RPAREN
                    )?  
                    {type_flag = AstFactory.DYNAMIC_TYPE;}
                )?
                COMMA
                form = formula
                RPAREN
                {
                    term = af.createBigTerm(createLocation(big), id, form, and_flag, type_flag, prop_flag, rules);
                }
            )
        )
    ;

protected
big_term_rules returns [Identifier[] rules]
{
    List list = createList();
    Identifier id;
}
    :
        id = ident
        {
            list.add(id);
        }
        (
            COMMA id=ident
            {
                list.add(id);
            }
        )*
        {
            rules = (Identifier[]) list.toArray(new Identifier[list.size()]);
        }
    ;


protected
term_list returns [AstTerm[] terms]
{
    List list = createList();
    AstTerm t;
}
    :
        t=term
        {
            list.add(t);
        }
        (
            COMMA t=term
            {
                list.add(t);
            }
        )*
        { 
            terms = (AstTerm[]) list.toArray(new AstTerm[list.size()]);
        }
    ;

protected
varterm returns [AstTerm a]
{
    Identifier id;
}
    :
        id=ident
        {
            a = af.createTerm(id.getLocation(), id, null);
        }
    ;

protected
funcpredvarterm returns [AstTerm a]
{
    Identifier id;
    AstTerm[] l=null;
    AstTerm t=null;
}
    :
        (
            id=ident l=term_args
        |
            id=identtimed t=simpleterm l=term_args 
        |
            id=identnarity
        )
        {
            if (t != null) {
                AstTerm[] temp;
                temp = new AstTerm[l.length + 1];
                for (int i=1; i<= l.length; i++) {
                    temp[i] = l[i-1];
                }
                temp[0] = t;
                l = temp;
            }
            a = af.createTerm(id.getLocation(), id, l);
        }
    ;

protected
term_args returns [AstTerm[] terms = null]
{
    Identifier id;
}
    :
        ( 
            LPAREN
            (
                terms=term_list RPAREN
            |
                RPAREN
                {
                    terms = EMPTY_TERMS;
                }
            )
        )?
    ;

protected
set_term returns [AstTerm t]
{
    AstTerm[] elems = null;
}
    :
        set:LSETBRACK
        (elems=term_list)?
        RBRACK
        {
            t = af.createSetTerm(createLocation(set), elems);
        }
    ;

protected
sequence_term returns [AstTerm t]
{
    AstTerm[] elems = null;
    AstTerm tail = null;
}
    :
        lbrack:LSEQBRACK
        (elems=term_list)?
        (COLON tail=term)? // OR is used as a bar, cf. prolog lists
        {
            t = af.createSequenceTerm(createLocation(lbrack), elems, tail);
        }
        GT
    ;

protected
seq returns [AstSequent s]
{
    AstTerm[] ant = EMPTY_TERMS, suc = EMPTY_TERMS;
} 
    : 
        (
            ant=term_list 
        )?
        SEQARROW
        (
            suc=term_list
        )?
        { 
            s = new AstSequent(getLocation(ant), ant, suc);
        }
    ;


protected
termorseq returns [TermSequent ts]
{
    AstTerm t;
    AstTerm[] ant, suc = EMPTY_TERMS;
}
    :
        (
            t=term
            (
                /* empty */
                {
                    ts = new TermSequent(t);
                }
            |
                SEQARROW
                (
                    suc=term_list
                )?
                {
                    ts = new TermSequent(new AstSequent(t.getLocation(), new AstTerm[]{ t }, suc));
                }
            |
                COMMA ant=term_list SEQARROW
                (
                    suc=term_list
                )?
                {
                    ts = new TermSequent(new AstSequent(t.getLocation(), prependTerm(t, ant), suc));
                }
            )
        |
            seqarrow:SEQARROW 
            (
                suc=term_list
            )?
            {
                ts = new TermSequent(new AstSequent(createLocation(seqarrow), EMPTY_TERMS, suc));
            }
        )
    ;


protected
taclet returns [AstTaclet r]
{ 
    AstSequent s = null;
    TermSequent ts;
    AstCondition[] conds = EMPTY_CONDITIONS;
    AstGoal[] goals;
    AstTacletModifier[] mods;
    Location location = null;
}
	: 
      ((
		( 
            if_:IF LPAREN s=seq RPAREN 
            {
                location = createLocation(if_);
            }
        )?
		( 
            find:FIND LPAREN ts=termorseq RPAREN 
            {
                if (location == null) {
                    location = createLocation(find);
                }
            }
        )
        (
            ( VARCOND | COND ) LPAREN conds=conditions RPAREN
        )?
		goals=goalspecs
        mods=modifiers
        {
            if (ts.term != null) {
                r = new AstTaclet(location, s, ts.term, conds, goals, mods);
            } else {
                r = new AstTaclet(location, s, ts.sequent, conds, goals, mods);
            }
        }
      ) |
        (builtin:BUILTIN 
                {
                    r = null;
                }
            ))
    ;


protected
conditions returns [AstCondition[] conds]
{
    AstCondition cond;
    List list = createList();
}
    :
        cond=condition
        {
            list.add(cond);
        }
        (
            COMMA cond=condition
            {
                list.add(cond);
            }
        )*
        {
            conds = (AstCondition[]) list.toArray(new AstCondition[list.size()]);
        }
    ;


protected
condition returns [AstCondition cond]
{
    Identifier v, w, r, f;
}
    :
        op:OP v=ident 
        (
            EQUALS w=ident
            {
                cond = af.createOperatorEqualsCondition(createLocation(op), true, v, w);
            }
        |
            NOTEQUALS w=ident
            
            {
                cond = af.createOperatorEqualsCondition(createLocation(op), false, v, w);
            }
        )
    |
        v=ident NOT  (
            FREE ASMIN w=ident
            {
                cond = af.createNotFreeCondition(v.getLocation(), v, w);
            }
        |
            UPDATE f=ident
            {
                cond = af.createNotUpdateCondition(v.getLocation(), v,f);
            }
        |
            ACCESS f=ident
            {
                cond = af.createNotAccessCondition(v.getLocation(), v, f);
            }
        )
    |
		v=ident NEW DEPENDING ON w=ident
        { 
            cond = af.createSkolemCondition(v.getLocation(), v, w);
        }
    |
        pure:PURE v=ident
        {
            cond = af.createPureCondition(createLocation(pure), v);
        }
    |
        static_:STATIC v=ident
        {
            cond = af.createStaticCondition(createLocation(static_), v);
        }
    |
        staticargs:STATICARGS v=ident
        {
            cond = af.createStaticArgsCondition(createLocation(staticargs), v);
        }
    |
        dynamic:DYNAMIC v=ident
        {
            cond = af.createDynamicCondition(createLocation(dynamic), v);
        }
    |
        atomic:ATOMIC v=ident
        {
            cond = af.createAtomicCondition(createLocation(atomic), v);
        }
    |
        call:CALL v=ident
        {
            cond = af.createCallCondition(createLocation(call), v);
        }
    |
        r=ident MAY  (
            UDPATE f=ident
            {
                cond = af.createMayUpdateCondition(r.getLocation(), r, f);
            }
        | 
            ACCESS f=ident
            {
                cond = af.createMayAccessCondition(r.getLocation(), r, f);
            }
        )
    ;


protected
goalspecs returns [AstGoal[] goals]
    :
        CLOSE GOAL
        {
            goals = new AstGoal[]{ };
        }
    | 
        goals=goalspec_list
    ;


protected
goalspec_list returns [AstGoal[] goals]
{
    List list = createList();
    AstGoal g;
}
    :
        ((
            g=goalspec
            {
                list.add(g);
            } 
            (
                SEMI g=goalspec
                {
                    list.add(g);
                }
            )*
        )
        |
        (
            g=replacebranch
            {
                list.add(g);
            }
        ))
        { 
            goals = (AstGoal[]) list.toArray(new AstGoal[list.size()]);
        }
    ;


protected
goalspec returns [AstGoal goal] 
    :
        goal=replacewith
    ;


protected
replacewith returns [AstGoal goal]
{
    Location location = null;
    TermSequent ts = null;
    AstSequent s = null;
    AstTacletDeclaration[] l = null;
    String name = null;
}
    :
		(
            rw:REPLACEWITH 
            (
                str:STRING_LITERAL 
                {
                    name = str.getText();
                }
            )?
            LPAREN ts=termorseq RPAREN
            {
                if (location == null) {
                    location = createLocation(rw);
                }
            }
        )?
        (
            add:ADD LPAREN s=seq RPAREN
            {
                if (location == null) {
                    location = createLocation(add);
                }
            }
        )?
        (
            addrules:ADDRULES LPAREN l=taclet_decl_list RPAREN
            {
                if (location == null) {
                    location = createLocation(addrules);
                }
            }
        )?
        {
            if (ts == null) {
                goal = af.createReplaceGoal(location, name, (AstTerm) null, s, l);
            } else if (ts.term != null) {
                goal = af.createReplaceGoal(location, name, ts.term, s, l);
            } else {
                goal = af.createReplaceGoal(location, name, ts.sequent, s, l);
            }
        }
    ;

protected
replacebranch returns [AstGoal g]
{
    AstSequent s;
}
    :
        replacebranch:REPLACEBRANCH
        LPAREN s=seq RPAREN
        {
            g = af.createReplaceGoal(createLocation(replacebranch), null, s);
        }
    ;

protected
taclet_decl_list returns [AstTacletDeclaration[] l]
{
    AstTacletDeclaration decl;
    List list = createList();
}
    :
        decl=taclet_decl
        {
            list.add(decl);
        }
        (
            COMMA decl=taclet_decl
            {
                list.add(decl);
            }
        )*
        {
            l = (AstTacletDeclaration[]) list.toArray(new AstTacletDeclaration[list.size()]);
        }
    ;


protected
modifiers returns [AstTacletModifier[] l]
{
    List list = createList();
    AstTacletModifier tm;
}
    :
        (
            tm=modifier
            {
                list.add(tm);
            }
        )*
        {
            l = (AstTacletModifier[]) list.toArray(new AstTacletModifier[list.size()]);
        }
    ;


protected
modifier returns [AstTacletModifier tm]
{
    Identifier[] ids;
}
    :
        heurs:HEURISTICS LPAREN ids=id_list RPAREN
        {
            tm = af.createTacletHeuristics(createLocation(heurs), ids);
        }
    |
        ni:NONINTERACTIVE
        {
            tm = af.createTacletNonInteractiveModifier(createLocation(ni));
        }
    |
        re:RECURSIVE
        {
            tm = af.createTacletRecursiveModifier(createLocation(re));
        }
    |
        dn:DISPLAYNAME dname:STRING_LITERAL
        {
            tm = af.createTacletDisplayNameModifier(createLocation(dn), dname.getText());
        }
    |
        ht:HELPTEXT htext:STRING_LITERAL
        {
            tm = af.createTacletHelpTextModifier(createLocation(ht), htext.getText());
            // b.setHelpText(htext.getText());
        }
    ;
    

protected 
package_decls returns [AstDeclaration [] decls]
{
    AstDeclaration[] l;
    boolean ps;
    boolean ls;
}
    :
        (ps=packages_section)?
        l=decls_without_lemmas
        (ls=lemmas_section)?
        {
            decls=l;
        }
    ;

protected 
package_lemmas returns [AstDeclaration [] decls]
{
    AstDeclaration[] l;
    boolean ps;
    AstDeclaration[] ls=null;
}
    :
        (ps=packages_section)?
        l=decls_without_lemmas
        (ls=lemma_decls)?
        {
            decls=ls==null?(new AstDeclaration[0]):ls;
        }
    ;

protected
problem returns [AstProblem p]
{
    AstDeclaration[] l;
    AstTerm a;
    AstProofExpression pe = null;
    AstImportFile[] im, il;
}
	:
        (
            im=importdecls
        )?
        (
            il=importlemmas
        )?
 		l=decls
        PROBLEM LBRACE a=formula RBRACE
        (
            pe=proof
        )?
        {
            p = new AstProblem(getLocation(l), l, a, pe);
        }
	;

protected
problem_importdecls returns [AstImportFile[] importfiles]
{
    AstImportFile[] imfs=null, il;
    AstProofExpression pe;
    AstDeclaration[] l;
    AstTerm a;
}
	:
        (
            imfs=importdecls
        )?
        (
            il=importlemmas
        )?
 		l=decls
        "problem" LBRACE a=formula RBRACE
        (
            pe=proof
        )?
        {
            importfiles = imfs==null?(new AstImportFile[0]):imfs;
        }
    ;

protected
problem_importlemmas returns [AstImportFile[] importfiles]
{
    AstImportFile[] imfs, il=null;
    AstProofExpression pe;
    AstDeclaration[] l;
    AstTerm a;
}
	:
        (
            imfs=importdecls
        )?
        (
            il=importlemmas
        )?
 		l=decls
        "problem" LBRACE a=formula RBRACE
        (
            pe=proof
        )?
        {
            importfiles = il==null?(new AstImportFile[0]):il;
        }
    ;

protected
packages_section returns [boolean b] 
{
    boolean up;
}
    :
        PACKAGES LBRACE
        (
            up=use_package
        )+
        RBRACE
        {
            b=true;
        }
    ;

protected
use_package returns [boolean b]
{
    Identifier id;
}
    :
        USE id=ident SEMI
        {
            b=true;
        }
    ;

protected
lemmas_section returns [boolean b]
{
    boolean l;
}
    :
        LEMMAS LBRACE
        (
            l=lemma_def
        )+
        RBRACE
        {
            b = true;
        }
    ;

protected
lemma_def returns [boolean b]
{
    Identifier id;
    AstTerm f;
}
    :
        id=ident LBRACE FORMULA LBRACE
        f=formula
        RBRACE RBRACE SEMI
        {
            b=true;
        }
    ;

protected
importdecls returns [AstImportFile[] impfiles]
{
    List list = createList();
    AstImportFile imfile;
}
    :
        IMPORTFILES LBRACE 
        (
            imfile=importfile
            {
                list.add(imfile);
            }
        )*
        RBRACE
        {
            if (list.size() == 0) {
                impfiles = null;
            } else {
                impfiles = (AstImportFile[]) list.toArray(new AstImportFile[list.size()]);
            }
        }
    ;

protected
importlemmas returns [AstImportFile[] impfiles]
{
    List list = createList();
    AstImportFile imfile;
}
    :
        IMPORTLEMMAS LBRACE 
        (
            imfile=importfile
            {
                list.add(imfile);
            }
        )*
        RBRACE
        {
            if (list.size() == 0) {
                impfiles = null;
            } else {
                impfiles = (AstImportFile[]) list.toArray(new AstImportFile[list.size()]);
            }
        }
    ;


protected
importfile returns [AstImportFile imfile]
{
    String p;
}
    :
        path:FILE p=path SEMI
        {
            imfile = new AstImportFile(createLocation(path),p);
        }
    ;

protected
path returns [String p]
    :
        pa:STRING_LITERAL
        {
            p = pa.getText();
        }
    ;
        

protected
decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration[] l;
}
    :
        l=decls_without_lemmas
        {
            list.addAll(Arrays.asList(l));
        }
        (
            l=lemma_decls
            {
                list.addAll(Arrays.asList(l));
            }
        )?
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;

protected
decls_without_lemmas returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration[] l;
}
    :
		(
            (
                l=sort_decls
            |
                l=schema_var_decls
            |
                l=pred_decls
            |
                l=func_decls
            |
                l=heur_decls
            |
                l=taclet_decls
            |
                l=hyper_decls
            |
                l=procedure_decls
            )
            {
                list.addAll(Arrays.asList(l));
            }
		)*
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
sort_decls returns [AstDeclaration[] decls] 
{
	List list = createList();
    AstDeclaration decl;
}
    :
		SORTS 
		LBRACE
		(
			decl=sort_decl
            {
                list.add(decl);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
sort_decl returns [AstDeclaration decl]
{
    Identifier id;
    boolean isGenericSort=false;
}
    :
        (GENERIC {isGenericSort=true;})? id=ident SEMI
        {
            decl = af.createSortDeclaration(id.getLocation(), id, isGenericSort);
        }
    ;


protected
schema_var_decls returns [AstDeclaration[] decls]
{
	List list = createList();
    AstDeclaration[] decl;
}
    :
		SCHEMA VARIABLES
		LBRACE
		(
			decl=schema_var_decl
            {
                list.addAll(Arrays.asList(decl));
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
schema_var_decl returns [AstDeclaration[] decls]
{
	AstSchemaType r;
    Identifier[] l;
}
    :
        r=schema_type_ref l=id_list SEMI
        {
            decls = new AstDeclaration[l.length];
            for (int i = 0; i < l.length; ++i) {
                decls[i] = af.createSchemaVariableDeclaration(r.getLocation(), l[i], r);
            }
        }
    ;


protected
schema_type_ref returns [AstSchemaType r]
{
    Identifier sort;
    boolean variable = false;
    boolean depending = false;
}
    :
        (
            asm:ASM RULE
            {
                r = af.createSchemaType(createLocation(asm), PrimitiveSchemaType.ASMRULE);
            }
        | 
            formula:FORMULA
            { 
                r = af.createSchemaType(createLocation(formula), PrimitiveSchemaType.FORMULA);
            }			
        | 
            sort=ident 
            ( 
                VARIABLES 
                { 
                    variable = true; 
                }
            |
                DEPENDING
                {
                    depending = true;
                }
            )?
            { 
                if (depending) {
                    r = af.createDependingSchemaType(sort.getLocation(), sort);
                } else {
                    r = af.createSchemaType(sort.getLocation(), sort, variable);
                }
            }
        )

    ;


protected
pred_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		PREDICATES
		LBRACE
		(
            decl=pred_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
pred_decl returns [AstDeclaration decl]
{
    Identifier id;
    Identifier[] args = null;
    Location location = null;
    boolean dynamic = false;
    AstTerm t = null;
}
    :
        (
            (
                der:DERIVED
                id=ident args=arg_ids EQUALS t=formula
                {
                    decl = af.createDerivedPredicateDeclaration(createLocation(der), id, args, t);
                }
            )   
        |
            (
                (
                    dyn:DYNAMIC
                    {
                        dynamic = true;
                    }
                )?
                id=ident args=arg_ids
                {
                    decl = af.createPredicateDeclaration(location == null ? id.getLocation() : location, id, args);
                }
            )
        )
        SEMI
    ;


protected
func_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		FUNCTIONS
		LBRACE
		(
            decl=func_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
func_decl returns [AstDeclaration decl]
{
    Identifier ret;
    Identifier id;
	Identifier[] args=null;
    boolean dynamic = false;
    Location location = null;
    AstTerm t = null;
}
	:
        (
            (
                der:DERIVED
                ret=ident id=ident args=arg_ids EQUALS t=term
                {
                    decl = af.createDerivedFunctionDeclaration(createLocation(der), id, args, ret, t);
                }
            )
        |
            (
                (
                    d:DYNAMIC 
                    { 
                        dynamic = true;
                        location = createLocation(d);
                    }
                )?
                ret=ident id=ident args=arg_ids
                {
                    decl = af.createFunctionDeclaration(location != null ? location : ret.getLocation(),
                                                        id, args, ret, dynamic);
                }
            )
        )
        SEMI
	;


protected
heur_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		HEURISTICS
		LBRACE
		(
            decl=heur_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
heur_decl returns [AstDeclaration decl]
{
    Identifier id;
}
    :
        id=ident SEMI
        {
            decl = af.createHeuristicDeclaration(id.getLocation(), id);
        }
	;


protected
arg_ids returns [Identifier[] sorts = null] 
	:
		(
			LPAREN
            (
                sorts=id_asm_rule_list RPAREN
            |
                RPAREN
                {
                    sorts = EMPTY_IDS;
                }
            )
		)?
	;


protected
taclet_decls returns [AstDeclaration[] decls] 
{
	List list = createList();
    AstDeclaration d;
}
    :
		RULES LBRACE
		(
			d=taclet_decl SEMI
            {
                list.add(d);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
taclet_decl returns [AstTacletDeclaration decl]
{
    Identifier id;
    AstTaclet t;
}
    :
		id=ident LBRACE
        t=taclet
        RBRACE
        {
            decl = new AstTacletDeclaration(id.getLocation(), id, t);
        }
    ;

protected
hyper_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    //AstHyper h;
}
    :
        HYPERRULES LBRACE

        RBRACE
        {
            decls = null;
        }
    ;

protected
procedure_decls returns [AstDeclaration[] decls]
{
	List list = createList();
    AstDeclaration d;
}
    :
		PROCEDURES LBRACE
		(
			d=procedure_decl SEMI
            {
                list.add(d);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
procedure_decl returns [AstDeclaration d]
{
    Identifier id;
    AstParameter[] params = EMPTY_PARAMETERS;
    AstAsmRule r;
}
    :
        id=ident 
        (
            LPAREN 
            (
                params=param_list 
            )?
            RPAREN 
        )?
        LBRACE r=asm_rule RBRACE
        {
            d = af.createProcedureDeclaration(id.getLocation(), id, params, r);
        }
    ;


protected
lemma_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration d;
}
    :
        LEMMAS LBRACE
        (
            d=lemma_decl SEMI
            {
                list.add(d);
            }
        )* RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;

protected
lemma_decl returns [AstDeclaration d]
{
    Identifier id;
    AstTerm phi;
}
    :
        id=ident LBRACE FORMULA LBRACE
        phi=formula
        RBRACE RBRACE
        {
            d = af.createLemmaDeclaration(id.getLocation(), id, phi);
        }
    ;

protected
asm_rule returns [AstAsmRule r]
    :
        r=asm_rule_seq
    ;

protected
asm_rule_seq returns [AstAsmRule r1]
{
    AstAsmRule r2;
}
    :
        r1=asm_rule_par 
        ( 
            ASMSEQ r2=asm_rule_seq 
            { 
                r1 = af.createAsmSeq(r1.getLocation(), r1, r2);
            }
        )?
    ;

protected
asm_rule_par returns [AstAsmRule r1]
{
    AstAsmRule r2;
}
    :
        r1=asm_rule_atom 
        ( 
            ASMPAR r2=asm_rule_par 
            {
                r1 = af.createAsmPar(r1.getLocation(), r1, r2);
            }
        )?
    ;

protected
asm_rule_atom returns [AstAsmRule r]
{
    Identifier id;
}
    :
        (
            r=asm_rule_skip
        |
            r=asm_rule_assign
        |
            r=asm_rule_call
        |
            r=asm_rule_if
        |
            r=asm_rule_elseif
        |
            r=asm_rule_let
        |
            r=asm_rule_forall
        |
            r=asm_rule_try
        |
            r=asm_rule_subst
        |
            r=asm_rule_meta
        |
            id=ident
            {
                r = af.createAsmSchemaVariable(id.getLocation(), id);
            }
        |
            LPAREN r=asm_rule RPAREN
        )
    ;


protected
asm_rule_skip returns [AstAsmRule r]
    :
        skip:"skip"
        {
            r = af.createAsmSkip(createLocation(skip));
        }
    ;


protected
asm_rule_assign returns [AstAsmRule r]
{
    AstTerm t1, t2;
}
    :
        t1=funcpredvarterm ASSIGN t2=term
        {
            r = af.createAsmAssign(t1.getLocation(), t1, t2);
        }
    ;


protected
asm_rule_call returns [AstAsmRule r]
{
    Identifier id;
    AstTerm[] l;
}
    :
        CALL id=ident l=term_args
        {
            r = af.createAsmCall(id.getLocation(), id, l);
        }
    ;


protected
asm_rule_if returns [AstAsmRule r]
{
    AstTerm phi;
    AstAsmRule r1, r2;
}
    :
        if_:IF phi=formula ASMTHEN r1=asm_rule
        r2=asm_rule_if_end
        {
            r = af.createAsmIf(createLocation(if_), phi, r1, r2);
        }
    ;

protected
asm_rule_elseif returns [AstAsmRule r]
{
    AstTerm phi;
    AstAsmRule r1, r2;
}
    :
        elseif:ASMELSEIF phi=formula ASMTHEN r1=asm_rule
        r2=asm_rule_if_end
        {
            r = af.createAsmElseif(createLocation(elseif), phi, r1, r2);
        }
    ;

protected
asm_rule_if_end returns [AstAsmRule r]
{
    AstTerm phi, phi1;
    AstAsmRule r1, r2=null;
    List phis = createList();
    List elsifs = createList();
    List locations = createList();
    int ll;
}
    :
        (elseif:ASMELSEIF phi = formula ASMTHEN r1=asm_rule {phis.add(phi); elsifs.add(r1);
                                                               locations.add(createLocation(elseif));})*
        (else_: ASMELSE r2=asm_rule)
        ASMEND
        {
            //first create the tail.
            if (r2 == null) { r2 = af.createAsmSkip(createLocation(else_)); }
            
            //simple if or elsif
            ll = phis.size();
            if (ll == 0) {
                r = r2;
            } else {
                r = af.createAsmElseif((Location) locations.get(ll-1),
                    (AstTerm) phis.get(ll-1),
                    (AstAsmRule) elsifs.get(ll-1), r2);
                for(int i = ll-2; i>=0; i--) {
                    r = af.createAsmElseif((Location) locations.get(i),
                            (AstTerm)phis.get(i),
                            (AstAsmRule)elsifs.get(i), r);
                }
            }
        }
    ;

protected
asm_rule_let returns [AstAsmRule r]
{
    Identifier v;
    Identifier sort = null;
    AstTerm t;
}
    :
        let:ASMLET v=ident ( COLON sort=ident )? EQUALS t=funcpredvarterm ASMIN r=asm_rule ASMEND
        {
            r = af.createAsmLet(createLocation(let), v, sort, t, r);
        }
    ;


protected
asm_rule_forall returns [AstAsmRule r]
{
    Identifier v;
    Identifier sort = null;
    AstTerm t;
}
    :
        forall:ASMFORALL v=ident ( COLON sort=ident )? ASMWITH t=formula ASMDO r=asm_rule ASMEND
        {
            r = af.createAsmForall(createLocation(forall), v, sort, t, r);
        }
    ;


protected
asm_rule_try returns [AstAsmRule r]
{
    AstAsmRule r1, r2;
}
    :
        try_:ASMTRY r1=asm_rule ASMELSE r2=asm_rule ASMEND
        {
            r = af.createAsmTry(createLocation(try_), r1, r2);
        }
    ;


protected
asm_rule_subst returns [AstAsmRule r]
{
    Identifier var;
    Identifier sort = null;
    AstTerm t;
}
    :
        lbrace:LBRACE var=ident ( COLON sort=ident )? t=term RBRACE r=asm_rule_atom
        {
            r = af.createAsmSubstitution(createLocation(lbrace), var, sort, t, r);
        }
    ;


protected
asm_rule_meta returns [AstAsmRule r]
{
    Identifier id;
}
    :
        at:AT id=ident LPAREN r=asm_rule RPAREN
        {
            r = af.createAsmRuleMeta(createLocation(at), id, r);
        }
    ;


protected
param_list returns [AstParameter[] l]
{
    AstParameter p;
    List list = createList();
}
    :
        p=param
        {
            list.add(p);
        }
        (
            COMMA p=param
            {
                list.add(p);
            }
        )*
        {
            l = (AstParameter[]) list.toArray(new AstParameter[list.size()]);
        }
    ;


protected
param returns [AstParameter p]
{
    Identifier id;
    Identifier sort;
}
    :   
        id=ident COLON sort=ident
        {
            p = new AstParameter(id.getLocation(), id, sort);
        }
    ;



proof returns [AstProofExpression pe]
    :
        PROOF LBRACE pe=proof_expr RBRACE
    ;


protected
proof_expr returns [AstProofExpression pe = null]
{
    AstProofExpression child;
    AstProofExpressionType type;
    List list = createList();
    String id = null;
}
    :
        paren:LPAREN
        (
            type=proof_expr_type
            (
                str:STRING_LITERAL
                {
                    id = str.getText();
                }
            )?
            (
                child=proof_expr
                {
                    list.add(child);
                }
            )*
            {
                AstProofExpression[] children = (AstProofExpression[]) list.toArray(new AstProofExpression[list.size()]);
                pe = new AstProofExpression(createLocation(paren), type, id, children);
            }
        )?
        RPAREN
    ;


/* Das ist aus der Parser-Spezifikation des Java-KeY-Systems. Ich
verstehe allerdings nicht, was der allgemeine (und ungenutzte) IDENT
hier soll. Er wird beispielsweise bei "opengoal" verwendet. */
protected
proof_expr_type returns [AstProofExpressionType type]
    :
        (
            FORMULA
            {
                type = AstProofExpressionType.FORMULA;
            }
        |
            TERM
            {
                type = AstProofExpressionType.TERM;
            }
        |
            BUILTIN
            {
                type = AstProofExpressionType.BUILTIN;
            }
        |
            RULE
            {
                type = AstProofExpressionType.RULE;
            }
        | 
            BRANCH
            {
                type = AstProofExpressionType.BRANCH;
            }
        |   
            INST
            {
                type = AstProofExpressionType.INST;
            }
        |
            IFSEQFORMULA
            { 
                type = AstProofExpressionType.IFSEQFORMULA;
            }
        |
            INTERACTIVE
            { 
                type = AstProofExpressionType.INTERACTIVE;
            }
        |
            HEUR
            {
                type = AstProofExpressionType.HEURISTICS;
            }
        |
            IDENT
            {
                type = AstProofExpressionType.IDENT;
            }
        )
    ;
