// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.ast;


import de.uka.ilkd.key.parser.Location;

/** Abstract factory for creating AST nodes. */

public interface AstFactory {

    /* --- first order logic terms --- */

    /** Operator term. */
    AstTerm createTerm(Location location, AstOperator op, AstTerm[] terms);

    /** Quantifier term. If sort is null, then var is a schema variable */
    AstTerm createTerm(Location location, AstQuantifier q, Identifier var, Identifier sort, AstTerm term);

    /** Function or predicate term. */
    AstTerm createTerm(Location location, Identifier id, AstTerm[] terms);

    /** Substitution term. If sort is null, then var is a schema variable */
    AstTerm createSubstitutionTerm(Location location, Identifier var, Identifier sort, AstTerm subst, AstTerm term);

    /** Big Operators in taclets */
    AstTerm createBigTerm(Location location, AstTerm term, boolean and_flag);

    /** Big Operators in formulas */
    AstTerm createBigTerm(Location location, Identifier id, AstTerm form, boolean and_flag,
			  int type_flag, int prop_flag, Identifier[] rules);

    /** Function Operator in formula */
    AstTerm createFunctionTerm(Location location, Identifier id, AstTerm func, AstTerm form);

    /** Abbreviation. */
    AstTerm createAbbreviationTerm(Location location, Identifier name);

    /* --- sequences and sets --- */

    AstTerm createSetTerm(Location location, AstTerm[] elems);

    AstTerm createSequenceTerm(Location location, AstTerm[] elems, AstTerm tail);

    /* --- dynamic logic --- */

    /** Box operator: <tt>[rule] term</tt>. */
    AstTerm createDlBox(Location location, AstAsmRule rule, AstTerm term, AstTerm step);

    /** Diamond operator: <tt>&lt;rule&gt; term</tt>. */
    AstTerm createDlDiamond(Location location, AstAsmRule rule, AstTerm term, AstTerm step);

    /** Modal def operator: <tt>def(rule)</tt>. */
    AstTerm createDlDef(Location location, AstAsmRule rule, AstTerm step);

    /** Modal update operator: <tt>upd(lhs := rhs in rule)</tt>. */
    AstTerm createDlUpd(Location location, AstTerm lhs, AstTerm rhs, AstAsmRule rule, AstTerm step);

    /** Modal Con operator: <tt>Con(rule)</tt>. */
    AstTerm createDlCon(Location location, AstAsmRule rule, AstTerm step);

    /** Modal inv operator: <tt>inv(term in rule)</tt>. */
    AstTerm createDlInv(Location location, AstTerm term, AstAsmRule rule, AstTerm step);

    /** Modal joinable operator: <tt>joinable(rule1 and rule2)</tt>. */
    AstTerm createDlJoinable(Location location, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Modal equi operator: <tt>equi(rule1, rule2)</tt>. */
    AstTerm createDlEqui(Location location, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Modal equiS operator: <tt>equiS(rule1, rule2)</tt>. */
    AstTerm createDlEquiS(Location location, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Modal equiA operator: <tt>equiA(rule1, rule2)</tt>. */
    AstTerm createDlEquiA(Location location, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Modal accT operator: <tt>accT(term in term)</tt> */
    AstTerm createDlAccT(Location location, AstTerm term1, AstTerm term2, AstTerm step);

    /** Modal accF operator: <tt>accF(term in formula)</tt> */
    AstTerm createDlAccF(Location location, AstTerm term, AstTerm formula, AstTerm step);

    /** Modal acc operator: <tt>acc(term in rule)</tt> */
    AstTerm createDlAcc(Location location, AstTerm term, AstAsmRule rule, AstTerm step);

    ///** Short circuit andalso operator */
    //AstTerm createAndalso(Location location, AstTerm term1, AstTerm term2);

    ///** Short circuit orelse operator */
    //AstTerm createOrelse(Location location, AstTerm term1, AstTerm term2);

    /* --- taclets --- */

    /** Taclet condition: The top level funtion symbol of the terms represented by v and w equal. */
    AstCondition createOperatorEqualsCondition(Location location, boolean equal, Identifier v, Identifier w);

    /** Taclet condition: The variable represented by v occurs not free in the term represented by w. */
    AstCondition createNotFreeCondition(Location location, Identifier v, Identifier w);

    /** Taclet condition. The term represented by the schema variable is pure. */
    AstCondition createPureCondition(Location location, Identifier v);

    /** Taclet condition. The term represented by the schema variable is static. */
    AstCondition createStaticCondition(Location location, Identifier v);

    /** Taclet condition. All subterms of the the term represented by the schema variable are static. */
    AstCondition createStaticArgsCondition(Location location, Identifier v);

    /** Taclet condition. The top level operator of the term represented by the schema variable is dynamic. */
    AstCondition createDynamicCondition(Location location, Identifier v);

    /** Taclet condition. The top level operator of the formula is a predicate */
    AstCondition createAtomicCondition(Location location, Identifier v);

    /** Taclet condition. The top level operator of the term represented by the schema variable is an asm named rull cale. */
    AstCondition createCallCondition(Location location, Identifier v);

    /** Taclet condition. New skolem symbol with meta variables occuring in second term. */
    AstCondition createSkolemCondition(Location location, Identifier v, Identifier w);

    /** Taclet condition. if a rule r may update function f */
    AstCondition createMayUpdateCondition(Location location, Identifier r, Identifier f);

    /** Taclet condition. if a rule cannot update function f */
    AstCondition createNotUpdateCondition(Location location, Identifier r, Identifier f);

    /** Taclet condition. if a rule r may access a function f */
    AstCondition createMayAccessCondition(Location location, Identifier r, Identifier f);

    /** Taclet condition. if a rule cannot access a function f*/
    AstCondition createNotAccessCondition(Location location, Identifier r, Identifier f);

    /** Replace goal. */
    AstGoal createReplaceGoal(Location location, String name, 
			      AstTerm term, AstSequent addSequent, 
			      AstTacletDeclaration[] addRules);

    /** Replace goal. */
    AstGoal createReplaceGoal(Location location,
                              String name,
                              AstSequent sequent,
                              AstSequent addSequent,
                              AstTacletDeclaration[] addRules);

    /** Replace goal for elseif */
    AstGoal createReplaceGoal(Location location,
			      String name,
			      AstSequent sequent);

    /** Taclet modifier: heuristics. */
    AstTacletModifier createTacletHeuristics(Location location, Identifier[] ids);

    /** Taclet modifier: non interactive. */
    AstTacletModifier createTacletNonInteractiveModifier(Location location);

    /** Tacet modifier: recursive. */
    AstTacletModifier createTacletRecursiveModifier(Location location);

    /** Taclet modifier: display name. */
    AstTacletModifier createTacletDisplayNameModifier(Location location, String displayName);

    /** Taclet modifier: help text. */
    AstTacletModifier createTacletHelpTextModifier(Location location, String helpText);

    /* --- terms for asm rules --- */

    /** Sequential ASM rule: <tt>rule1 ; rule2</tt>. */
    AstAsmRule createAsmSeq(Location location, AstAsmRule rule1, AstAsmRule rule2);

    /** Parallel ASM rule: <tt>rule1 rule2</tt>. */
    AstAsmRule createAsmPar(Location location, AstAsmRule rule1, AstAsmRule rule2);

    /** ASM skip rule. */
    AstAsmRule createAsmSkip(Location location);

    /** ASM update rule: <tt>lhs := rhs</tt>. */
    AstAsmRule createAsmAssign(Location location, AstTerm lhs, AstTerm rhs);

    /** ASM named rule call: <tt>id(terms)</tt>. */
    AstAsmRule createAsmCall(Location location, Identifier id, AstTerm[] terms);

    /** ASM if rule: <tt>if formula then thenRule</tt>. */
    AstAsmRule createAsmIf(Location location, AstTerm formula, AstAsmRule thenRule, AstAsmRule elseRule);

    /** ASM if rule: <tt>if formula then thenRule</tt>. */
    AstAsmRule createAsmElseif(Location location, AstTerm formula, AstAsmRule thenRule, AstAsmRule elseRule);

    /** ASM let rule: <tt>let var:sort := term in rule</tt>. */
    AstAsmRule createAsmLet(Location location, Identifier var, Identifier sort, AstTerm term, AstAsmRule rule);

    /** ASM forall rule: <tt>forall var:sort with formula do rule</tt>. */
    AstAsmRule createAsmForall(Location location, Identifier var, Identifier sort, AstTerm formula, AstAsmRule rule);

    /** ASM try rule: <tt>try tryRule else elseRule</tt>. */
    AstAsmRule createAsmTry(Location location, AstAsmRule tryRule, AstAsmRule elseRule);

    /** ASM schema variable. */
    AstAsmRule createAsmSchemaVariable(Location location, Identifier id);

    /** ASM substitution: <tt>{var:sort = subst} term</tt>. */
    AstAsmRule createAsmSubstitution(Location location,
                                     Identifier var,
                                     Identifier sort,
                                     AstTerm subst,
                                     AstAsmRule rule);

    /** ASM meta operator: <tt>id(rule)</tt>. The meta operator has exactly one argument. */
    AstAsmRule createAsmRuleMeta(Location location, Identifier id, AstAsmRule rule);

    /* --- declarations --- */

    /** Primitive schema type. */
    AstSchemaType createSchemaType(Location location, PrimitiveSchemaType type);

    /** Composite schema type. */
    AstSchemaType createSchemaType(Location location, Identifier sort, boolean variable);

    /** Schema type for new depending variables (skolem functions). */
    AstSchemaType createDependingSchemaType(Location location, Identifier sort);

    /** Sort declaration. */
    AstDeclaration createSortDeclaration(Location location, Identifier id, boolean isGenericSort);

    /** Schema variable declaration. */
    AstDeclaration createSchemaVariableDeclaration(Location location, Identifier id, AstSchemaType type);

    /** Predicate declaration. */
    AstDeclaration createPredicateDeclaration(Location location, Identifier id, Identifier[] argTypes);

    /** Derived Predicate declaration */
    AstDeclaration createDerivedPredicateDeclaration(Location location, Identifier id, Identifier[] args, AstTerm formula);

    /** Function declaration. */
    AstDeclaration createFunctionDeclaration(Location location,
                                             Identifier id,
                                             Identifier[] argTypes,
                                             Identifier returnType,
                                             boolean dynamic);

    /** Function declaration. */
    AstDeclaration createDerivedFunctionDeclaration(Location location,
						    Identifier id,
						    Identifier[] args,
						    Identifier returnType,
						    AstTerm t);

    /** Heuristic declaration. */
    AstDeclaration createHeuristicDeclaration(Location location, Identifier id);

    /** Named ASM rule (procedure) declaration. */
    AstDeclaration createProcedureDeclaration(Location location, Identifier id, AstParameter[] params, AstAsmRule rule);

    /** Lemma declaration (lemma a proof may use.) */
    AstDeclaration createLemmaDeclaration(Location location, Identifier id, AstTerm phi);


    /** Constants for the big boolean operator */
    final int ALL_TYPE = 0;
    final int STATIC_TYPE = 1;
    final int DYNAMIC_TYPE = 2;

    final int ALL_PROPERTY = 0;
    final int ACCESS_PROPERTY = 1;
    final int UPDATE_PROPERTY = 2;
}
