// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.asmkey.parser.ast;


/** Vistor for ASM terms.
 *
 * @see AstTerm
 * @author Hubert Schmid
 * @author Stanislas Nanchen  
 */

public interface AstTermVisitor {

    /* --- Terms of first order logic --- */

    /** Operator term: <tt>op(terms)</tt>. */
    Object visitOperatorTerm(Object closure, AstOperator op, AstTerm[] terms);

    /** Quantifier term: <tt>q var:sort. term</tt>. If sort is null,
     * then var is a schema variable. */
    Object visitQuantifierTerm(Object closure, AstQuantifier q, Identifier var, Identifier sort, AstTerm term);

    /** Funtion term: <tt>id(terms)</tt>. */
    Object visitFunctionTerm(Object closure, Identifier id, AstTerm[] terms);

    /** Substitution term: <tt>{var:sort = subst} term</tt>. If sort
     * is null, then var is a schema variable. */
    Object visitSubstitutionTerm(Object closure, Identifier var, Identifier sort,
				 AstTerm subst, AstTerm term);

    /** BigOperator in taclets */
    Object visitBigTerm(Object closure, AstTerm formula, boolean and_flag);


    /** BigOperator in formulas */
    Object visitBigTerm(Object closure, Identifier id, AstTerm formula, boolean and_flag, int type_flag,
			int prop_flag, Identifier[] rules);

    /** functions in formulas */
    Object visitFunctionTerm(Object closure, Identifier id, AstTerm func, AstTerm form);

    /** Term abbreviation. */
    Object visitAbbreviationTerm(Object closure, Identifier name);

    /* --- Sequences and sets --- */

    Object visitSetTerm (Object closure, AstTerm[] elems);

    Object visitSequenceTerm (Object closure, AstTerm[] elems, AstTerm tail);

    /* --- Terms of dynamic logic --- */

    /** Term with modal box operator: <tt>[rule] term</tt>. */
    Object visitBox(Object closure, AstAsmRule rule, AstTerm term, AstTerm step);

    /** Term with modal diamond operator: <tt>&lt;rule&gt;
     * term</tt>. */
    Object visitDiamond(Object closure, AstAsmRule rule, AstTerm term, AstTerm step);

    /** Term with modal def operator: <tt>def(rule)</tt>. */
    Object visitDef(Object closure, AstAsmRule rule, AstTerm step);

    /** Term with modal update operator: <tt>upd(lhs := rhs in
     * rule)</tt>. */
    Object visitUpd(Object closure, AstTerm lhs, AstTerm rhs, AstAsmRule rule, AstTerm step);

    /** Term with modal Con operator: <tt>Con(rule)</tt>. */
    Object visitCon(Object closure, AstAsmRule rule, AstTerm step);

    /** Term with modal inv operator: <tt>inv(term in rule)</tt>. */
    Object visitInv(Object closure, AstTerm term, AstAsmRule rule, AstTerm step);

    /** Term with modal joinable operator: <tt>joinable(rule1 and
     * rule2)</tt>. */
    Object visitJoinable(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Term with modal equi operator: <tt>equi(rule1, rule2)</tt>. */
    Object visitEqui(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Term with modal equiS operator: <tt>equiS(rule1,
     * rule2)</tt>. */
    Object visitEquiS(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Term with modal equiA operator: <tt>equiA(rule1,
     * rule2)</tt>. */
    Object visitEquiA(Object closure, AstAsmRule rule1, AstAsmRule rule2, AstTerm step);

    /** Term with modal accT operator: <tt>accT(term1, term2)</tt>. */
    Object visitAccT(Object closure, AstTerm term1, AstTerm term2, AstTerm step);

    /** Term with modal accF operator: <tt>accF(term, formula)</tt>. */
    Object visitAccF(Object closure, AstTerm term, AstTerm formula, AstTerm step);

    /** Term with modal acc operator: <tt>acc(term, rule)</tt>. */
    Object visitAcc(Object closure, AstTerm term, AstAsmRule rule, AstTerm step);

    /* asm rules */

    /** Sequential ASM rule: <tt>rule1 ; rule2</tt>. */
    Object visitSeq(Object closure, AstAsmRule rule1, AstAsmRule rule2);

    /** Parallel ASM rule: <tt>rule1 rule2</tt>. */
    Object visitPar(Object closure, AstAsmRule rule1, AstAsmRule rule2);

    /** Skip ASM rule: <tt>skip</tt>. */
    Object visitSkip(Object closure);

    /** Update ASM rule: <tt>lhs := rhs</tt>. */
    Object visitAssign(Object closure, AstTerm lhs, AstTerm rhs);

    /** Call named ASM rule: <tt>id(terms)</tt>. */
    Object visitCall(Object closure, Identifier id, AstTerm[] terms);

    /** If ASM rule: <tt>if formula then thenRule else
     * elseRule</tt>. */
    Object visitIf(Object closure, AstTerm formula, AstAsmRule thenRule, AstAsmRule elseRule);

    /** If ASM rule: <tt>elsif</tt>. */
    Object visitElseif(Object closure, AstTerm formula, AstAsmRule thenRule, AstAsmRule elseRule);

    /** Let ASM rule: <tt>let var:sort := term in rule</tt>. */
    Object visitLet(Object closure, Identifier var, Identifier sort, AstTerm term, AstAsmRule rule);

    /** Forall ASM rule: <tt>forall var:sort with formula do
     * rule</tt>. */
    Object visitForall(Object closure, Identifier var, Identifier sort, AstTerm formula, AstAsmRule rule);

    /** Try ASM rule: <tt>try tryRule else elseRule</tt>. */
    Object visitTry(Object closure, AstAsmRule tryRule, AstAsmRule elseRule);

    /** SchemaVariable for ASM rule: <tt>var</tt>. */
    Object visitSchemaVariable(Object closure, Identifier var);

    /** Substitution for ASM rule: <tt>{var:sort = subst}
     * term</tt>. */
    Object visitSubstitution(Object closure, Identifier var, Identifier sort, AstTerm subst, AstAsmRule rule);

    /** Metaoperator for ASM rule: <tt>id(rule)</tt>. The Metaoperator
     * has only one argument, an ASM rule. */
    Object visitMeta(Object closure, Identifier id, AstAsmRule rule);
}
