// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.casetool;

import tudresden.ocl.check.types.Type;
import java.io.IOException;
import java.util.Iterator;

import recoder.abstraction.Method; 
import recoder.abstraction.Field; 
import recoder.abstraction.ClassType; 
import recoder.list.CompilationUnitList;
import recoder.list.MethodList;
import recoder.list.FieldList;
import recoder.list.StringArrayList;
import recoder.list.TypeList;
import recoder.list.ClassTypeList;
import recoder.java.declaration.TypeDeclaration; 
import recoder.io.SourceFileRepository;
import recoder.io.DefaultSourceFileRepository;
import recoder.util.FileCollector;
import recoder.java.CompilationUnit;
import recoder.service.DefaultSourceInfo;
import recoder.ServiceConfiguration;
import java.io.File;

public class UMLOCLJavaModel implements UMLOCLModel {
    //debug flag
    private static final boolean DEBUG = false;

    protected SourceFileRepository sfr;
    protected String projectRoot;
    protected HashMapOfClassifier classifiers;
    protected HashMapOfAssociations associations;
    protected JavaOCLTypeMap typeMap = null;
    
    private UMLOCLJavaModel() {
    }

    public UMLOCLJavaModel(String projectRoot) {
	this.projectRoot  = projectRoot;
	File f = new File(projectRoot);
	this.projectRoot  = f.getAbsolutePath();
	this.classifiers  = new HashMapOfClassifier();
	this.associations = new HashMapOfAssociations();
	initRecoder();
    }

    protected JavaOCLTypeMap getJavaOCLTypeMap
	(HashMapOfClassifier classifier) {
	
	return new JavaOCLTypeMap(classifiers);
    }
    
    public HashMapOfClassifier getUMLOCLClassifiers() {
	String[] files = collectJavaFiles(projectRoot);
	try{
	    for (int i=0; i<files.length; i++) {
                if (files[i] != null) {
		    sfr.getCompilationUnitFromFile(files[i]);
                }
            }
	} catch (recoder.ParserException e){
            System.out.println("  Exception occured while "+
			       "reading in Java sources: " + e);
        }
	CompilationUnitList cuList = sfr.getCompilationUnits();
	
	// find classes
	findAllClasses(cuList);
	if (DEBUG) {
	    System.out.println("------------- found the following "+
			       "classes ------------------");
	    Iterator it = classifiers.values().iterator();
	    UMLOCLClassifier c;
	    while (it.hasNext()) {
		c = (UMLOCLClassifier)it.next();
		System.out.println(c.getName());
	    }
	}

	typeMap = getJavaOCLTypeMap(classifiers);
	
	// find attributes and methods
	CompilationUnit unit;
	for (int i=0; i < cuList.size(); i++) {
	    unit = cuList.getCompilationUnit(i);
	    TypeDeclaration td;
	    for (int j = 0; j < unit.getTypeDeclarationCount(); j++) {
		td = unit.getTypeDeclarationAt(j);	
		findAllAttributes(td);		
		findAllMethods(td);
	    }
	}
	if (DEBUG) {
	    System.out.println("------------- found the following "+
			       "methods and attributes ------------------");
	    Iterator cit = classifiers.values().iterator();
	    UMLOCLClassifier c;
	    while (cit.hasNext()) {
		c = (UMLOCLClassifier)cit.next();
		HashMapOfFeatures feat = c.features();
		Iterator it = feat.values().iterator();
		UMLOCLFeature f;
		while (it.hasNext()) {
		    f = (UMLOCLFeature)it.next();
		    if (f instanceof UMLOCLBehaviouralFeature)
			System.out.println("Method    "+c.getName()+
					   " :: "+f.getName()+"  "+f.getType());
		    if (f instanceof UMLOCLStructuralFeature)
			System.out.println("Attribute "+c.getName()+
					   " :: "+f.getName());
		} 
	    }
	}
	return classifiers;
    }


    protected UMLOCLClassifier createUMLOCLClassifier(ClassType ct) {
	UMLOCLClassifier cls = new UMLOCLClassifier(ct.getName(),
						    ct.getFullName());
	return cls;
    }


    protected void findAllClasses(CompilationUnitList cuList) {
	UMLOCLClassifier supertype;
	ClassType st;
	UMLOCLClassifier cls;
	for (int i=0; i < cuList.size(); i++) {
	    CompilationUnit pa = cuList.getCompilationUnit(i);
	    for (int j = 0; j < pa.getTypeDeclarationCount(); j++) {
		TypeDeclaration td = pa.getTypeDeclarationAt(j);
		if (!classifiers.containsKey(td.getName())) {			
		    cls = createUMLOCLClassifier(td);
		    classifiers.putClassifier(td.getName(),cls);
		}
		else 
		    cls = classifiers.getClassifier(td.getName());

		// now find all supertypes of this class
		ClassTypeList supertypes = td.getAllSupertypes();
		for (int k = 1; k < supertypes.size(); k++) {
		    st = supertypes.getClassType(k);
		    if (classifiers.containsKey(st.getName())) {			
			cls.addSupertype(st.getName(), 
					 classifiers.getClassifier(st.getName()));
		    } else {
			supertype = createUMLOCLClassifier(st);
			classifiers.putClassifier(st.getName(), supertype);
			cls.addSupertype(st.getName(), supertype);
		    }
		}

	    }	
	}
    }

    private void findAllMethods(TypeDeclaration td) {
	String nameString;
	UMLOCLClassifier cls = classifiers.getClassifierByFullName(td.getFullName());

	MethodList ml = td.getAllMethods();
	Method meth;
	for (int l=0; l < ml.size(); l++) {
	    meth = ml.getMethod(l);
	    recoder.abstraction.Type retType = meth.getReturnType();
	    tudresden.ocl.check.types.Type ret2 = null;
	    if (typeMap.getTypeFor(retType, classifiers) != null) {
		ret2 = typeMap.getTypeFor(retType, classifiers);
	    } else {
		ret2 = new UMLOCLClassifier(retType.getName(), retType.getFullName());
		classifiers.putClassifier(retType.getName(), (UMLOCLClassifier) ret2);
	    }
	    nameString = new String(meth.getName() + "(");
	    TypeList ctl = meth.getSignature();
	    Type[] paramsArray = new Type[ctl.size()];
	    recoder.abstraction.Type par;
	    for (int k=0; k < ctl.size(); k++) {		
		par = ctl.getType(k);
		if (k < ctl.size() - 1) {
		    nameString = nameString + par.getFullName() + ",";
		} else {
		    nameString = nameString + par.getFullName();
		}
		if (typeMap.getTypeFor(par, classifiers) != null) {
		    paramsArray[k] = typeMap.getTypeFor(par, 
							classifiers);
		    
		} else{
		    UMLOCLClassifier c = new 
			UMLOCLClassifier(par.getName(), par.getFullName());
		    classifiers.putClassifier(par.getName(), c);
		    paramsArray[k] = typeMap.getTypeFor(par, 
							classifiers);
		}
	    }
	    nameString = nameString + ")";
	    if (retType!=null) {
		cls.addFeature(nameString, new UMLOCLBehaviouralFeature
		    (nameString,ret2,paramsArray));
	    } else {
		cls.addFeature(nameString, new UMLOCLBehaviouralFeature
		    (nameString,paramsArray));
	    }	    
	}	
    }
    
	
    private void findAllAttributes(TypeDeclaration td) {
	FieldList fl = td.getAllFields();
	Field field;
	recoder.abstraction.Type fieldType;
	for (int l=0; l < fl.size(); l++) {
	    field = fl.getField(l);
	    fieldType = field.getType();
	    Type type = typeMap.getTypeFor(fieldType, classifiers);	    
	    if (type == null) {
		UMLOCLClassifier c = new 
		    UMLOCLClassifier(fieldType.getName(), 
				     fieldType.getFullName());
		classifiers.putClassifier(fieldType.getName(), c);
		type=c;
	    }
	    UMLOCLClassifier clf
		= classifiers.getClassifierByFullName(td.getFullName());
	    if (clf==null) {		
		throw new IllegalStateException("Classifier not found: "
						+td.getFullName());
	    }
	    clf.addFeature(field.getName(), new UMLOCLStructuralFeature
		    (field.getName(), type));
	}		
    }

    // replace recoder2key by pure recoder 
    private void initRecoder() {
	de.uka.ilkd.key.java.Services services = 
	    new de.uka.ilkd.key.java.Services();
	de.uka.ilkd.key.java.Recoder2KeY r2k = 
	    new de.uka.ilkd.key.java.Recoder2KeY
		(services, new de.uka.ilkd.key.logic.NamespaceSet());
	// comment this if OCLDL does not needs to access the java classes
	r2k.parseSpecialClasses();
	ServiceConfiguration crsc = 
	    services.getJavaInfo().getKeYProgModelInfo().getServConf();
	sfr = crsc.getSourceFileRepository();
	new DefaultSourceInfo(crsc);
    }

    private String[] collectJavaFiles(String dir) {
 	FileCollector col = new FileCollector(dir);
 	StringArrayList list = new StringArrayList();
 	while (col.next(DefaultSourceFileRepository.JAVA_FILENAME_FILTER)) {
 	    String path;
 	    try {
 		path = col.getFile().getCanonicalPath();
 	    } catch (IOException ioe) {
 		path = col.getFile().getAbsolutePath();
 	    }
 	    list.add(path);
 	}
 	String[] result = new String[list.size()];
 	for (int i = list.size() - 1; i >= 0; i -= 1) {
 	    result[i] = list.getString(i);
 	}
 	return result;
	
    }

}
