// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.reuse.*;
import de.uka.ilkd.key.rule.*;
import java.util.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.Debug;
import org.apache.log4j.Logger;


public class ReuseListener implements KeYSelectionListener {

   private KeYMediator medi;
   private Logger reuseLogger = Logger.getLogger("key.proof.reuse");
   
   private List reusePoints;
   private ReusePoint best;
   
   private ReuseFindTaclet findTacletLogic;
   private ReuseUpdateSimplificationRule updateSimpRuleLogic;
   
   
   public ReuseListener(KeYMediator medi) {
      this.medi = medi;
      initLogicNewProof();
   }

   private void initLogicNewProof() {
      reusePoints = new LinkedList();
      findTacletLogic = new ReuseFindTaclet(medi, reusePoints);
      updateSimpRuleLogic = new ReuseUpdateSimplificationRule(medi, reusePoints);
   }
   
   
   
   public void selectedNodeChanged(KeYSelectionEvent e) {
      if (medi.reuseInProgress()) return; // we are still in the loop
      Node target = e.getSource().getSelectedNode();
      if (medi.getProof().getGoal(target) == null) return; // inner node
      Node targetParent = target.parent();
      if (targetParent == null) return;
      ReusePoint rp = targetParent.getReuseSource();
      if (rp == null) return;
      int siblingNr = targetParent.getChildNr(target);
      Node source = rp.source().child(siblingNr);
      try{
         Main.getInstance().setStatusLine("Template proof continues with: "+
                                          source.name());
//         source.getAppliedRuleApp().rule().name());
      } catch (NullPointerException npe) {
      // there is no ruleApp at this point in the source proof (=open goal)
      // forget about the hint then
      ;
      }
   
   }

   public void selectedProofChanged(KeYSelectionEvent e) {
      recomputeReuseTotal();
      if (reusePossible()) medi.indicateReuse(getBestReusePoint());
        else medi.indicateNoReuse();
   }


   
   private void recomputeReuseTotal() {
      initLogicNewProof();
      Iterator markerIt = Node.reuseCandidatesIterator();
      while (markerIt.hasNext()) {
         Node marker = (Node) markerIt.next();
            IteratorOfGoal goalIt = medi.getProof().openGoals().iterator();
//         reuseLogger.info("***********************************************");
            while (goalIt.hasNext()) {
               analyzeCandidate(marker, goalIt.next());
            }
      }
   }



   
//remove RP with consumed goal
   public void removeRPConsumedGoal(Goal usedGoal) {
      ListIterator it = reusePoints.listIterator();
      while (it.hasNext()) {
         ReusePoint rp = (ReusePoint) it.next();
         if ((rp.target()==usedGoal)) {
            it.remove();
         }
      }
   }


//add: old markers - new goals; call after removeRPConsumedMarker()
   public void addRPOldMarkersNewGoals(ListOfGoal newGoals) {
      Iterator markerIt = Node.reuseCandidatesIterator();
      while (markerIt.hasNext()) {
         Node marker = (Node) markerIt.next();
         IteratorOfGoal goalIt = newGoals.iterator();
         while (goalIt.hasNext()) {
            analyzeCandidate(marker, goalIt.next());
         }
      }
   }


   
   
   public void removeRPConsumedMarker(Node usedMarker) {
      usedMarker.unmarkReuseCandidate();

//remove RP with consumed marker
      ListIterator it = reusePoints.listIterator();
      while (it.hasNext()) {
         ReusePoint rp = (ReusePoint) it.next();
         if ((rp.source()==usedMarker)) {
            it.remove();
         }
      }
   }


//add RP: new markers - all goals
   public void addRPNewMarkersAllGoals(Node usedMarker) {
      IteratorOfNode newMarkerIt = usedMarker.childrenIterator();
      while (newMarkerIt.hasNext()) {
         Node n = newMarkerIt.next();
         if ("Closed goal".equals(n.name())) continue; // this is ugly
         if ("OPEN GOAL".equals(n.name())) continue;
//         while ("Update Simplification".equals(n.name())) {
//            n = n.child(0);
//         }
         n.markReuseCandidate();
         IteratorOfGoal goalIt = medi.getProof().openGoals().iterator();
//         reuseLogger.info("***********************************************");
         while (goalIt.hasNext()) {
            Goal g = goalIt.next();
            analyzeCandidate(n, g);
         }
      }
   }
   
   
   private ReusePoint highestScored(List l) {
      ReusePoint max = (ReusePoint) l.get(0);
      ListIterator it = l.listIterator();
      while (it.hasNext()) {
         ReusePoint curr = (ReusePoint) it.next();
         if (curr.score()>max.score()) max=curr;
      }
      return max;
   }



   public boolean reusePossible() {
      reuseLogger.warn(">Possible applications detected: "+reusePoints.size());
      if (reusePoints.size() > 0) {
         best = highestScored(reusePoints);
         return !best.notGoodEnough();
      } else {
         return false;
      }
   }

   
   public ReusePoint getBestReusePoint() {
       reuseLogger.info("############## RE-USING "+
           best.getApp().rule().name()+" "+best.score());
       return best;
   }
   
   
   void analyzeCandidate(Node source, Goal target) {
      reuseLogger.info("***** (next goal) candidate nodes in system: "+
                       Node.reuseCandidatesNumber());
      reuseLogger.info("goal: "+target.node().sequent().toString());
                       
      ReusePoint blank = new ReusePoint(source, target);
      if ( blank == null || blank.getApp() == null ) return; // %%%%
      Rule r = blank.getApp().rule();
      if (r instanceof FindTaclet) {
          findTacletLogic.applicableWhere(blank);
      } else if (r instanceof NoFindTaclet) {
          NoPosTacletApp app = NoPosTacletApp.createNoPosTacletApp((Taclet)r);
          ReusePoint p = blank.initializeNoFind(app, medi);
          if ((p != null)){ // RuleApp can be transferred // && goodEnough!
             reusePoints.add(p);
          }
      } else if (r instanceof UpdateSimplificationRule) {
          updateSimpRuleLogic.applicableWhere(blank);
      } else {
          String ruleName = (blank == null) ? 
              "null" : blank.getApp().rule().getClass().toString();
          reuseLogger.error("Cannot re-use "+ruleName);
      }

   }
   


}
