// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.gui;

import java.util.Properties;
import java.util.LinkedList;
import java.util.Iterator;
import org.apache.log4j.Logger;

/** 
 * This class encapsulates information about:
 * 1) relative font size in the prover view
 * 2) the maximal number of lines a tooltip with instantiated SchemaVariables
 *    is allowed to have. If this number is exceeded no SchemaVariables get
 *    instantiated in the displayed tooltip.
 */
public class ViewSettings implements Settings {
	static Logger logger = Logger.getLogger(ViewSettings.class.getName());
    private static final String MAX_TOOLTIP_LINES_KEY = "[View]MaxTooltipLines";
    private static final String SHOW_WHOLE_TACLET = "[View]ShowWholeTaclet";
    private static final String FONT_INDEX = "[View]FontIndex";


    /** default max number of displayed tooltip lines is 40 */
    private int maxTooltipLines = 40; 
    /** do not print the find, varcond and heuristics part of taclets in
     * the TacletMenu by default */
    private boolean showWholeTaclet = false;
    /** default fontsize */
    private int sizeIndex = 2;


    private LinkedList listenerList = new LinkedList();


    /**
     * @return the current maxTooltipLines
     */
    public int getMaxTooltipLines() {
    	return maxTooltipLines;
    }

    /**
     * Sets maxTooltipLines 
     * @param b The new value for maxTooltipLines 
     */
    public void setMaxTooltipLines(int b) {
		maxTooltipLines = b;
		fireSettingsChanged();
    }
    
    /**
	 * returns whether the Find and VarCond part of Taclets should be
	 * pretty-printed with instantiations of schema-variables or not
	 * 
	 * @return true iff the find part should be pretty-printed instantiated
	 */
    public boolean getShowWholeTaclet() {
    	return showWholeTaclet;
    }
    
    /**
	 * Sets whether the Find and VarCond part of Taclets should be
	 * pretty-printed with instantiations of schema-variables or not
	 * 
	 * @param b
	 *           indicates whether the Find and VarCond part of Taclets should
	 *           be pretty-printed with instantiations of schema-variables or
	 *           not
	 */
    public void setShowWholeTaclet(boolean b) {
    	showWholeTaclet = b;
    	fireSettingsChanged();
    }
    
    /**
     * @return the current sizeIndex
     */
    public int sizeIndex() {
    	return sizeIndex;
    }

     /**
     * Sets FontIndex
     * @param b The new value for SizeIndex 
     */
    public void setFontIndex(int b) {
		sizeIndex = b;
		fireSettingsChanged();
    }


    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     * @param The aforementioned Properties object
     */
    public void readSettings(Properties props) {
		String val1 = props.getProperty(MAX_TOOLTIP_LINES_KEY);
		String val2 = props.getProperty(FONT_INDEX);
		String val3 = props.getProperty(SHOW_WHOLE_TACLET);
		if (val1 != null) {
			setMaxTooltipLines(Integer.valueOf(val1).intValue());
		} else {
			setMaxTooltipLines(maxTooltipLines);
		}
		if (val2 != null) {
			setFontIndex(Integer.valueOf(val2).intValue());
		} else {
			setFontIndex(sizeIndex);
		}
		if (val3 != null) {
			setShowWholeTaclet(Boolean.valueOf(val3).booleanValue());
		}
	}

  
    /**
	 * implements the method required by the Settings interface. The settings
	 * are written to the given Properties object. Only entries of the form
	 * <key>=<value>(, <value>)* are allowed.
	 * 
	 * @param props
	 *           the Properties object where to write the settings as (key,
	 *           value) pair
	 */
    public void writeSettings(Properties props) {
    	props.setProperty(MAX_TOOLTIP_LINES_KEY, "" + maxTooltipLines);
    	props.setProperty(SHOW_WHOLE_TACLET, "" + showWholeTaclet);
		props.setProperty(FONT_INDEX, "" + sizeIndex);
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** 
     * adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }
    
}
