// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java;

import java.io.*;
import java.util.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.annotation.Annotation;

/**
 *  Top level implementation of a Java {@link ProgramElement}.
 * taken from COMPOST and changed to achieve an immutable structure
 */

public abstract class JavaProgramElement extends JavaSourceElement
                                         implements ProgramElement {

    /**
 *        Comments.
    */

    protected final Comment[] comments;
    

    /**
 *      Java program element.
     */

    public JavaProgramElement() {
	comments=null;
    }

    /**
     * Java program element. 
     * @param list ExtList with comments
     */

    public JavaProgramElement(ExtList list) {
	super(list);
	comments=(Comment[])list.collect(Comment.class);
    }

    /**
     *      Get comments.
     *      @return the comments.
     */

    public Comment[] getComments() {
        return comments;
    }

    /**
     *@return the annotations.
     */
    public Annotation[] getAnnotations(){
	return new Annotation[0];
    }

    public int getAnnotationCount(){
	return 0;
    }

    /**
     *      Template Method. Prints DocComments if existing, calls
     *      prettyPrintMain and prints other Comments.
     */

    public void prettyPrint(PrettyPrinter w) throws IOException {
        int s = (comments != null) ? comments.length : 0;
        int t = 0;
        for (int i = 0; i < s; i += 1) {
            Comment c = comments[i];
            if (c.isPrefixed()) {
                c.prettyPrint(w);
            } else {
                t += 1;
            }
        }
        prettyPrintMain(w);
        if (t > 0) {
            for (int i = 0; i < s; i += 1) {
                Comment c = comments[i];
                if (!c.isPrefixed()) {
                    if (c instanceof SingleLineComment) {
                        w.scheduleComment((SingleLineComment)c);
                    } else {
                        c.prettyPrint(w);
                    }
                }
            }
        }
    }

    /**
     *      Prints main content of current node and all syntactical children.
     *      Hook method of prettyPrint; defaults to do nothing.
     */
    protected void prettyPrintMain(PrettyPrinter w) throws IOException {}



    /** commented in interface SourceElement. The default equals
     * method compares  two elements by testing if they have the 
     * same type and calling the default equals method.
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
    	return (this.getClass() == se.getClass());
    }
    
    public int hashCode(){
    	int result = 17;
    	result = 37 * result + this.getClass().hashCode();
    	return result;
    }
    
    public boolean equals(Object o){
    	if(o == this)
    		return true;
    	if(!(o instanceof JavaProgramElement))
    		return false;
    	JavaProgramElement cmp = (JavaProgramElement)o;
    	return equalsModRenaming(cmp, NameAbstractionTableDisabled.INSTANCE);
    }

    /** this is the default implementation of the signature, which is
     *  used to determine program similarity.
     */
    public String reuseSignature(Services services) {
       String s = getClass().toString();
       return s.substring(s.lastIndexOf('.')+1, s.length());
    }
    
    /** this class is used by method call. As in this case we do not
     * want to abstract from  names
     */
    static class NameAbstractionTableDisabled extends NameAbstractionTable {
	
    public static final NameAbstractionTableDisabled INSTANCE = new NameAbstractionTableDisabled();
    	
	public void add(SourceElement pe1, 
			SourceElement pe2) {
	}

	public boolean contains(SourceElement pe) {
	    return false;
	    
	}
	
	public boolean sameAbstractName(SourceElement pe1, 
					SourceElement pe2) {
	    return false;
	}
    }

}
