// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.java.statement.MethodFrame;
import de.uka.ilkd.key.java.NonTerminalProgramElement;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.reference.SimpleStackOfIExecutionContext;
import de.uka.ilkd.key.java.reference.ExecutionContext;

import de.uka.ilkd.key.logic.PosInProgram;

public class JavaIterator {
    

    private ProgramElement root;
    private PosTracker rootPos;
    
    /** 
     * stack which contains the execution contexts from the top up to
     * the current position of the branch
     */
    private SimpleStackOfIExecutionContext executionContextStack;


    private PosInProgram pip;

    /**
     * the top element of the position stack
     */
    private PosTracker top;    


    public JavaIterator(ProgramElement root) {
	this.root = root;
	rootPos   = new PosTracker(root, -1);

	pip = PosInProgram.TOP;
	top = null;
	executionContextStack = new SimpleStackOfIExecutionContext();
	push(rootPos);
    }

    /**
     * push pt on the stack and updates the execution context stack
     * @param pt the PosTracker object 
     */
    private void push(PosTracker pt) {

	if (top != null) pip = pip.down(top.pos());
	pt.setParent(top);		

	if (pt.node() instanceof MethodFrame) {
	    executionContextStack.push(((MethodFrame)pt.node()).getExecutionContext());
	}

	top = pt;

    }

    /**
     * performs a pop pt on the position stack, updates the execution
     * context stack and returns the popped element
     * @return the popped element
     */
    private PosTracker pop() {
	final PosTracker pt = top;
	top = top.parent();

	if (pt.node() instanceof MethodFrame) {
	    executionContextStack.pop();
	}

	pip = pip.up();
	return pt;
    }
   

    /** returns the position of the element relative to its parent
     * @return returns the position of the element relative to its parent
     */
    public int relativePos() {
	final ProgramElement current = element();
	if (current != null && current != root) {
	    return top.parent().pos();
	}
	throw new ArrayIndexOutOfBoundsException();
    }



    /** returns the enclosing parent of the current element 
     * @return the enclosing parent of the current element 
     */
    public ProgramElement parent() {
	final ProgramElement current = element();
	if (current != null && current != root) {
	    return top.parent().node();
	}
	return null;
    }


    /**
     * sets back to the parent and returns the iterator;
     */
    public JavaIterator up() {
	final ProgramElement current = element();
	if (current != null && current != root) {
	    pop();	    
	}
	return this;	
    }

    /**
     *
     */
    public void restartAtCurrentPosition() {
	top.setPos(-1);
    }



    /** skips the subtree starting at this element and goes to the
     * next sibling or if not available to the next possible sibling
     * of one of the ancestors of this node     
     */
    public ProgramElement skip() {
	if (hasNextSibling() || element() != root) {
	    pop();
	} else if (hasNext()) {
	    top.setLast();
	}	           
	return next();	
    }
    
    /** moves the iterator to the next node (depth-left-first)  */
    public ProgramElement next() {	
	while (hasNext()) {
	    top.incPos();
	    if (top.hasNext()) {
		push
		    (new PosTracker(((NonTerminalProgramElement)element()).
				    getChildAt(top.pos()), -1));
		break;
	    } else if (top != root) { // do not remove root from
		                      // stack 
		pop();
	    }
	}
	return element();
    }

    /** moves the iterator to the next sibling of the node. Iterator
     * stays at the current node if no next sibling exists
     */
    public ProgramElement nextSibling() {	
	if (hasNextSibling()) {
	    pop();
	    return next();
	}       	
	return element();
    }

    /** returns true if next sibling exists */
    public boolean hasNextSibling() {
	return element() == root ? false : 
	    top.parent().pos() < top.parent().getChildCount() - 1;
    }

    /** moves the iterator to the previous sibling of the node. Iterator
     * stays at the current node if no previous sibling exists
     */
    public ProgramElement prevSibling() {		
	if (hasPrevSibling()) {
	    pop();
	    if (top != null) {
		top.decPos();
	    }
	    push(new PosTracker
		(((NonTerminalProgramElement)element()).getChildAt(top.pos()), -1));	    
	}
	return element();
    }

    /** returns true if next sibling exists */
    public boolean hasPrevSibling() {
	return root == element() ? false : top.parent().pos() > 0;
    }    

    /** moves the iterator to the previous node (depth-left-first)  */
    public ProgramElement back() {	
	ProgramElement current = element();
	if (current == null) { // the end had been reached
	    top = rootPos;
	    goToRightLeave();
	} else if (!atTop()) {
	    top.decPos();
	    if (top.pos() < -1) {
		pop();
		back();
	    } else if (top.pos() > -1) {
		// go to the most right leave
		push
		    (new PosTracker(((NonTerminalProgramElement)current).
				    getChildAt(top.pos()), -1));
		goToRightLeave();
	    }
	}
	return element();
    }

    /** goes to the rightest leave in the current subtree */
    private void goToRightLeave() {	
	do { 
	    top.setPos(top.getChildCount() - 1);
	    if (top.pos() != -1) {
		push(new PosTracker		     
		     (((NonTerminalProgramElement)top.node()).getChildAt(top.pos()), -1));
	    } else {
		break;
	    }
	} while (top.getChildCount() > 0);
    }
    

    /** 
     * returns the position in program of the current element. It is
     * null if the end of the program has been reached. 
     */
    public PosInProgram position() {
	return hasNext() ? pip : null;
    }

    /**
     * returns the last execution context found above the current
     * position
     */
    public ExecutionContext lastExecutionContext() {
	return (ExecutionContext) 
	    (executionContextStack.empty() ? 
	     null : executionContextStack.peek());
    }
    
    /** 
     * returns the current node where the iterator stand or null if
     * the end of the programelement to iterate through has been
     * reached 
     */
    public ProgramElement element() {
	return hasNext() ? top.node() : null;
    }

    /** returns true if there is a non visited element */
    public boolean hasNext() {
	return rootPos.hasNext();
    }

    /** returns true if the iterator is back at root */
    public boolean atTop() {
	return rootPos.pos() == -1;
    }

    
    private static class PosTracker {

	/**
	 * reference to the next element in stack
	 */
	private PosTracker parent;

	/**
	 * 
	 */
	private ProgramElement pe;

	/**
	 * max. amount of children
	 */
	private int childCount;

	/**
	 * the current visited child
	 */
	private int currentChildPos;
	
	
	/** 
	 * the program element and the number of 
	 * the child that is visited 
	 */
	public PosTracker(ProgramElement pe, 
			  int currentChildPos) {
	    this.pe = pe;
	    childCount = pe instanceof NonTerminalProgramElement ? 
		((NonTerminalProgramElement)pe).getChildCount() : 0;
	    this.currentChildPos = currentChildPos;
	}


	/** 
	 * the program element and the number of 
	 * the child that is visited 
	 */
	public PosTracker(NonTerminalProgramElement pe, 
			  int currentChildPos) {
	    this.pe = pe;
	    childCount = pe.getChildCount();
	    this.currentChildPos = currentChildPos;
	}

	public void setParent(PosTracker parent) {
	    this.parent = parent;
	}
 	public ProgramElement node() {
	    return pe;
	}

 	public PosTracker parent() {
	    return parent;
	}

	public int pos() {
	    return currentChildPos;
	}

	public void setLast() {
	    if (pe instanceof NonTerminalProgramElement) {
		currentChildPos = childCount;
	    } else {
		currentChildPos = -1;
	    }
	}

	public int getChildCount() {
	    return childCount;
	}

	public boolean hasNext() {
	    return currentChildPos < childCount;
	}
	
	public void setPos(int n) {
	    currentChildPos = n;
	}

	public void incPos() {
	    currentChildPos++;
	}

	public void decPos() {
	    currentChildPos--;
	}
	
	public String toString() {
	    return "PE:"+pe+"|| Pos:"+currentChildPos;
	}

    }   

}
