// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.java.visitor;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.logic.PosInProgram;
import de.uka.ilkd.key.logic.IntIterator;
import de.uka.ilkd.key.util.Debug;


/** 
 * A context given as {@link ContextStatementBlockInstantiation} is wrapped 
 * around a given {@ProgramElement}. 
 */
public class ProgramContextAdder {
    
    /**
     * singleton instance of the program context adder
     */
    public static ProgramContextAdder INSTANCE = new ProgramContextAdder();

    /**
     * an empty private constructor to ensure the singleton property
     */
    private ProgramContextAdder() {}
    
    /**
     * wraps the context around the statements found in the putIn block   
     */
    public JavaNonTerminalProgramElement start
	(JavaNonTerminalProgramElement   context,
	 StatementBlock                  putIn, 
	 ContextStatementBlockInstantiation ct) {

	return wrap(context, putIn, 
		    ct.prefix().iterator(), 
		    ct.prefix().depth(),
		    ct.suffix());
    }

    protected JavaNonTerminalProgramElement wrap
	(JavaNonTerminalProgramElement      context,
	 StatementBlock                     putIn, 
	 IntIterator                        prefixPos,
	 int                                prefixDepth,
	 PosInProgram                       suffix) {

	JavaNonTerminalProgramElement body = null;


	ProgramElement next = context.getChildAt(prefixPos.next());

	if (!prefixPos.hasNext()) {
	    body = createWrapperBody
		((JavaNonTerminalProgramElement)context, 
		 putIn, prefixDepth, suffix);
	    // special case labeled statement as a label must not be
	    // succeeded by a statement block
	    if (context instanceof LabeledStatement) {
		body = createLabeledStatementWrapper
		    ((LabeledStatement)context, body);
	    } 
	    return body;
	} else {    
	    body = wrap((JavaNonTerminalProgramElement)next, 
			putIn, 
			prefixPos,
			prefixDepth, 
			suffix);
	    if (context instanceof StatementBlock) {	   
		return createStatementBlockWrapper((StatementBlock)context,
						   body); 
	    } else if (context instanceof Try) {
		return createTryStatementWrapper((StatementBlock)body, 
						 (Try)context);
	    } else if (context instanceof MethodFrame) {
		return createMethodFrameWrapper((MethodFrame)context,
						(StatementBlock)body);
	    }  else if (context instanceof LabeledStatement) {
		return createLabeledStatementWrapper
		    ((LabeledStatement)context, body);
	    } else {
		return createSynchronizedBlockWrapper
		    ((SynchronizedBlock)context, (StatementBlock)body);
	    }
	}
    }

    /** 
     * <p>The context instantiation term contains a position list to the     
     * last element that is not in the context.</p><p>
     * We need the entry position to this child in the enclosing context
     * braces, for example in <code> {{{{ int a; a = a +1; }}}} </code>
     * assuming the first element not hidden context in the
     * context is <code> int a; </code> and the next statement 
     * <code> a = a + 1; </code> disappears in the suffix of the
     * context. This method returns then <code> 0 </code> as this is the
     * position of <code> int a; </code> in the enclosing context
     * braces. </p>
     */
    private int lastChildNotInContext(PosInProgram suffix, int prefixDepth) {
	IntIterator it = suffix.iterator();
	int addChildrenAfter = 0;	
	int depthCounter     = prefixDepth;
	// Decending to the depth given by prefix depth. 
	// Special case: if prefix depth is zero, we have to descend to
	// depth one (that is why we have here a do-while loop
	do {
	    addChildrenAfter = it.next();
	} while ((--depthCounter) > 0);

	return addChildrenAfter;
    }

    /**
     * inserts the content of the statement block <code>putIn</code> and
     * adds succedding children of the innermost non terminal element
     * (usually statement block) in the context.
     * @param wrapper the JavaNonTerminalProgramElement with the context
     *   that has to be wrapped around the content of <code>putIn</code>
     * @param putIn the StatementBlock with content that has to be
     * wrapped by the elements hidden in the context 
     * @param prefixDepth the int describing the depth of the prefix
     * (i.e. the size of the position list to the first element not
     * belonging to the prefix itself)
     * @param suffix the PosInProgram describing the position of the
     * first element before the suffix of the context 
     * @return the StatementBlock which encloses the content of
     * <code>putIn</code> together with the succeeding context elements
     * of the innermost context statemebt block (attention:
     * in a case like 
     *   <code>{{{oldStmnt; list of further stmnt;}} moreStmnts; }</code> 
     * only the underscored part is returned
     *   <code>{{ __{putIn;....}__ }moreStmnts;}</code>
     * adding the other braces including the <code>moreStmnts;</code>
     * part has to be done elsewhere.
     */
    private final StatementBlock createWrapperBody
	(JavaNonTerminalProgramElement wrapper,
	 StatementBlock putIn, int prefixDepth, PosInProgram suffix) {


	final int putInLength   = putIn.getChildCount();
	final int lastChild     = lastChildNotInContext(suffix,
							prefixDepth);
	final int childLeft     = 
	    (wrapper.getChildCount() <= lastChild ? 0 : 
	     wrapper.getChildCount() - (lastChild + 1));


	final int childrenToAdd = putInLength + childLeft;

	if (childLeft == 0) { 
	    return putIn;
	}
	
	final Statement[] body = new Statement[childrenToAdd]; 
	
	putIn.getBody().arraycopy(0, body, 0, putInLength);


	for (int i = putInLength; i < childrenToAdd; i++) {
	    body[i] = (Statement) wrapper.getChildAt
		((lastChild + 1) + (i - putInLength));
	}

	/*
	  Example: 
           In <code>{{{oldStmnt; list of further stmnt;}} moreStmnts; }</code>
	   where <code>oldStmnt;</code> has to be replaced by the content of
	   <code>putIn</code>. Up to here (including the return
	   statement) the underscored part has been created:
	         <code>{{ __{putIn;....}__ }moreStmnts;}</code>
	   Attention: we have not yet added the enclosing braces or even
	   the <code>moreStmnts</code>
	 */
	return new StatementBlock(new ArrayOfStatement(body));
    }

    /**
     * Replaces the first statement in the wrapper block. The
     * replacement is optimized as it just returns the replacement block
     * if it is the only child of the statement block to be constructed
     * and the chld is a statementblock too.
     * @param wrapper the StatementBlock where to replace the first
     * statement
     * @param replacement the StatementBlock that replaces the first
     * statement of the block 
     * @return the resulting statement block
     */
    protected StatementBlock createStatementBlockWrapper
	(StatementBlock wrapper, JavaNonTerminalProgramElement replacement) {
	int childrenCount = wrapper.getStatementCount();
	if (childrenCount <= 1 && replacement instanceof StatementBlock) {
	    return (StatementBlock) replacement;
	} else {	    
	    Statement[] body = new Statement[childrenCount > 0 ?
					     childrenCount : 1];
	    /* reconstruct block */
	    body[0] = (Statement)replacement;
	    if (childrenCount - 1 > 0) {
		wrapper.getBody().arraycopy(1, body, 1, childrenCount - 1);
	    }
	    return new StatementBlock(new ArrayOfStatement(body));
	}
    }

    protected Try createTryStatementWrapper(StatementBlock body, 
					    Try old) {
	return new Try(body, old.getBranchList());
    }

    protected MethodFrame createMethodFrameWrapper
	(MethodFrame old, StatementBlock body) {
	return new MethodFrame(old.getProgramVariable(),
			       old.getExecutionContext(),
			       body,
                               old.getProgramMethod());
    }

    protected LabeledStatement createLabeledStatementWrapper
	(LabeledStatement old, JavaNonTerminalProgramElement body) {
	return new LabeledStatement(old.getLabel(),
				    body instanceof StatementBlock 
				    && body.getChildCount() == 1 && 
				    !(body.getChildAt(0) 
				      instanceof LocalVariableDeclaration) ? 
				    (Statement)body.getChildAt(0) : 
				    (Statement)body);
    }

    protected SynchronizedBlock createSynchronizedBlockWrapper
	(SynchronizedBlock old, StatementBlock body) {
	return new SynchronizedBlock(old.getExpression(), body);
    }

}


