// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

/** this class describes a position in a term 
 */
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.util.Debug;


public class PosInTerm implements java.io.Serializable {
  
    /** top level term position */
    public static PosInTerm TOP_LEVEL = new PosInTerm();

    private final PosInTerm prev;
    
    /** 
     * the position number
     */
    private final int pos;

    /** 
     * the depth
     */
    private final int depth;

    private final int hashCode;

    /**
     * iterator cache
     */
    private ListOfInteger cache;

    private PosInTerm(PosInTerm pit, int posNr) {
	prev = pit;
	pos = posNr;
	depth = pit.depth + 1;
	hashCode = prev.hashCode * 715 + pos;
    }   

    /**
     * creates a new PosInTerm 
     * position. 
     */
    private PosInTerm() {
	pos = -1;
	prev = null;
	depth = 0;
	hashCode = 13;
    }   

    /** size of the position list */
    public int depth() {
	return depth;
    }

  
    /** descending downwards choosing the n'th subterm, subformula 
     * @param n int describs the chosen subterm 
     * @return Position like old Position but with a deeper
     * subterm.
     */    
    public PosInTerm down(int n) {
	return new PosInTerm(this, n);
    }
    
    public PosInTerm up() {
       return prev;
    }


    /** 
     * compares this PosInTerm with another PosInTerm
     * and returns true if both describe the same position 
     */
    public boolean equals(Object obj) {
        if ( this == obj )
            return true;
        
	if (!(obj instanceof PosInTerm)) {
	    return false;
	} 

	PosInTerm cmp = (PosInTerm) obj;

	if (depth != cmp.depth) {
	    return false;
	}
		
	PosInTerm thisPIT = this;
    
	while ( thisPIT != TOP_LEVEL ) {
	    if ( thisPIT.pos != cmp.pos )
	        return false;
	    thisPIT = thisPIT.prev;
	    cmp     = cmp.prev;
	    if ( thisPIT == cmp )
	        return true;
	}
    
	return true;
    }

    /**
     * get subterm/-formula information.  A PosInTerm defining the xn'th
     * subterm of the x(n-1)'th subterm of ... of the x1'th subterm of a
     * term returns an iterator subsequently resulting in the list xn...x1.
     * @return IteratorOfInteger that iterates through the subterms
     * of a sequent in the reverse order as the PosInTerm has been defined.
     */ 
    public IntIterator reverseIterator() {
	return new PosIntIterator(this);
    }


    /**
     * returns an iterator defining the position in a term according to
     * subterm relation. A PosInTerm defining the xn'th subterm of the
     * x(n-1)'th subterm of ... of the x1'th subterm of a term returns an
     * iterator subsequently resulting in the list x1...xn.
     * @return an iterator over the list defining the position in a term.
     */
    public IntIterator iterator() {	
	if (cache == null) fillCache();
	return new PosArrayIntIterator(cache);
    }

    
    public static PosInTerm parseReverseString(String s) {
	PosInTerm result = TOP_LEVEL;
	if ("".equals(s)) {
	    return result;
	}
	ListOfInteger list = SLListOfInteger.EMPTY_LIST;
	java.util.StringTokenizer tker = 
	    new java.util.StringTokenizer(s, ",", false);
	while (tker.hasMoreTokens()) 
	    list = list.prepend(Integer.decode(tker.nextToken()));

	IteratorOfInteger it = list.iterator();
	while (it.hasNext()) {
	    result = new PosInTerm(result, it.next().intValue());
	}

	return result;       
    }

    
    private void fillCache() {
        cache = SLListOfInteger.EMPTY_LIST;
        IntIterator it = reverseIterator();
        while (it.hasNext()) {
	    cache = cache.prepend(new Integer(it.next()));
        }
    }

    public ListOfInteger toPositionList() {
	if (cache == null) fillCache();
	return cache;
    }

    /** toString */
    public String integerList(IntIterator it) {
	String list = "["; 
	while (it.hasNext()) {
	    list += ""+it.next();
	    if (it.hasNext()) {
		list +=","; 
	    }
	}
	list += "]";
	return list;
    }

    public String toString() {
	if (this==TOP_LEVEL) {
	    return "top level";
	}

	return "subterm: " + integerList(iterator());
    }



    private static class PosIntIterator implements IntIterator {
	private PosInTerm p;
	
	public PosIntIterator(PosInTerm p) {
	    this.p = p;
	}

	public boolean hasNext() {
	    return p != null && p != TOP_LEVEL;
	}

	public int next() {
	    Debug.assertTrue(p != TOP_LEVEL && p!=null);
	    int result = p.pos; 
	    p = p.prev;
	    return result;
	}	
    }

    private static class PosArrayIntIterator implements IntIterator {
	private IteratorOfInteger it;

	public PosArrayIntIterator(ListOfInteger pos) {
	    it = pos.iterator();
	}

	public boolean hasNext() {
	    return it.hasNext();
	}

	public int next() {
	    return it.next().intValue();
	}		
    }

    
    public int hashCode () {
        return hashCode;
    }
}
