// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.IteratorOfConstrainedFormula;
import de.uka.ilkd.key.java.Services;

import java.util.*;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.rule.SetOfInScopePair;
import de.uka.ilkd.key.rule.SetAsListOfInScopePair;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.pp.SequentPrintFilter;

/** This class represents a sequent. A sequent consists of an
 * antecedent and succedent. As a sequent is persistent there is no
 * public constructor. To create a Sequent use one of the 
 * @see{Sequent#createSequent Semisequent Semisequent}
 * @see{Sequent#createAnteSequent Semisequent}
 * @see{Sequent#createSuccSequent Semisequent}
 * methods or the @see{Sequent#EMPTY_SEQUENT} and insert 
 * @see{ConstrainedFormula}s into it.  
 */ 
public class Sequent implements java.io.Serializable {

    public static final Sequent EMPTY_SEQUENT=new Sequent();

    private Semisequent antecedent=Semisequent.EMPTY_SEMISEQUENT;
    private Semisequent succedent=Semisequent.EMPTY_SEMISEQUENT;
    
    private Sequent() {
    }    

    /** creates new Sequent with antecedence and succedence */
    private Sequent(Semisequent antecedent, Semisequent succedent) {
	this.antecedent=antecedent;
	this.succedent=succedent;	
    }

    private Object readResolve() 
	throws java.io.ObjectStreamException {
	if (antecedent == Semisequent.EMPTY_SEMISEQUENT &&
	    succedent == Semisequent.EMPTY_SEMISEQUENT){
	    return Sequent.EMPTY_SEQUENT;
	}
	return this;
    }


    /**
     * creates a new Sequent 
     * @param ante the Semisequent that plays the antecedent part
     * @param succ the Semisequent that plays the succedent part
     * @return the new sequent or the EMPTY_SEQUENT if both antec and succ
     * are same as EMPTY_SEMISEQUENT
     */
    public static Sequent createSequent(Semisequent ante, Semisequent succ) {
	return new Sequent(ante,succ).checked();
    }

    /**
     * creates a new Sequent with empty succedent 
     * @param ante the Semisequent that plays the antecedent part
     * @return the new sequent or the EMPTY_SEQUENT if both antec and succ
     * are same as EMPTY_SEMISEQUENT
     */
    public static Sequent createAnteSequent(Semisequent ante) {
	return new Sequent(ante,Semisequent.EMPTY_SEMISEQUENT).checked();
    }

    /**
     * creates a new Sequent with empty antecedent 
     * @param succ the Semisequent that plays the succedent part
     * @return the new sequent or the EMPTY_SEQUENT if both antec and succ
     * are same as EMPTY_SEMISEQUENT
     */
    public static Sequent createSuccSequent(Semisequent succ) {
	return new Sequent(Semisequent.EMPTY_SEMISEQUENT,succ).checked();
    }

    /** returns semisequent of the antecedent to work with */
    public Semisequent antecedent() {
	return antecedent;
    }

    /** returns semisequent of the succedent to work with */
    public Semisequent succedent() {
	return succedent;
    }

    /**
     * determines if the sequent is empty.
     * @return true iff the sequent consists of two instances of
     * Semisequent.EMPTY_SEMISEQUENT
     */
    public boolean empty() {
	return ((antecedent()==Semisequent.EMPTY_SEMISEQUENT) &&
		(succedent()==Semisequent.EMPTY_SEMISEQUENT));
    }

    /** returns the semisequent in which the ConstrainedFormula described
     * by PosInOccurrence p lies 
     */
    private Semisequent getSemisequent(PosInOccurrence p) {
	if (p.isInAntec()) {
	    return antecedent();  	    
	} 
	return succedent();	
    }

    /** returns new Sequent with replaced antecedent if
     * PosInOccurrence describes a ConstrainedFormula in the antecedent, 
     * the succedent is replaced otherwise. 
     * The new semisequent is seq.
     */
    private Sequent composeSequent(PosInOccurrence p, Semisequent seq) {
	return new Sequent(p.isInAntec() ? seq : antecedent(), 
			   p.isInAntec() ? succedent() : seq).checked();
    }

    /** returns creates a new sequent out of the current sequent with a
     * replaced antecedent if antec is true, otherwise the succedent is
     * replaced by the given semisequent seq. 
     */
    private Sequent composeSequent(boolean antec, Semisequent seq) {
	return new Sequent(antec ? seq : antecedent(), 
			   antec ? succedent() : seq).checked();
    }

    /**
     * returns this if the two semisequents are not the empty semisequents
     * and returns EMPTY_SEQUENT otherwise. Return the result of this method
     * in all methods that can 
     * remove elements from the Sequent, to avoid empty sequents that
     * are not the EMPTY_SEQUENT.
     * @return a Sequent that has either at least one Semisequent that is not
     * the empty semisequent or the empty sequent.
     */
    private Sequent checked() {
	return ((antecedent() == Semisequent.EMPTY_SEMISEQUENT) &&
		(succedent()  == Semisequent.EMPTY_SEMISEQUENT)) ? 
	    EMPTY_SEQUENT : this;
    }


    /** 
     * adds list of formulas to the antecedent (or succedent) of the
     * sequent. Depending on the value of first the formulas are
     * inserted at the beginning or end of the ante-/succedent.
     *  (NOTICE:Sequent determines 
     * index using identy (==) not equality.)
     * @param insertions the ListOfConstrainedFormula to be added
     * @param antec boolean selecting the correct semisequent where to
     * insert the formulas. If set to true, the antecedent is taken
     * otherwise the succedent.
     * @param first boolean if true the formulas are added at the
     * beginning of the ante-/succedent, otherwise to the end
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo addFormula(ListOfConstrainedFormula insertions,
					boolean antec, boolean first) {

	Semisequent seq = antec ? antecedent : succedent;
	
	SemisequentChangeInfo semiCI = first ? seq.insertFirst(insertions) :
	    seq.insertLast(insertions);	

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (antec, semiCI, composeSequent(antec, semiCI.semisequent()), this);
	return sci;
    }

    /** 
     * adds a formula to the antecedent (or succedent) of the
     * sequent. Depending on the value of first the formulas are
     * inserted at the beginning or end of the ante-/succedent.
     *  (NOTICE:Sequent determines 
     * index using identy (==) not equality.)
     * @param cf the ConstrainedFormula to be added
     * @param antec boolean selecting the correct semisequent where to
     * insert the formulas. If set to true, the antecedent is taken
     * otherwise the succedent.
     * @param first boolean if true the formula is added at the
     * beginning of the ante-/succedent, otherwise to the end
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo addFormula(ConstrainedFormula cf, 
					boolean antec, boolean first) {

	Semisequent seq = antec ? antecedent : succedent;
	
	SemisequentChangeInfo semiCI = first ? 
	    seq.insertFirst(cf) : seq.insertLast(cf);	

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (antec, semiCI, composeSequent(antec, semiCI.semisequent()), this);

	return sci;
    }
    
    /** 
     * adds a formula to the sequent at the given
     * position. (NOTICE:Sequent determines 
     * index using identy (==) not equality.)
     * @param cf a ConstrainedFormula to be added
     * @param p a PosInOccurrence describes position in the sequent 
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo addFormula(ConstrainedFormula cf, PosInOccurrence p) {
	Semisequent seq = getSemisequent(p);

	SemisequentChangeInfo semiCI = seq.insert(seq.indexOf(p.constrainedFormula()),cf);
	
	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (p.isInAntec(), semiCI, composeSequent(p, semiCI.semisequent()), this);

	return sci;
    }


    /** adds the formulas of list insertions to the sequent starting at
     * position p. (NOTICE:Sequent determines 
     * index using identy (==) not equality.)
     * @param insertions a ListOfConstrainedFormula with the formulas to be added
     * @param p the PosInOccurrence describing the position where to
     * insert the formulas 
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo addFormula(ListOfConstrainedFormula insertions, 
					PosInOccurrence p) {
	Semisequent seq = getSemisequent(p);

	SemisequentChangeInfo semiCI = 
	    seq.insert(seq.indexOf(p.constrainedFormula()), insertions);

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (p.isInAntec(), semiCI, composeSequent(p, semiCI.semisequent()), this);

	return sci;
    }

    /** 
     * replaces the formula at the given position with another one
     * (NOTICE:Sequent determines 
     * index using identity (==) not equality.)  
     * @param newCF the ConstrainedFormula replacing the old one
     * @param p a PosInOccurrence describes position in the sequent 
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo changeFormula(ConstrainedFormula newCF,
					   PosInOccurrence p) { 
	Semisequent seq = getSemisequent(p);
	SemisequentChangeInfo semiCI = seq.replace(p, newCF);

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (p.isInAntec(), semiCI, composeSequent(p, semiCI.semisequent()), this);

	return sci;

    }

    /** 
     * replaces the formula at position p with the head of the given
     * list and adds the remaining list elements to the sequent
     * (NOTICE:Sequent determines 
     * index using identity (==) not equality.)  
     * @param replacements the ListOfConstrainedFormula whose head
     * replaces the formula at position p and adds the rest of the list
     * behind the changed formula
     * @param p a PosInOccurrence describing the position of the formula
     * to be replaced
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo changeFormula(ListOfConstrainedFormula replacements,
					   PosInOccurrence p) {

	Semisequent seq = getSemisequent(p);       
	SemisequentChangeInfo semiCI = seq.replace(p, replacements);

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (p.isInAntec(), semiCI, composeSequent(p, semiCI.semisequent()), this);

	return sci;
    }

    /** removes the formula at position p (NOTICE:Sequent determines
     * index using identity (==) not equality.) 
     * @param newCF the ConstrainedFormula replacing the old one
     * @param p a PosInOccurrence that describes position in the sequent 
     * @return a SequentChangeInfo which contains the new sequent and
     * information which formulas have been added or removed 
     */
    public SequentChangeInfo removeFormula(PosInOccurrence p) { 
	Semisequent seq=getSemisequent(p);       

	SemisequentChangeInfo semiCI = 
	    seq.remove(seq.indexOf(p.constrainedFormula()));

	SequentChangeInfo sci = SequentChangeInfo.createSequentChangeInfo
	    (p.isInAntec(), semiCI, composeSequent(p, semiCI.semisequent()), this);

	return sci; 
    }
    
    
    public int formulaNumberInSequent(boolean inAntec, 
                                      ConstrainedFormula cfma) {
       IteratorOfConstrainedFormula formIter = iterator();
       int n=0;
       
       if (inAntec) {
          formIter = antecedent().iterator();
          while (formIter.hasNext()) {
             n++;
             if (formIter.next().equals(cfma)) return n;
          }
       }
       else {
          formIter = succedent().iterator();
          while (formIter.hasNext()) {
             n++;
             if (formIter.next().equals(cfma)) return n+antecedent().size();
          }
       }
       throw new RuntimeException("Ghost formula "+cfma+" in sequent "+
                                  this+" [antec="+inAntec+"]");
    }
    
    
     public ConstrainedFormula getFormulabyNr(int formulaNumber) {
	 ConstrainedFormula cfma = null;
	 IteratorOfConstrainedFormula formIter = iterator();
	 int n=0;
	 while (n != formulaNumber && formIter.hasNext()) {
	     n++;
	     cfma = formIter.next();
	 }
	 if (cfma == null) throw new RuntimeException("No formula nr. "+
				      formulaNumber+" in seq. "+this);
	 return cfma;
     }
    
    
    public boolean numberInAntec(int formulaNumber) {
       return formulaNumber <= antecedent.size();
    }
    
    
    public int size () {
        return antecedent ().size () + succedent ().size ();
    }
    

    // does the real work for varIsBound in a given semiseq    
    private boolean varIsBoundHelp(QuantifiableVariable v, Semisequent semi) {
	IteratorOfConstrainedFormula it=semi.iterator();	
	while (it.hasNext()) {
	    BoundVarsVisitor bvv=new BoundVarsVisitor();
	    it.next().formula().execPostOrder(bvv);
	    if (bvv.getBoundVariables().contains(v)) {
		return true;
	    }
	}
	return false;
    }

    
    /**
     * returns true iff the given variable is bound in a formula of a
     * ConstrainedFormula in this sequent.
     * @param v the bound variable to search for
     */
    public boolean varIsBound(QuantifiableVariable v) {
	return varIsBoundHelp(v, antecedent()) || 
	    varIsBoundHelp(v, succedent);
    }

    /**
     * returns pairs of variables (x,y) where y is bound in the formulae
     * (i.e. the formulae of all ConstrainedFormulae, Constraints are not
     * considered) of this Sequent and x
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y.
     * @return a set of variable pairs as specified above
     */
    private SetOfInScopePair getVarInScope(Semisequent semi) {
	IteratorOfConstrainedFormula it = semi.iterator();	
	SetOfInScopePair result = SetAsListOfInScopePair.EMPTY_SET;
	while (it.hasNext()) {
	    VarInScopeCollector bc = new VarInScopeCollector();
	    it.next().formula().execPostOrder(bc);
	    result=result.union(bc.getVarInScope());
	}	
	return result;
    }
    
    public SetOfInScopePair getVarInScope() {
	return getVarInScope(antecedent()).union(getVarInScope(succedent()));
    }

    /**
     * executes the syntactical Replacement defined by the given map on all
     * formulas of the Sequent and returns a new Sequent where the
     * replacement is done. The Constraints of the ConstrainedFormulas are ignored
     * @param services the Services encapsulating all java information
     * @param svInst the SVInstantiations containing the replacements
     * @return a new Sequent with the replaced terms 
     */
    protected Sequent syntacticalReplace(Services services, SVInstantiations svInst) {
	return Sequent.createSequent(antecedent().syntacticalReplace(services, svInst),
				     succedent().syntacticalReplace(services, svInst));
    }


    /** looks for variable var in the sequent 
     * @param var the SchemaVariable to be looked for
     * @return true iff variable has been found in some
     * ConstrainedFormula of the sequent
     */
    public boolean lookup(SchemaVariable var) {
	return antecedent().lookup(var) ||
	    succedent().lookup(var); 
    }

    
    /** returns iterator about all ConstrainedFormulae of the sequent
     * @return iterator about all ConstrainedFormulae of the sequent 
     */
    public IteratorOfConstrainedFormula iterator() {
	return new SequentIterator( antecedent(), succedent() );
    }

    public boolean equals(Object o) {
 	if ( ! ( o instanceof Sequent ) ) 
	    return false;
	Sequent o1 = (Sequent) o;
	return antecedent.equals(o1.antecedent())
	    && succedent.equals(o1.succedent());
    }

    public int hashCode () {
        return antecedent.hashCode () * 17 + succedent.hashCode ();
    }
    
    /** String representation of the sequent
     * @return String representation of the sequent */
    public String toString() {
	return antecedent().toString()+"==>"+succedent().toString();
    }

    public void prettyprint(de.uka.ilkd.key.pp.LogicPrinter printer) {
	printer.printSequent(this);
    }

    public void prettyprint(de.uka.ilkd.key.pp.LogicPrinter printer, SequentPrintFilter filter) {
	printer.printSequent(this, filter);
    }

    public StringBuffer prettyprint() {
	de.uka.ilkd.key.pp.LogicPrinter lp = (new de.uka.ilkd.key.pp.LogicPrinter 
					       (new de.uka.ilkd.key.pp.ProgramPrinter(null), 
						de.uka.ilkd.key.pp.NotationInfo.createInstance()));
	lp.printSequent(this);
	return lp.result();
    }

    static class SequentIterator implements IteratorOfConstrainedFormula {

	private IteratorOfConstrainedFormula anteIt;
	private IteratorOfConstrainedFormula succIt;


	SequentIterator(Semisequent ante, 
			Semisequent succ) {
	    this.anteIt = ante.iterator();
	    this.succIt = succ.iterator();
	}

	public ConstrainedFormula next() {
	    if (anteIt.hasNext()) {
		return anteIt.next();
	    } 	    
	    return succIt.next();
	}

	public boolean hasNext() {
	    return anteIt.hasNext() || succIt.hasNext();
	}

    } 


}







