// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.java.NameAbstractionTable;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.Rulevariable;

/** 
 * In contrast to the distinction of formulas and terms as made by most of the 
 * inductive definition of the syntax of a logic, an instance of this class can
 * stand for a term or a formula. This is done for implementation reasons, as
 * their structure is quite similar and there are good reasons concerning the
 * software's design/architecture (for example using same visitors, reduction of
 * case distinction, unified interfaces etc.). However, they are strictly
 * separated by their sorts. A formula (and just a formula) must have 
 * the sort {@link Sort#FORMULA}. Terms of a different sort are terms in the
 * customary logic sense. A term of sort formula is allowed exact there, where a
 * formuala in logic is allowed to appear, same for terms of different sorts. 
 *   Some words about other design decisions: 
 * <ol> 
 *  <li> terms are immutable, this means after a term object is created, it
 *  cannot be changed. The advantage is that we can use ecessive term sharing and
 *  saving a lot of memory space. 
 *  </li>
 *  <li> Term has to be created using the {@link TermFactory} and
 *    <emph>not</emph> by using the constructors itself. 
 *  </li>
 *  <li> Term is subclassed, but all subclasses have to be package private, so
 *    that all other classes except {@link TermFactory} know only the class
 *    Term and its interface. Even most classes of the logic package.
 *  </li>
 *  <li> as it is immutable, most (all) attributes should be declared final
 * </li>
 * </ol>
 */
public abstract class Term implements SVSubstitute {

    /** empty list of variables */
    public static final ArrayOfQuantifiableVariable EMPTY_VAR_LIST = 
	new ArrayOfQuantifiableVariable(new QuantifiableVariable[0]);

    /**
     * @link aggregation
     * @clientCardinality *
     * @supplierCardinality 1
     */
    private final Operator op;

    /** caches the sort of the term */
    private final Sort sort;
    
    /** caches the hashcode of the term */
    private int hashcode;
    
    /** true iff this term is rigid */
    private boolean rigid;

    protected SetOfQuantifiableVariable freeVars =
	SetAsListOfQuantifiableVariable.EMPTY_SET; 

    protected SetOfMetavariable metaVars =
        SetAsListOfMetavariable.EMPTY_SET; 


    /** 
     * creates a Term with top operator op
     * @param op Operator at the top of the termstructure that starts
     * here.
     */
    protected Term(Operator op, Sort sort) {
	this.op       = op;
	this.sort     = sort;
    }

    public SetOfQuantifiableVariable freeVars(){
	return freeVars;
    }

    public SetOfMetavariable metaVars () {
        return metaVars;
    }

    /**
     * fills the cache variables
     */
    protected void fillCaches() {
	determineFreeVars();
	determineMetaVars();
	determineRigidness();
	hashcode = calculateHash();	
    }

    /**
     * this method has to be called by subclasses after they determined the
     * arity of the term. The method then can determine the free vars of the
     * term and put them in a cache.
     */
    private void determineFreeVars() {
	for (int i=0; i<arity(); i++) {
	    if (sub(i)==null) {System.err.println("FREE "+op+" "+i);}
	    SetOfQuantifiableVariable subFreeVars = sub(i).freeVars;
	    for (int j=0; j<varsBoundHere(i).size(); j++) {
		subFreeVars = subFreeVars.
		    remove(varsBoundHere(i)
			   .getQuantifiableVariable(j));
	    }
	    freeVars = freeVars.union(subFreeVars);
	}
    }

    private void determineMetaVars () {
        for ( int i = 0; i < arity (); i++ ) {
            if ( sub ( i ) == null ) {
                System.err.println ( "FREE " + op + " " + i );
            }
            metaVars = metaVars.union ( sub ( i ).metaVars );
        }
    }


    /**
     * this method has to be called by subclasses after they have
     * assigned the subterms of the term. The method then can
     * determine whether the operator and the subterms are rigid and
     * put this information in a cache.
     *
     * the implementation of this method uses the fact that
     * quantifiable variables are rigid; otherwise, the results are
     * too pessimistic (but still valid)
     */
    private void determineRigidness () {
	this.rigid = op ().isRigid ( this ) ;
    }

    /**
     * @return true iff all subterms of this term are rigid according
     * to the value the method "isRigid" returns for them (this does
     * not imply this term to be rigid itself!)
     */
    public boolean hasRigidSubterms () {
	for (int i=0; i<arity(); i++) {
	    if ( sub(i) == null )
		System.err.println("FREE "+op+" "+i);
	    if ( !sub(i).isRigid () )
		return false;
	}
	return true;
    }

    /**
     * @return false if the value of this term may be changed by
     * modalities (otherwise, however, the result may also be false)
     */
    public final boolean isRigid () {
	return rigid;
    }

    /** @return arity of the term */
    public abstract int arity();

    /** the nr-th subterm */
    public abstract Term sub(int nr);

    /** the top operator in "A and B" it is "and", in f(x,y) it is "f"
     * @return operator at the top
     */
    public Operator op() {
	return op;
    }
    
    /** @return the sort of the term */
    public Sort sort() {
	return sort;
    }

    /** returns the depth of the term */
    public abstract int depth();

   /**
    * @return the hashcode of the Term.
    */    
    public int hashCode(){
	if (hashcode == 0) {
	    hashcode = calculateHash();
	}
	return hashcode;
    }

    protected int calculateHash(){
	int hashValue = 5;
	hashValue = hashValue*17 + op().hashCode();  
	hashValue = hashValue*17 + sort().hashCode();
	for(int i=0 ; i<arity() ; i++){
	   hashValue = hashValue*17 + varsBoundHere(i).size();
	   hashValue = hashValue*17 + sub(i).hashCode();
	}
	hashValue = hashValue*17 + javaBlock().hashCode();
	return hashValue;
    }

    /**
     * @return true iff the given Term has the same values in
     * operator, sort, arity, varsBoundHere and javaBlock as this object.
     */
    public boolean equals(Object o) {
	if (o == null) return false;
	if (o == this) return true;
	if (!(o instanceof Term)) return false;	

	Term t = (Term) o;	

	if (!(t.hashCode() == hashCode() && 
	      //sort().equals(t.sort()) && 
	      sort() == t.sort() &&
	      arity() == t.arity() && 
	      op().equals(t.op()) && 
	      javaBlock().equals(t.javaBlock()))) {
	    return false;
	}

	for (int i = 0; i < arity(); i++) {
	    if (varsBoundHere(i).size()!=t.varsBoundHere(i).size()){
		return false;
	    }
	    for (int j = 0; j<varsBoundHere(i).size(); j++) {
		if (!varsBoundHere(i).getQuantifiableVariable(j).equals
		    (t.varsBoundHere(i).getQuantifiableVariable(j))) {
		    return false;
		}
	    }

	    if (!sub(i).equals(t.sub(i))) {
		return false;
	    }
	}

	return true;
    }

    /**
     * checks if the o occurs in this term.
     * @return true if o occurs in the term and is instance of Term,
     * ProgramVariable, QuantifiableVariable, SchemaVariable or Operator.
     */
    public boolean occurCheck(Object o){
        if(op().equals(o) || this.equals(o)) return true;
        if(o instanceof ProgramVariable && javaBlock().program()!=null){
            ProgramVariableCollector collector=
                new ProgramVariableCollector(javaBlock().program());
            collector.start();
            if(collector.result().contains(o)) return true;
        }
        if(o instanceof QuantifiableVariable){
            if(freeVars.contains((QuantifiableVariable) o)) return true;
            for(int i=0;i<arity();i++){
                for (int j=0;j<varsBoundHere(i).size();j++){
                    if(varsBoundHere(i).getQuantifiableVariable(j).equals(o))
                        return true;
                }
            }
        }            
        for(int i=0;i<arity();i++){
            if(sub(i).occurCheck(o)) return true;
        }
        return false;
    } 
    
    /**
     * compares to terms modulo bound renaming
     * @return true iff the given Term has the same values in
     * operator, sort, arity, varsBoundHere and javaBlock as this object
     * modulo bound renaming
     */  
    public boolean equalsModRenaming(Object o) {
        if (o == null || !(o instanceof Term)) {
	    return false;
	}
	if (o == this){
	    return true;
	}	
	Term cmp = (Term) o;
	
	final Constraint result = Constraint.BOTTOM.unify(this, cmp);
	return result == Constraint.BOTTOM;	
    }

    /** the visitor is handed through till the bottom of the tree and
     * then it walks upwards while at each upstep the method visit of
     * the visitor is called
     * @param visitor the Visitor
     */
    public void execPostOrder(Visitor visitor) {
	visitor.subtreeEntered(this);
	for (int i=0;i<arity();i++) {
	    sub(i).execPostOrder(visitor);
	}
	visitor.visit(this);
	visitor.subtreeLeft(this);
    }

    /** the visitor walks downwards the tree
     * while at each downstep the method visit of
     * the visitor is called
     * @param visitor the Visitor
     */
    public void execPreOrder(Visitor visitor) {
	visitor.subtreeEntered(this);
	visitor.visit(this);
	for (int i=0;i<arity();i++) {
	    sub(i).execPreOrder(visitor);
	}
	visitor.subtreeLeft(this);
	
    }



    /** @return the bound variables for the n-th subterm */
    public abstract ArrayOfQuantifiableVariable varsBoundHere(int n);

    /** @return JavaBlock if term has diamond at the top level */
    public JavaBlock javaBlock() {
	return JavaBlock.EMPTY_JAVABLOCK;
    }
    
    /** The primary diamond in this formula. Note that the default
     *  implementation is the same as <code>javaBlock()</code> but
     *  the semantics is different. See <code>SimultaneousUpdateTerm</code>.
     */
    public JavaBlock executableJavaBlock() {
        return javaBlock();
    }
    
     /** @return rulevariable if term is proofobligation term  */
    public Rulevariable ruleVar() {
	return null;
    }

    /**
     * checks whether the Term is valid on the top level. If this is
     * the case this method returns the Term unmodified. Otherwise a
     * TermCreationException is thrown.  
     * @return this Term if the top level of the Term is valid.
     */
    public Term checked() {
	if (op().validTopLevel(this)) {
	    return this;	    
	} else {
	    String subs="";
	    for (int i=0; i<arity(); i++) {
		subs=subs+" "+sub(i)+":"+sub(i).sort()+"(hashcode: "+
		    sub(i).sort().hashCode()+")";
	    }

	    final String errorMsg = 
		"Building a term with top level operator " + op() + "failed." +
		"Usually this indicates that the arity or on of the subterm's sort " +
		"is not as required (e.g. like the \'2\' in \"true & 2\")" +
		"The subterms were:\n "+subs;
	    throw new TermCreationException(errorMsg);
	}
    }

    /**
     * returns subterm at the position specified by the given PosInTerm pos.
     * @param pos the position of the wanted subterm
     * @return the subterm that is specified by pos
     */
    public Term subAt(PosInTerm pos) {
	Term sub = this;
	for (IntIterator it = pos.iterator(); it.hasNext(); ) {	
	    sub = sub.sub(it.next());
	}
	return sub;
    }
	
    public String toString() {
	StringBuffer sb = new StringBuffer(op().name().toString());
	if (arity()==0) return sb.toString();
	sb.append("(");
	for (int i=0;i<arity();i++) {
	    for (int j=0; j<varsBoundHere(i).size(); j++) {
		if (j==0) {
		    sb.append("{");
		}
		sb.append(varsBoundHere(i).getQuantifiableVariable(j));
		if (j!=varsBoundHere(i).size()-1) {
		    sb.append(",");
		} else {
		    sb.append("}");
		}
	    }
	    sb.append(sub(i));
	    if (i<arity()-1) {
		sb.append(",");
	    }
	}
	sb.append(")");
	return sb.toString();
    }


// =============== For debugging, used in Main (COMPUTE_SPEC_OP-Z) =================

    public void tree() {
       System.out.println("==============================");
       tree(this, 0);
    }
    
    public void tree(Term t, int off) {
       int i;
       for (i=0; i<off; i++) System.out.print(" ");
       String s = t.op().toString();
       s = s.substring(s.lastIndexOf(".")+1, s.length());
       System.out.println(t+" ["+s+"]");
       for (i=0; i<t.arity(); i++) tree(t.sub(i), off+3);
    }

}
