// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.logic;

import java.util.ArrayList;
import java.util.List;

import de.uka.ilkd.key.logic.op.AccessOp;
import de.uka.ilkd.key.logic.op.AnonymousUpdate;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.ArrayOp;
import de.uka.ilkd.key.logic.op.AttributeOp;
import de.uka.ilkd.key.logic.op.ConjCond;
import de.uka.ilkd.key.logic.op.Equality;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.logic.op.IfElse;
import de.uka.ilkd.key.logic.op.Junctor;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.op.MetaOperator;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.Operator2;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.logic.op.Quantifier;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.ShadowArrayOp;
import de.uka.ilkd.key.logic.op.ShadowAttributeOp;
import de.uka.ilkd.key.logic.op.ShadowedOperator;
import de.uka.ilkd.key.logic.op.SubstOp;
import de.uka.ilkd.key.logic.op.TermSymbol;
import de.uka.ilkd.key.logic.op.UpdateOperator;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.ListOfUpdatePair;
import de.uka.ilkd.key.rule.Rulevariable;
import de.uka.ilkd.key.rule.UpdatePair;
import de.uka.ilkd.key.util.Debug;

/*
 * creates instances of subclasses of {@link
 * de.uka.ilkd.key.logic.Term}.  As class {@link
 * de.uka.ilkd.key.logic.Term} is abstract, but access to
 * subclasses is only allowed via {@link
 * de.uka.ilkd.key.logic.Term} this class is used to create
 * special kind of Terms.
 */
public class TermFactory {

    /** An instance of TermFactory */
    public static final TermFactory DEFAULT=new TermFactory();

    private static final Term[] NO_SUBTERMS = new Term[0];

    /** creates t[index] with top operand op
     */
    public OpTerm createArrayTerm(ArrayOp op, Term t, Term index){
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory"); 
        Term[] t1 = new Term[2];
        t1[0]=t;
	t1[1] = index;
	return (OpTerm) (new OpTerm(op,t1)).checked();
    }

    /** creates t[index] with top operand op
     */
    public OpTerm createShadowArrayTerm(ShadowArrayOp op, Term t, Term index, Term shadownum){
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory"); 
	}
	return (OpTerm) (new OpTerm(op, new Term[]{t, index, shadownum})).checked();
    }
	
    /** creates t[index[0]]..[index[n]] with top operand op
     */
    public OpTerm createArrayTerm(ArrayOp op, Term t, Term[] index){
	if (op == null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory"); 
	}
        Term[] t1 = new Term[2];
        t1[0]=t;
        for(int i=0;i<index.length;i++){
	    t1[1] = index[i];
	    t1[0] = (OpTerm) (new OpTerm(op, t1)).checked();
	}
	return (OpTerm) t1[0];
    }


    public OpTerm createShadowArrayTerm(ShadowArrayOp op, Term t, Term[] index, Term shadownum){
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory"); 
	}
        Term[] t1 = new Term[3];
        t1[0] = t;
        for(int i=0;i<index.length;i++){
	    t1[1] = index[i];
	    t1[2] = shadownum;
	    t1[0] = (OpTerm) (new OpTerm(op, t1)).checked();
	}
	return (OpTerm) t1[0];
    }


    /**
     * @param op the shadowed or normal version of the array access operator
     * @param subs array of subterms
     * @return the term representing an array access
     */
    public OpTerm createArrayTerm(ArrayOp arrayOp, Term[] subs) {
       if (arrayOp==null) {
           throw new IllegalArgumentException
               ("null-Operator at TermFactory");
       }       
       if (arrayOp instanceof ShadowedOperator) arrayOp = 
           ShadowArrayOp.getShadowArrayOp(subs[0].sort());
       else 
           arrayOp = ArrayOp.getArrayOp(subs[0].sort());
       return (OpTerm) (new OpTerm(arrayOp, subs)).checked();              
    }
    
    public OpTerm createAttributeTerm(AttributeOp op, Term term) {
	Debug.assertFalse(op instanceof ShadowedOperator, 
			  "Tried to create a shadowed attribute.");
	if (op.attribute() instanceof ProgramVariable && 
	    ((ProgramVariable)op.attribute()).isStatic()) {
	    return (OpTerm) new OpTerm((ProgramVariable)op.attribute(), 
				       NO_SUBTERMS).checked();
	} 
	return (OpTerm) new OpTerm(op, new Term[]{term}).checked();
    }
    
    /**
     * creates a term representing an attribute access
     * @param attrOp the AttributeOp representing the attribute to be accessed
     * @param subs an array of Term containing the subterms (usually of 
     * length 1 but may have length 2 for shadowed accesses) 
     * @return the term <code>subs[0].attr</code> 
     * (or <code>subs[0]^(subs[1]).attr)</code>)
     */
    public Term createAttributeTerm(AttributeOp attrOp, Term[] subs) {
        if (attrOp instanceof ShadowedOperator) {
            return createShadowAttributeTerm
                ((ShadowAttributeOp)attrOp, subs[0], subs[1]);
        }
        return createAttributeTerm(attrOp, subs[0]);
    }

    /** creates an attribute term that references to a field of a class
     * @param var the variable the attribute term references to
     * @param term the Term describing the class/object of which the
     * attribute value has to be determined
     * @return the attribute term "term.var"
     */
    public OpTerm createAttributeTerm(ProgramVariable var, 
				      Term term) {
	if (var.isStatic()) {
	    return (OpTerm) new OpTerm(var, NO_SUBTERMS).checked();
	}
	return (OpTerm) new OpTerm(AttributeOp.getAttributeOp(var),
				   new Term[]{term}).checked();
    }
    
    /** creates an attribute term that references to a field of a class 
     * @param var the variable the attribute term references to
     * @param term the Term describing the class/object of which the
     * attribute value has to be determined      
     * @return the attribute term "term.var"
     */
    public OpTerm createAttributeTerm(SchemaVariable var, Term term) {
	return (OpTerm) new OpTerm
	    (AttributeOp.getAttributeOp((IProgramVariable)var), 
	     new Term[]{term}).checked();
    }


    public ProgramTerm createBoxTerm(JavaBlock javaBlock, Term subTerm) {
	return (ProgramTerm) createProgramTerm(Op.BOX, javaBlock, subTerm);
    }

    /**
     * creates a conditonal term where the condition is a conjuntion 
     * of equalities        
     * @param subs the subterms of the conditonal term to be created
     * @return simplified version of
     *     subs[0] * subs[1] : subs[2] ?= subs[3] * subs[4] : subs[5] 
     */
    public Term createConjConditionalTermAndSimplify(Term[] subs) {
        
        if (subs[2].equals(subs[5])) {// same branches
            return subs[2];
        } else if (subs[0].equals(subs[3]) &&
                subs[1].equals(subs[4])) { // syntactical equal conditions
                return subs[5];
        } else {   
            return createConjCondTerm(subs);
        }                
    }
    
    /**
     * creates a conditonal term where the condition is a conjuntion 
     * of equalities        
     * @return simplified version of
     *     fstCmp[0] * fstCmp[1] : _else ?= sndCmp[0] * sndCmp[1] : _then 
     */
    public Term createConjConditionalTermAndSimplify(Term[] fstCmp, 
            Term[] sndCmp, Term _then, Term _else) {
        
        if (_then.equals(_else)) { // same branches
            return _then;
        } else if (fstCmp[0].equals(sndCmp[0]) &&
                fstCmp[1].equals(sndCmp[1])) { // syntactical equal conditions
                return _then;
        } else {   
            final Term[] allSubs = new Term[6];
            allSubs[0] = fstCmp[0];
            allSubs[1] = fstCmp[1];
            allSubs[2] = _else;
            allSubs[3] = sndCmp[0];
            allSubs[4] = sndCmp[1];
            allSubs[5] = _then;
            return createConjCondTerm(allSubs);
        }                
    }


    /** 
     * creates a chained conjunctive conditional term used 
     * @param subs the subterms 
     */
    public OpTerm createConjCondTerm (Term[] subs) {
	return (OpTerm) new OpTerm(ConjCond.createConjCond(subs.length), 
				   subs).checked();
    }    

    public ProgramTerm createDiamondTerm(JavaBlock javaBlock, Term subTerm) {
	return (ProgramTerm) createProgramTerm(Op.DIA, javaBlock, subTerm);
    }

    /**
     * creates a EqualityTerm with a given equality operator. USE THIS WITH
     * CARE! THERE IS NO CHECK THAT THE EQUALITY OPERATOR MATCHES THE TERMS.
     */
    public OpTerm createEqualityTerm(Equality op, Term t1, Term t2) {
	return (OpTerm) (new OpTerm(op, new Term[]{t1, t2} )).checked();
    }
    

    /**
     * creates a EqualityTerm with a given equality operator. USE THIS WITH
     * CARE! THERE IS NO CHECK THAT THE EQUALITY OPERATOR MATCHES THE TERMS.
     */
    public OpTerm createEqualityTerm(Equality op, Term[] subTerms) {
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

 
    /** create an EqualityTerm with the correct equality symbol for
     * the sorts involved, according to {@link Sort#getEqualitySymbol}
     */
    public OpTerm createEqualityTerm(Term t1, Term t2) {
	Equality eq = t1.sort().getEqualitySymbol();
	if (t1.op() instanceof SchemaVariable) {
	    eq = t2.sort().getEqualitySymbol();
	} 
	if (eq == null) eq = Op.EQUALS;
	return (OpTerm) new OpTerm(eq, new Term[]{t1, t2} ).checked();
    }


    /** create an EqualityTerm with the correct equality symbol for
     * the sorts involved, according to {@link Sort#getEqualitySymbol}
     */
    public OpTerm createEqualityTerm(Term[] terms) {
	return createEqualityTerm(terms[0], terms[1]);
    }



    public OpTerm createFunctionTerm(TermSymbol op) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");
	return (OpTerm) (new OpTerm(op, NO_SUBTERMS)).checked();
    }

    public OpTerm createFunctionTerm(TermSymbol op, Term s1) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, new Term[]{s1})).checked();
    }

    public OpTerm createFunctionTerm(TermSymbol op, Term s1,Term s2) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, new Term[]{s1,s2})).checked();
    }



     /** creates a FunctionTerm with top operator op, some subterms
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms
     */
    public OpTerm createFunctionTerm(TermSymbol op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }


    //For OCL simplification

    /** creates a term representing an OCL expression with a
      * collection operation as top operator that 
      * takes an OclExpression as argument (not "iterate")
      * @param op the OCL collection operation
      * @param varBoundHere the iterator variable
      * @param subs subs[0] is the collection and subs[1] is the 
      *        expression in which the iterator variable is bound
      */

    public Term createFunctionWithBoundVarsTerm(TermSymbol op,
						PairOfTermArrayAndBoundVarsArray subs) {
	return createFunctionWithBoundVarsTerm(op, subs.getTerms(), subs.getBoundVars());
    }

    public Term createFunctionWithBoundVarsTerm(TermSymbol op,
						Term[] subTerms,
						ArrayOfQuantifiableVariable[] boundVars) {
	if (boundVars != null) {
	   return (BoundVarsTerm)(new BoundVarsTerm(op, subTerms, boundVars)).checked(); 
	} else {
	    return createFunctionTerm(op, subTerms);
	}
     }


    /** 
     * creates an `if-else' term  
     * <code> (cond1 ?= cond2) _then : _else </code>
     * @param op the IfElse operator
     * @param subs the subterms of the 'if-else' term
     */
    public OpTerm createIfElseTerm (IfElse op, Term[] subs) {

	return (OpTerm) new OpTerm(op,new Term[]{subs[0], subs[1], 
						 subs[2], subs[3]}).checked();
    }

    /** creates an 'if-else' term  
     * <code> (cond1 ?= cond2) _then : _else </code>
     * @param cond1 the first part of the (reference) equality condition
     * @param cond2 the second part of the (reference) equality condition
     * @param _then the 'then' part of the term
     * @param _else the 'else' part of the term
     */
    public OpTerm createIfElseTerm (Term cond1, Term cond2, 
				    Term _then, Term _else) {

	return (OpTerm) new OpTerm
	    (Op.IF_ELSE, new Term[]{cond1, cond2, _then, _else}).checked();
    }


    /** creates an `if-else' term given in the short form 
     * <code> (cond1 ?= cond2).attr /-> _then </code> that is equivalent to
     * <code> (cond1 ?= cond2) _then : cond1.attr </code>
     * @param attr the TermSymbol (ProgramVariable/SchemaVariable) an attribute
     * of cond1 and cond2
     * @param cond1 the first part of the (reference) equality condition
     * @param cond2 the second part of the (reference) equality condition
     * @param _else the 'then' part of the term
     */
    public OpTerm createIfElseTerm (TermSymbol attr, Term cond1, 
				    Term cond2, Term _then) {

	Term _else = null;
	if (attr instanceof ProgramVariable) {
	    _else = createAttributeTerm((ProgramVariable)attr, 
					cond1);
	} else {
	    _else = createAttributeTerm((SchemaVariable)attr, cond1);
	}

	return (OpTerm) new OpTerm
	    (Op.IF_ELSE, new Term[]{cond1, cond2, _then, _else}).checked();
    }

    /**
     * creates an if-else term and tries to simplify as far as possible
     * 
     * @param op the IfElse operator
     * @param subs array of Term of length <em>4</em> where 
     *        (subs[0] ?= subs[1]) subs[2] : subs[3] 
     * @return the simplifed version of the if else term
     */
    public Term createIfElseTermAndSimplify(Term[] subs) {
        if (subs[0].equals(subs[1])) {
            return subs[2];
        } else {
            if (subs[2].equals(subs[3])) {
                return subs[2];
            }
            return createIfElseTerm(Op.IF_ELSE, subs);
        }
    }
    
    

    /** some methods to handle the equality for formulas (equiv - operator) ... */
    public OpTerm createJunctorTerm(Equality op, Term t1, Term t2) {
	return createEqualityTerm(op, new Term[]{t1, t2});
    }
    
    public OpTerm createJunctorTerm(Equality op, Term[] subTerms) {
	return createEqualityTerm(op, subTerms);
    }

    public OpTerm createJunctorTerm(Junctor op) {
	return createJunctorTerm(op, NO_SUBTERMS);
    }

    public OpTerm createJunctorTerm(Junctor op, Term t1) {
	return createJunctorTerm(op, new Term[]{t1});
    } 

    public OpTerm createJunctorTerm(Junctor op, Term t1, Term t2) {
	return createJunctorTerm(op, new Term[]{t1, t2});
    }
 
    

    /** creates a JunctorTerm with top operator op, some subterms
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms */
    public OpTerm createJunctorTerm(Junctor op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    /** some methods for the creation of junctor terms with automatically performed simplification
     * like  ( b /\ true ) == (b) ...
     * Currently only the AND, OR, IMP Operators will be simplified (if possible)
     */

    public Term createJunctorTermAndSimplify(Junctor op, Term t1, Term t2) {
	if (op == Op.AND) {
	    // if one of the terms is false the expression is false as a whole
	    if (t1.op() == Op.FALSE || t2.op() == Op.FALSE)
	        return createJunctorTerm(Op.FALSE);
	    // if one of the terms is true skip the subterm.
	    if (t1.op() == Op.TRUE) {
		return  t2;
	    } else if(t2.op() == Op.TRUE) {
		return t1;
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    }
	} else if (op == Op.OR) {
	    // if one of the terms is true the expression is true as a whole
	    if (t1.op() == Op.TRUE || t2.op() == Op.TRUE)
		return createJunctorTerm(Op.TRUE);
	    // if one of the terms is false skip the subterm.
	    if (t1.op() == Op.FALSE) {
		return t2;
	    } else if(t2.op() == Op.FALSE) {
		return t1;
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    } 
	} else if (op == Op.IMP) {
	    if (t1.op() == Op.FALSE || t2.op() == Op.TRUE)
		// then the expression is true as a whole
		return createJunctorTerm(Op.TRUE);
	    // if t1 is true or t2 is false skip that subterm.
	    if (t1.op() == Op.TRUE) {
		return t2;
	    } else if(t2.op() == Op.FALSE) {
		return createJunctorTerm(Op.NOT, t1);
	    } else { // nothing to simplifiy ...
		return createJunctorTerm(op, t1, t2);
	    }
	
	} else { // all other Junctors ...
	    return createJunctorTerm(op, t1, t2);
	}
    }

     /** creates a OpTerm with a meta operator as top operator. These
     * terms are only used in the replacewith areas of taclets. And are
     * replaced by the SyntacticalReplaceVisitor
     * @param op Operator at the top of the termstructure that starts
     * here
     * @param an array containing subTerms (an array with length 0 if
     * there are no more subterms
     */
    public OpTerm createMetaTerm(MetaOperator op, Term[] subTerms) {
	if (op==null) throw new IllegalArgumentException("null-Operator at"+
							 "TermFactory");    
	return (OpTerm) (new OpTerm(op, subTerms)).checked();
    }

    public ProgramTerm createProgramTerm(Modality op, 
					 JavaBlock javaBlock, 
					 Term subTerm) {
	return (ProgramTerm) new ProgramTerm(op, javaBlock, subTerm).checked();
    }


    public ProgramTerm createProgramTerm(Operator op, 
					 JavaBlock javaBlock, 
					 Term[] subTerms) {
	return (ProgramTerm) new ProgramTerm(op, javaBlock, subTerms).checked();
    }



    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere an
     * array of QuantifiableVariable containing all variables bound by the
     * quantifier
     */
    public Term createQuantifierTerm(Quantifier quant,
				     ArrayOfQuantifiableVariable varsBoundHere, 
				     Term subTerm) {
	if (varsBoundHere.size()<=1) {
	    return (QuantifierTerm) new QuantifierTerm(quant, varsBoundHere, 
						       subTerm).checked();
	} else {
	    Term qt = subTerm;
	    for (int i=varsBoundHere.size()-1; i>=0; i--) {
		QuantifiableVariable qv 
		    = varsBoundHere.getQuantifiableVariable(i);
		qt = createQuantifierTerm(quant, qv, qt);
	    }
	    return qt;
	}	
    }



    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere a QuantifiableVariable representing the only bound
     * variable of this quantifier.
     */
    public Term createQuantifierTerm(Quantifier quant,
				     QuantifiableVariable var, 
				     Term subTerm) {
       return createQuantifierTerm
	   (quant, new QuantifiableVariable[]{var}, subTerm);
    }


    /**
     * creates a quantifier term 
     * @param op Operator representing the
     * Quantifier (all, exist) of this term 
     * @param varsBoundHere an
     * array of QuantifiableVariable containing all variables bound by the
     * quantifier
     */
    public Term createQuantifierTerm(Quantifier quant, 
				     QuantifiableVariable[] varsBoundHere, 
				     Term subTerm) {
	return createQuantifierTerm(quant, new ArrayOfQuantifiableVariable
	    (varsBoundHere), subTerm);
    }

    public OpTerm createShadowAttributeTerm(ProgramVariable var, 
					    Term term, Term shadownum) {
	return (OpTerm) new OpTerm(ShadowAttributeOp.getShadowAttributeOp(var), 
				   new Term[]{term, shadownum}).checked();
    }


    public OpTerm createShadowAttributeTerm(SchemaVariable var,
					    Term term, Term shadownum) {
	return (OpTerm) 
	    new OpTerm(ShadowAttributeOp.getShadowAttributeOp((IProgramVariable)var), 
		       new Term[]{term, shadownum}).checked();
    }
    
    public OpTerm createShadowAttributeTerm(ShadowAttributeOp op, 
					    Term term, Term shadownum) {
	return (OpTerm) new OpTerm(op, new Term[]{term, shadownum}).checked();
    }


     /** creates a substitution term
     * @param substVar the QuantifiableVariable to be substituted
     * @param substTerm the Term that replaces substVar
     * @param origTerm the Term that is substituted
     */
    public SubstitutionTerm createSubstitutionTerm
	(SubstOp op, QuantifiableVariable substVar, 
	 Term substTerm, Term origTerm) {
	return (SubstitutionTerm) new SubstitutionTerm
	    (op, substVar, new Term[]{substTerm, origTerm}).checked();
    }
    

     /** creates a substitution term
      * @param op the replacement variable
      * @param substVar the QuantifiableVariable to be substituted
      * @param subs an array of Term where subs[0] is the term that
      * replaces the variable substVar in subs[1] 
      */
    public SubstitutionTerm createSubstitutionTerm(SubstOp op,
            QuantifiableVariable substVar, Term[] subs) {
	return (SubstitutionTerm) new SubstitutionTerm
	    (op, substVar, subs).checked();
    }


    /**
     * helper method for term creation (reduces duplicate code)
     */
    private Term createTermWithNoBoundVariables(Operator op, 
						Term[] subTerms, 
						JavaBlock javaBlock) {
	if (op instanceof Equality) {
	    return createEqualityTerm(subTerms);
	} else if (op instanceof TermSymbol) {
	    return createFunctionTerm((TermSymbol)op, subTerms);
	} else if (op instanceof Junctor) {
	    return createJunctorTerm((Junctor)op,subTerms);
	} else if (op instanceof UpdateOperator) {
	    if (op instanceof AnonymousUpdate) {
		return createAnonymousUpdateTerm
		    ((AnonymousUpdate)op, subTerms);
	    } else { 
		return createNormalizedUpdateTerm((UpdateOperator)op, subTerms);
	    }
	} else if (op instanceof Modality) {
	    return createProgramTerm((Modality)op, javaBlock, subTerms[0]); 
	} else if (op instanceof AccessOp) {
	    if (op instanceof ShadowAttributeOp) {
		return createShadowAttributeTerm((ShadowAttributeOp)op, 
						 subTerms[0], subTerms[1]);
	    } else if (op instanceof AttributeOp) {
		return createAttributeTerm((AttributeOp)op, subTerms[0]);
	    } else if (op instanceof ArrayOp) {
		return createArrayTerm((ArrayOp)op, subTerms);
	    } else {
		Debug.fail("Unknown access operator" + op);
		return null;
	    }
	} else if (op instanceof IfElse) {
	    return op instanceof ConjCond ? 
		createConjCondTerm(subTerms) : createIfElseTerm((IfElse)op, subTerms);
	} else if (op instanceof MetaOperator) {
	    return createMetaTerm((MetaOperator)op, subTerms);
	} else {
	    de.uka.ilkd.key.util.Debug.fail("Should never be"+
					    " reached. Missing case for class", 
					    op.getClass());
	    return null;
	}       	
    }

    public Term createTerm(Operator op, Term[] subTerms, 
			   ArrayOfQuantifiableVariable vars, 
			   JavaBlock javaBlock, Rulevariable ruleVar) {
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory");
	} else if (op instanceof Operator2) {
	    /* If the operator is an instance of Operator2, the
	     * operator can create the term by itself. Operator2
	     * supports no Java blocks and no ruleVar. */
	    if (javaBlock != JavaBlock.EMPTY_JAVABLOCK) {
		throw new IllegalArgumentException("Non-empty JavaBlock in Operator2.");
	    } else if (ruleVar != null) {
		throw new IllegalArgumentException("Operator2 can not have Rulevariable.");
	    }
	    return ((Operator2) op).createTerm(vars, subTerms);
	} else if (op instanceof Quantifier) {
	    return createQuantifierTerm((Quantifier)op, vars, subTerms[0]);
	} else if (op instanceof SubstOp) {
	    return createSubstitutionTerm((SubstOp)op, 
					  vars.getQuantifiableVariable(0),
					  subTerms);
	} else {
	    return createTermWithNoBoundVariables(op, subTerms, javaBlock);
	}
    }


   public Term createTerm(Operator op, Term[] subTerms, 
			  ArrayOfQuantifiableVariable[] bv,
			  JavaBlock javaBlock, Rulevariable ruleVar) {
	if (op==null) {
	    throw new IllegalArgumentException("null-Operator at TermFactory");
	} else if (op instanceof Operator2) {
	    /* 
	     * If the operator is an instance of Operator2, the
	     * operator can create the term by itself. Operator2
	     * supports no Java blocks and no ruleVar. 
	     *
	     * Operator2 restrict the use of quantified variables. 
	     */
	    if (javaBlock != JavaBlock.EMPTY_JAVABLOCK) {
		throw new IllegalArgumentException("Non-empty JavaBlock in Operator2.");
	    } else if (ruleVar != null) {
		throw new IllegalArgumentException
		    ("Operator2 can not have Rulevariable."); 
	    }
	    List list = new ArrayList();
	    for (int i = 0; i < bv.length; ++i) {
		for (int j = 0; j < bv[i].size(); ++j) {
		    list.add(bv[i].getQuantifiableVariable(j));
		}
	    }
	    final QuantifiableVariable[] vars = (QuantifiableVariable[]) list.
		toArray(new QuantifiableVariable[list.size()]);
	    return ((Operator2) op).createTerm
		(new ArrayOfQuantifiableVariable(vars), subTerms);
	} else if (op instanceof Quantifier) {
	    return createQuantifierTerm((Quantifier)op, bv[0], subTerms[0]);
	} else if (op instanceof SubstOp) {	    
	    return createSubstitutionTerm((SubstOp)op, 
					  bv[1].getQuantifiableVariable(0),
					  subTerms);
	} else if (op instanceof TermSymbol) { 
	    // special treatment for OCL operators binding variables	    
	    return createFunctionWithBoundVarsTerm((TermSymbol)op, subTerms, bv);	 
	} else {
	    return createTermWithNoBoundVariables(op, subTerms, javaBlock);
	}       
   }
    //
    // CHANGE these two methods!  vars should be something like an 
    // array of arrays! 
    //


    /**
     * creates a term using the other methods of this class depending on the
     * given operator. If the kind of term is known before (without using
     * if-else cascades on the kind of operator) the other methods in this
     * factory should be preferred. Depending on the needed parameters for
     * the terms that should be created some of the parameters of this method
     * might be ignored.
     * @param op the top level operator for the new term.
     * @param subTerms the subterms for the new term. The first n elements
     * are taken if op is a Junctor or TermSymbol and n is the arity
     * of op. Only the first entry is taken if op is a Quantifier or
     * a Diamond. The first (representing the replacing term
     * for a variable) and the second (representing the term behind the
     * substitution operator) entries are taken if op is a SubstOp.
     * @param vars the variables that are bound to a subterm. Not considered
     * if op is a Junctor, TermSymbol or Diamond. If op is a
     * SubstOp only the first element is taken and the variable is bound to
     * the second subterm. In all other cases all variables are taken and
     * bound to the first subterm.
     * @param javaBlock representing a java code block. Only taken if op is a
     * Diamond.
     * @return the created Term
     */
    public Term createTerm(Operator op, Term[] subTerms, 
			   QuantifiableVariable[] vars, 
			   JavaBlock javaBlock, Rulevariable ruleVar) {
	return createTerm(op, subTerms, new ArrayOfQuantifiableVariable(vars),
			  javaBlock, ruleVar);
    }

    /**
     * creates an update term like
     *    <code>{pair0}..{pairN}target</code>     
     */
    public UpdateTerm createUpdateTerm(ListOfUpdatePair pairs, Term target) {
	if (pairs.size()>1) {
	    return createUpdateTerm(pairs.head(), 
				    (createUpdateTerm(pairs.tail(), 
						      target)));
	} else {
	    return createUpdateTerm(pairs.head(), target);
	}
    }

    /**
     * creates an update term where the given locations are updated
     * this method requires detailed knowledge of the update structure as
     * the subterms must have the correct order. 
     * @param locations the Location operators of the update
     * @param subs the array of Term with the subterms of the 
     * locations to be updated, the values and the target term 
     * @return the above described update term
     */
    public UpdateTerm createUpdateTerm(Location[] locations, Term[] subs) {                
        return (UpdateTerm) UpdateOperator.normalize(locations, subs);
    }
     
    
    /**
     * creates the update term <code>{loc:=value}target</code>
     * @param loc the Term representing the location to be updated
     * @param value the Term representing the value the location is updated to
     * @param target the Term on which the update is applied
     * @return the update term described above
     */
    public UpdateTerm createUpdateTerm(Term loc, Term value, Term target) {
        final Term[] subs = new Term[loc.arity()+2];
        for (int i = 0; i<loc.arity(); i++) {
            subs[i] = loc.sub(i);
        }
        subs[loc.arity()] = value;
        subs[loc.arity()+1] = target;
        
        return (UpdateTerm) new UpdateTerm
        (UpdateOperator.
            createUpdateOp(new Location[]{(Location)loc.op()}), subs).checked(); 
    }
    
   
    /** 
     * creates an update term 
     *   <code>{locs[0]:=values[0],...,locs[n]:=values[n]}target</code>
     * where <code>n</code> is the length of the location array. 
     * @param locs an array of Term describing the updates locations
     * @param values an array of Term describing the values to which 
     * the locations are updated
     * @param target the Term on which the update is applied to
     * @return the update term as described above
     */
    public UpdateTerm createUpdateTerm(Term[] locs, 
                                       Term[] values,
                                       Term target) {
        
        return (UpdateTerm)UpdateOperator.normalize(locs, values, target);
    }

    /** 
     * creates a normalized simultaneous update term
     * @param op the UpdateOperator 
     * @param subs the subterm of the simultaneous update term to be
     * created 
     * @return the normalized simultaneous update term
     */
    public UpdateTerm createNormalizedUpdateTerm(UpdateOperator op, Term[] subs) {
	return (UpdateTerm) 
	    UpdateOperator.normalize(op.locationsAsArray(), subs);
    }

    /** 
     * creates a simultaneous update-term
     * @param subs the sub-terms
     */
    public UpdateTerm createUpdateTerm(UpdateOperator op, Term[] subs) {
	return (UpdateTerm) new UpdateTerm (op, subs).checked();
    }

    public UpdateTerm createUpdateTerm(UpdatePair pair, 
				       Term target) {

	UpdateOperator op = pair.updateOperator();
	Term[] subs = new Term[pair.arity() + 1];
	
	for (int i = 0; i<subs.length-1; i++) {
	    subs[i] = pair.sub(i);	    
	}

	subs[subs.length-1] = target;

	return (UpdateTerm) new UpdateTerm(op, subs).checked();
    }
        
    /**
     * creates an update term which is not in normalform order
     * (usually usage of this method is discouraged)     
     * @return creates an update term which is not in normalform order
     */
    public UpdateTerm createUpdateTermUnordered
        (Location[] locations, Term[] subs) {
        return (UpdateTerm) new UpdateTerm
           (UpdateOperator.createUpdateOp(locations), subs).checked();              
    }

     /** 
      * creates a term consisting of the given variable.
      * @param v the variable
      */
    public OpTerm createVariableTerm(LogicVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }
    /**
     * creates a variable term representing the given programvariable
     * @param v the ProgramVariable to be represented
     * @return variable <code>v</code> as term 
     */
    public OpTerm createVariableTerm(ProgramVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }
    
    /**
     * creates a term with schemavariable <code>v</code> as top level operator     
     * @param v the SchemaVariable to be represented
     * @return the term <code>v</code>
     */
    public OpTerm createVariableTerm(SchemaVariable v) {
	return (OpTerm) (new OpTerm(v, NO_SUBTERMS)).checked();
    }

    /**
     * creates an anonymous update applied to the given target term 
     * @param name 
     * @param target
     * @return
     */
    public Term createAnonymousUpdateTerm(AnonymousUpdate op, Term[] subs) {       
        return createUpdateTerm(op, subs);
    }

    /**
     * creates an anonymous update applied to the given target term 
     * @param name 
     * @param target
     * @return
     */
    public Term createAnonymousUpdateTerm(Name name, Term target) {       
        return createUpdateTerm
        	(AnonymousUpdate.getAnonymousOperator(name),
        	     new Term[]{target});
    }

}
