// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.annotation.Annotation;
import de.uka.ilkd.key.logic.sort.Sort;
import java.io.IOException;

import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.ProgramConstruct;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.inst.IllegalInstantiationException;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 * This class represents the root of a schema variable hierarchy to be
 * express termstructures that match on logical terms. Schema variables
 * are used in Taclets where they act as placeholders for other
 * TermSymbols. The TermSymbols a SchemaVariable is allowed to match
 * is specified by their type and sort. If a match fits these
 * conditions can be checked using the method isCompatible(TermSymbol t)
 */
public abstract class SortedSchemaVariable extends TermSymbol
    implements SchemaVariable, ParsableVariable, QuantifiableVariable, ProgramConstruct {

    /** 
     * SchemaVariables are used as placeholder for other
     * TermSymbols. The allowed type of the symbols is stored here.
     */
    private Class matchType = TermSymbol.class;

    /**
     * this flag indicates if the schemavariable matches a list of syntactical
     * elements 
     */
    private boolean listSV = false;


    /** 
     * creates a new SchemaVariable. That is used as placeholder for
     * TermSymbols of a certain type.
     * @param name the Name of the SchemaVariable
     * @param matchType Class representing the type of symbols that
     * can be matched
     * @param sort the Sort of the SchemaVariable and the matched type     
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of syntactical elements
     */    
    protected SortedSchemaVariable(Name name,
				   Class matchType,
				   Sort sort,
				   boolean listSV) {
	super(name, sort);
	this.matchType = matchType;
	this.listSV    = listSV;
    }

    /** 
     * this method tests on object identity
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	return se == this;
    }
    
    public Class matchType() {
	return matchType;
    }

    /** returns true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     * @return true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     */
    public boolean isVariableSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a term but not a formula
     * @return true iff this SchemaVariable is used to match
     * a term but not a formula
     */
    public  boolean isTermSV() {
	return false;
    }


    /** returns true iff this SchemaVariable is used to match
     * a formula 
     * @return true iff this SchemaVariable is used to match
     * a formula
     */
    public boolean isFormulaSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a part of a program 
     * @return true iff this SchemaVariable is used to match
     * a part of a program
     */
    public boolean isProgramSV() {
	return false;
    }

    /**
     * returns true if the schemavariable can match a list of syntactical
     * elements 
     * @return true if the schemavariable can match a list of syntactical
     * elements 
     */
    public boolean isListSV() {
	return listSV;
    }


    /** 
     * @return true iff this SchemaVariable is used to match
     * modal operators
     */
    public boolean isModalOperatorSV() {
	return false;
    }
 
    /** returns true iff this SchemaVariable is used to match (create)
     * a skolem term
     * @return true iff this SchemaVariable is used to match (create)
     * a skolem term
     */
    public boolean isDependingSV() {
	return false;
    }

    /** @return arity of the Variable as int */
    public int arity() {
	return 0;
    }

    /** toString */
    public String toString(String sortSpec) {
	return name()+" ("+sortSpec+")"; 
    }
    
    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return isRigid ();
    }

    public boolean isRigid () {
	return false;
    }

    
    
    // Stuff needed for ProgramSV 
    /** @returns comments if the schemavariable stands for programm
     * construct and has comments attached to it (not supported yet)
     */
    public Comment[] getComments() {
	return new Comment[0];
    }


    public SourceElement getFirstElement(){
	return null;
    }

    public SourceElement getLastElement(){
	return null;
    }

    /**
     *        Returns the start position of the primary token of this element.
     *        To get the start position of the syntactical first token,
     *        call the corresponding method of <CODE>getFirstElement()</CODE>.
     *        @return the start position of the primary token.
     */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
     *        Returns the end position of the primary token of this element.
     *        To get the end position of the syntactical first token,
     *        call the corresponding method of <CODE>getLastElement()</CODE>.
     *        @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
     *        Returns the relative position (number of blank heading lines and 
     *        columns) of the primary token of this element.
     *        To get the relative position of the syntactical first token,
     *        call the corresponding method of <CODE>getFirstElement()</CODE>.
     *        @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return Position.UNDEFINED;
    }

    public PositionInfo getPositionInfo(){
	return  PositionInfo.UNDEFINED;
    }


    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public ReferencePrefix setReferencePrefix(ReferencePrefix r) {
	return this;
    }

    public int getDimensions(){
	return 0;
    }

    public int getTypeReferenceCount(){
	return 0;
    }

    public TypeReference getTypeReferenceAt(int index) {
	return this;
    }

    public PackageReference getPackageReference() {
	return null;
    }

    public int getExpressionCount() {
	return 0;
    }

    public Expression getExpressionAt(int index) {
	return null;
    }

    public int getChildCount() {
	return 0;
    }

    public ProgramElement getChildAt(int index) {
	return this;
    }
    
    public int getIndexOfChild(ProgramElement child) {
	return 0;
    }
    
    public int getChildPositionCode(ProgramElement child) {
	return 0;
    }

    public int getIndexOfChild(int positionCode) {
	return 0;
    }

    public int getRoleOfChild(int positionCode) {
	return 0;
    }

    public int getStatementCount() {
	return 0;
    }

    public int size() {
	return 0;
    }

    public ArrayOfExpression getUpdates() {
        return null;
    }

    public ArrayOfLoopInitializer getInits() {
	return null;
    }


    public Statement getStatementAt(int i) {
	return this;
    }

    public ProgramElementName getProgramElementName() {
	return new ProgramElementName(toString());
    }    

    public String getName() {
	return name().toString();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnSchemaVariable(this);
    }

    /** this pretty printer method is for the program pretty printer
     * and needs not to be overwritten by ProgramSV but at the moment
     * it is not 
     */
    public void prettyPrint(PrettyPrinter w) throws IOException {	
	w.printSchemaVariable(this);
    }

    public KeYJavaType getKeYJavaType() {
	return null;
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {
	return null;
    }

    public KeYJavaType getKeYJavaType(Services javaServ, ExecutionContext ec) {
	return null;
    }

    public boolean mayBeAliasedBy(Location loc) {
        return true;        
    }
    
    /**
     * checks if this schemavariable matches element <code>subst</code>
     * with respect to the given conditions and returns the resulting
     * match conditions (or null if the match is not possible)
     * @param subst the SVSubstitute to be matched
     * @param mc the MatchConditions posing restrictions on the
     * variables to be matched
     * @param services the Services with additional information about
     * the context (e.g. type hierarchy)
     * @return the resulting MatchConditions
     */
    public MatchConditions match(SVSubstitute subst, 
				 MatchConditions mc, Services services) {
        if (subst instanceof Term) {
            return addInstantiation((Term)subst, mc, services);                      
        }
        // overwritten in variable sv to match logic variables
        Debug.out("FAILED. Schemavariable of this kind only match terms.");
        return null;
    }
    
    /**
     * checks if this schemavariable matches element <code>subst</code>
     * with respect to the given conditions and returns the resulting
     * match conditions (or null if the match is not possible)
     * @param subst the Operator to be matched
     * @param mc the MatchConditions posing restrictions on the
     * variables to be matched
     * @param services the Services with additional information about
     * the context (e.g. type hierarchy)
     * @return the resulting MatchConditions
     */
    public MatchConditions match(Operator subst, 
				 MatchConditions mc, 
				 Services services) {
        if (subst instanceof SVSubstitute) {
            return match((SVSubstitute)subst, mc, services);
        } else {
            return null;
        }
    }
    

    /**
     * tries to add the pair <tt>(this,pe)</tt> to the match
     * conditions. If possible the resulting match conditions are
     * returned, otherwise <tt>null</tt>. Such an addition can fail,
     * e.g. if already a pair <tt>(this,x)</tt> exists where
     * <tt>x<>pe</tt>
     */
    protected MatchConditions addInstantiation(ProgramElement pe, 
					       MatchConditions matchCond, 
					       Services services) {       
        
        final SVSubstitute inMap = (SVSubstitute) matchCond.
	    getInstantiations ().getInstantiation(this);
        
        if (inMap == null) {                           
            try {
                return matchCond.setInstantiations ( matchCond.getInstantiations ()
						     .add(this, pe));
            } catch ( IllegalInstantiationException e ) {
                Debug.out("Exception thrown by class Taclet at setInstantiations");
            }
        } else {
	    Object peForCompare = pe;
            if (inMap instanceof Term) {
                try {
                    peForCompare = services.getTypeConverter().convertToLogicElement(pe);
                } catch (RuntimeException re) {
                    Debug.out("Cannot convert program element to term.", this, pe);
                    return null;
                }
            } 
	    if (inMap.equals(peForCompare)) {
		return matchCond;
	    }
        }
        Debug.out("FAILED. Illegal Instantiation.", this, pe);	    
        return null;        
    }
    
    /**
     * tries to add the pair <tt>(this,term)</tt> to the match
     * conditions. If successful the resulting conditions are returned,
     * otherwise null. Failure is possible e.g. if this schemavariable
     * has been already matched to a term <tt>t2</tt> which is not
     * unifiable with the given term.
     */
    protected MatchConditions addInstantiation 
	(Term term, MatchConditions matchCond, Services services) {           
        Term t = null;
        try {
            t = matchCond.getInstantiations ().
		getTermInstantiation(this, 
				     matchCond.
				     getInstantiations().getExecutionContext(), 
				     services);
        } catch (IllegalInstantiationException e) {
            return null;
        }
        if (t != null) {
            Constraint c = matchCond.getConstraint ().unify ( t, term );
            if ( !c.isSatisfiable () ) {
                Debug.out
		    ("FAILED. Adding instantiations leads to unsatisfiable constraint.",
		     this, term); 
                return null; //FAILED;
            } else {
                return matchCond.setConstraint ( c );
            }
        }                           
        try {
            return matchCond.setInstantiations ( matchCond.getInstantiations ()
						 .add(this, term) );
        } catch ( IllegalInstantiationException e ) {
            Debug.out("Exception thrown by class Taclet at setInstantiations");
        }
        Debug.out("FAILED. Illegal Instantiation");	    
        return null;
    }

    /**
     *@return the annotations.
     */
    public Annotation[] getAnnotations(){
	return new Annotation[0];
    }

    public int getAnnotationCount(){
	return 0;
    }
    
}
