// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.ListOfTerm;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.SLListOfTerm;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.util.Debug;

// to be replaced by HashMapFromIntegerToSimultaneousUpdateOp
import java.util.HashMap;
import java.util.TreeMap;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Comparator;

/**
 * <p>
 * State transitions are accumulated in updates as
 * <code>location := value</code> pairs. A formula <code> phi </code> as in
 * </p>
 * <p>
 * <code>
 *   {loc_1 := val_1, ..., loc_n := val_n} Phi
 * </code>
 * </p>
 * is evaluated in a state where <code>loc_j</code> has value
 * <code>val_j</code>.
 * </p>
 * <p>
 * The update operator is uniquely defined by the contained locations and their
 * order. This has some odd consequences of the sub terms of an update term as
 * given above:
 * <ol>
 * <li>Let <code> loc_1,...,loc_i </code> be local or static variables then
 * only the corresponding values <code>val_1, ..., val_n</code> are subterms
 * the locations itself are part of the operator.</li>
 * <li>For <code>loc_i (=r_i.a_i),...,loc_n(=r_n.a_n)</code> that refer to an
 * instance attribute, only their reference prefix <code>r_i</code> is a
 * subterm while the attribute <code> a_i </code> is part of the update
 * operator. If <code>loc_i</code> is an array <code>a[l]</code> then the
 * sort of <code> a </code> is fixed and <code>a, i</code> are sub terms</li>
 * </ol>
 * There are methods that return the location access part (attribute, local or
 * static variable) and also the complete n-th location as term.
 *
 * 
 * @invariant locations <>null -> arity = locations.size()
 */
public class UpdateOperator implements Operator, NonRigid {

    /**
     * a small class realising a tuple of integer values
     */
    private static class Tuple {

        public final int locPos;

        public final int oldValuePos;

        public Tuple(int locPos, int oldValuePos) {
            this.locPos = locPos;
            this.oldValuePos = oldValuePos;
        }

    }

    /**
     * name of the operator
     */
    private final Name name;

    /** the arity of this operator */
    private final int arity;

    /** the locations access operators used by the update op */
    private final ArrayOfLocation locations;

    /**
     * table to determine the position of the value to a given location
     */
    private final int[] valuePositionTable;

    private static final HashMap updates = new HashMap();

    /**
     * returns the update operator for the given location order
     */
    public static UpdateOperator createUpdateOp(ListOfLocation locs) {
        return createUpdateOp(locs.toArray());
    }

    /**
     * returns the update operator for the given location order
     */
    public static UpdateOperator createUpdateOp(Term[] locs) {
        Location[] locOps = new Location[locs.length];
        for (int i = locs.length - 1; i >= 0; i--) {
            locOps[i] = (Location) locs[i].op();
        }
        return createUpdateOp(locOps);
    }

    /**
     * returns the update operator for the given location order
     */
    public static UpdateOperator createUpdateOp(Location[] locs) {

        ArrayOfOperator locWrapper = new ArrayOfOperator(locs);
        UpdateOperator result = (UpdateOperator) updates.get(locWrapper);
        if (result == null) {
            result = new UpdateOperator(locs);
            updates.put(locWrapper, result);
        }
        return result;
    }

    /**
     * returns the normalized update term specified by the given 
     * locations, values and target term
     */
    public static Term normalize(Term[] locations, Term[] values, Term target) {
        final Location[] locs = new Location[locations.length];

        for (int i = 0; i < locs.length; i++) {
            locs[i] = (Location) locations[i].op();
        }

        Tuple[] normalformOrdering = normalize(locs);

        if (normalformOrdering == null) {
            ListOfTerm resultSubs = SLListOfTerm.EMPTY_LIST.prepend(target);
            for (int i = locs.length - 1; i >= 0; i--) {
                resultSubs = resultSubs.prepend(values[i]);
                for (int j = locs[i].arity() - 1; j >= 0; j--) {
                    resultSubs = resultSubs.prepend(locations[i].sub(j));
                }
            }
            return TermFactory.DEFAULT.createUpdateTerm(createUpdateOp(locs),
                    resultSubs.toArray());
        }

        LinkedList result = new LinkedList();

        final Location[] locsInNormalformOrder = new Location[normalformOrdering.length];
        for (int i = 0; i < normalformOrdering.length; i++) {
            final Term locTerm = locations[normalformOrdering[i].locPos];
            locsInNormalformOrder[i] = (Location) locTerm.op();
            for (int j = 0, arity = locTerm.arity(); j < arity; j++) {
                result.add(locTerm.sub(j));
            }
            result.add(values[normalformOrdering[i].locPos]);
        }

        result.add(target);

        return TermFactory.DEFAULT.createUpdateTermUnordered(
                locsInNormalformOrder, (Term[]) result.toArray(new Term[result
                        .size()]));
    }

    /**
     * returns the normalform of the update term specified by the given
     * locations and subterms
     */
    public static Term normalize(Location[] locations, Term[] subs) {

        final Tuple[] normalformOrdering = normalize(locations);

        if (normalformOrdering == null) {
            return TermFactory.DEFAULT.createUpdateTerm(
                    createUpdateOp(locations), subs);
        }

        LinkedList result = new LinkedList();

        final Location[] locsInNormalformOrder = new Location[normalformOrdering.length];
        for (int i = 0; i < normalformOrdering.length; i++) {
            locsInNormalformOrder[i] = locations[normalformOrdering[i].locPos];
            for (int j = locsInNormalformOrder[i].arity(); j >= 0; j--) {
                result.add(subs[normalformOrdering[i].oldValuePos - j]);
            }
        }

        result.add(subs[subs.length - 1]);

        return TermFactory.DEFAULT.createUpdateTermUnordered(
                locsInNormalformOrder, (Term[]) result.toArray(new Term[result
                        .size()]));
    }

    /**
     * returns the normalform of the update term specified by the given
     * locations and subterms
     */
    public static Term normalize(ArrayOfLocation locations, Term[] subs) {
	final Location[] locs = new Location[locations.size()];
	locations.arraycopy(0, locs, 0, locations.size());
	return normalize(locs, subs);
    }

    /**
     * determines the ordering of the locations according to the update
     * term normalform and returns an array of tuples describing the new
     * order of the locations
     */
    private static Tuple[] normalize(Location[] locations) {

        final TreeMap accessOperators = new TreeMap(
                Classifier.DEFAULT_CLASSIFIER);
        final TreeMap sortedStaticVars = new TreeMap(
                Classifier.DEFAULT_CLASSIFIER);
        final TreeMap sortedLocalVars = new TreeMap(
                Classifier.DEFAULT_CLASSIFIER);

        int valuePos = 0;
        for (int i = 0; i < locations.length; i++) {
            final Operator location = locations[i];
            valuePos += location.arity();
            if (location instanceof ProgramVariable) {
                if (((ProgramVariable) location).isStatic()) {
                    // extract static variables from locs and remove
                    // duplicates (just keep last occurence see also
                    // last-win semantics of updates)
                    sortedStaticVars.put(location, new Tuple(i, valuePos));
                } else {
                    // extract local variable from locs and remove
                    // duplicates (just keep last occurence; see also
                    // last-win semantics of updates)
                    sortedLocalVars.put(location, new Tuple(i, valuePos));
                }
            } else if (location instanceof AccessOp) {
                // extract attributes/arrays and attribute/arrays chains
                LinkedList positions = (LinkedList) accessOperators
                        .get(location);
                if (positions == null) {
                    positions = new LinkedList();
                }
                positions.add(new Tuple(i, valuePos));
                accessOperators.put(location, positions);
            } else if (location instanceof SchemaVariable
                    || location instanceof MetaOperator) {
                return null;
            }
            valuePos++;
        }

        final TreeMap[] maps = new TreeMap[] { sortedLocalVars,
                sortedStaticVars, accessOperators };

        final LinkedList normalformOrder = new LinkedList();

        for (int i = 0; i < maps.length; i++) {
            Iterator it = maps[i].values().iterator();
            while (it.hasNext()) {
                Object value = it.next();
                if (value instanceof LinkedList) {
                    normalformOrder.addAll((LinkedList) value);
                } else {
                    normalformOrder.add(value);
                }
            }
        }

        return (Tuple[]) 
        normalformOrder.toArray(new Tuple[normalformOrder.size()]);
    }

    /**
     * creates the update operator for the given locations in the given order
     * 
     * @param locs array of Locations
     */
    private UpdateOperator(Location[] locs) {
        this(new ArrayOfLocation(locs));     
    }

    /**
     * creates the update operator for the given locations in the given order
     * 
     * @param locs array of Locations
     */
    protected UpdateOperator(Name name, ArrayOfLocation locs) {
        this.name = name;
        this.locations = locs;
        this.valuePositionTable = new int[locations.size()];

        fillValuePositionTable();
        this.arity = locs.size() == 0 ? 1
                : valuePositionTable[valuePositionTable.length - 1] + 2;
    }

    /**
     * the given locations are the one to be updated by this operator  
     * @param location
     */
    public UpdateOperator(ArrayOfLocation location) {
        this(new Name("update("+location+")"), location); 
    }

    /**
     * returns the array of location operators which are updated
     */
    public ArrayOfLocation locationsAsArray() {
	return locations;
    }

    /**
     * returns the number of locations
     * 
     * @return the number of locations as int
     */
    public int locationCount() {
        return locations.size();
    }

    /**
     * returns the operator of <tt>n</tt>-th location
     */
    public Location location(int n) {
        return locations.getLocation(n);
    }

    /**
     * returns the number of the subterm representing the value to which
     * the locPos-th location is updated
     */
    public int valuePos(int locPos) {
        return valuePositionTable[locPos];
    }

    /**
     * returns an array with all location positions of <code>loc</code>
     * 
     * @param loc a n Operator accessing a location
     * @return all location positions of <code>loc</code>
     */
    public int[] getLocationPos(Operator loc) {
        int size = 0;
        final int[] tempResult = new int[locationCount()];
        for (int i = 0; i < tempResult.length; i++) {
            if (location(i) == loc) {
                tempResult[size] = i;
                size++;
            }
        }
        final int[] result = new int[size];
        System.arraycopy(tempResult, 0, result, 0, size);

        return result;
    }

    /**
     * computes the position table used to determine at which sub term position
     * the value used to update the i-th location can be found.
     */
    private void fillValuePositionTable() {
        int pos = 0;
        for (int i = 0; i < valuePositionTable.length; i++) {
            pos += location(i).arity();
            valuePositionTable[i] = pos;
            pos++;
        }
    }

    /**
     * returns the n-th location of the update as term
     * 
     * @param t
     *            Term with this operator as top level operator
     * @param n
     *            an int specifying the position of the requested location
     * @return the n-th location of term t updated by this operator
     */
    public Term location(Term t, int n) {
        Debug.assertTrue(t.op() == this, "Illegal update location access.");

        final Location op = location(n);
        final Term[] sub = new Term[op.arity()];
        final ArrayOfQuantifiableVariable[] vars = new ArrayOfQuantifiableVariable[op
                .arity()];
        final int basisPos = valuePositionTable[n];

        for (int i = 1; i <= sub.length; i++) {
            sub [sub.length - i] = t.sub(basisPos - i);
            vars[sub.length - i] = t.varsBoundHere(basisPos - i);
        }
        return TermFactory.DEFAULT.createTerm
	    (op, sub, vars, JavaBlock.EMPTY_JAVABLOCK, null);
    }

    /**
     * returns value the n-th location is updated with
     * 
     * @param t
     *            Term with this operator as top level operator
     * @param n
     *            an int specifying the position of the location
     * @return the value the n-th location is updated with
     */
    public Term value(Term t, int n) {
        Debug.assertTrue(t.op() == this, "Illegal update value access.");
        return t.sub(valuePositionTable[n]);
    }

    /** @return name of the operator */
    public Name name() {
        return name;
    }

    /**
     * returns the arity of the operator
     */
    public int arity() {
        return arity;
    }

    /**
     * returns the sort a term would have if constructed with this operator and
     * the given sunterms. It is assumed that the term would be allowed if
     * constructed.
     * 
     * @param term
     *            an array of Term containing the subterms of a (potential)
     *            simultaneous update term
     * @return sort of the simultaneous update term consisting of the given
     *         subterms, if this op would be the top symbol
     */
    public Sort sort(Term[] term) {
        return term[arity() - 1].sort();
    }

    /**
     * @return true if the value of "term" having this operator as top-level
     *         operator and may not be changed by modalities
     */
    public boolean isRigid(Term term) {
        return term.sub(arity() - 1).isRigid();
    }

    /**
     * checks whether the top level structure of the given {@link Term}is
     * syntactically valid, given the assumption that the top level operator of
     * the term is the same as this Operator. The assumption that the top level
     * operator and the operator are equal is NOT checked.
     * 
     * @return true iff the top level structure of the {@link Term}is valid.
     */
    public boolean validTopLevel(Term term) {
        return true;
    }

    // inner class that sorts the updates in a safe way (kind of normal form)
    static class Classifier implements Comparator {

        static final Classifier DEFAULT_CLASSIFIER = new Classifier();

        private int primitiveLongCompare(long i, long j) {
            return (i < j ? -1 : (i == j ? 0 : 1));
        }

        public int compare(Object o1, Object o2) {
            if (o1 == o2)
                return 0;
            Named pv1 = (Named) o1;
            Named pv2 = (Named) o2;
            int cmpResult = pv1.name().compareTo(pv2.name());

            if (cmpResult == 0) {
                boolean bPV1 = (pv1 instanceof ProgramVariable);
                boolean bPV2 = (pv2 instanceof ProgramVariable);
                cmpResult = bPV1 ? (bPV2 ? primitiveLongCompare(
                        ((ProgramVariable) pv1).id(), ((ProgramVariable) pv2)
                                .id()) : -1) : (bPV2 ? 1 : 0);
            }
            return cmpResult;
        }
    }

    /**    
     * two concrete update operators match if their defining locations
     * match 
     * @see de.uka.ilkd.key.logic.op.Operator#match
     * (de.uka.ilkd.key.logic.op.Operator,
     * de.uka.ilkd.key.rule.MatchConditions, de.uka.ilkd.key.java.Services)
     */
    public MatchConditions match(Operator op, MatchConditions mc, Services services) {
        MatchConditions result = mc;
        if (this == op) return result;
        if (op.getClass() != getClass()) {
            Debug.out("FAILED matching. Incomaptible operators " +
    		"(template, operator)", this, op);
            return null;
        }      
        final UpdateOperator update = (UpdateOperator)op;
        if (locationCount() == update.locationCount()) {
            for (int i = 0, locs = locationCount(); i<locs; i++) {
                result = location(i).match(update.location(i), result, services);
                if (result == null) { // fail fast
                    Debug.out
                    ("FAILED. UpdateOperator location mismatch " +
                            "(template, operator)", this, update);
                    return null;
                }
            }
            return result;
        }        
        Debug.out("FAILED matching. UpdateOperator match failed because of" +
                " incompatible locations (template, operator)", this, op);
        return null;
    }

    /**
     * returns a string representation of the update operator 
     */
    public String toString() {
        StringBuffer sb = new StringBuffer("update(");
        for (int i = 0; i < locations.size(); i++) {
            sb.append(locations.getLocation(i));
            if (i < locations.size() - 1) {
                sb.append(",");
            }
        }
        sb.append("}");
        return sb.toString();
    }

}
