// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.sort;


import java.util.WeakHashMap;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.op.SortDependingSymbol;
import de.uka.ilkd.key.logic.op.InstanceofSymbol;
import de.uka.ilkd.key.logic.op.ListOfSortDependingSymbol;
import de.uka.ilkd.key.logic.op.SLListOfSortDependingSymbol;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.MapAsListFromNameToSortDependingSymbol;
import de.uka.ilkd.key.logic.op.MapFromNameToSortDependingSymbol;
import de.uka.ilkd.key.logic.op.IteratorOfSortDependingSymbol;
import de.uka.ilkd.key.util.Debug;



public class ArraySortImpl extends AbstractCollectionSort implements ArraySort{

    private static WeakHashMap aSH = new WeakHashMap();
    private Sort oSort = null;
    private Sort cSort = null;

    private SetOfSort extendsSorts;

    private MapFromNameToSortDependingSymbol definedSymbols =
	MapAsListFromNameToSortDependingSymbol.EMPTY_MAP;

    private boolean symbolsCreated = false;


    private ArraySortImpl(Sort elemSort, SetOfSort extendsSorts, 
			  Sort objectSort, Sort cloneableSort) {
	super("ArrayOf"+elemSort.name(),elemSort);
	oSort = objectSort;
	cSort = cloneableSort;
	// uncommented because of ASMKeY:
	//	Debug.assertTrue(extendsSorts!=SetAsListOfSort.EMPTY_SET);
	this.extendsSorts = extendsSorts;
    }

    /**
     * @return an object of this class with elementSort().equals(p),
     * or null if such an object cannot be constructed (as p is an
     * incompatible sort).
     */
    public Sort cloneFor ( Sort p ) {
	return getArraySort ( p, oSort, cSort);
    }

    /**
     * @return the sorts of the predecessors of this sort
     */
    public SetOfSort extendsSorts() {
	return extendsSorts;
    }

    /** 
     * returns elemSort([])^n.
     */
    public static Sort getArraySortForDim(Sort elemSort, int n, 
					  Sort objectSort, 
					  Sort cloneableSort){
	Sort result = elemSort;
	while(n>0){
	    result = getArraySort(result, objectSort, cloneableSort);
	    n--;
	}
	return result;
    }

    /**
     * returns the ArraySort to the given elementsort. This method ensures that
     * only one ArraySort-object exists for each Arraysort.
     */
    public static ArraySort getArraySort(Sort elemSort, 
					 Sort objectSort,
					 Sort cloneableSort){
	if(aSH.containsKey(elemSort)){
	    return (ArraySort) aSH.get(elemSort);
	}else{ 
        // HACK: this simple handling of sort creation does not treat
        // depending symbols (i.e. ArrayOfS::instance)
	    SetOfSort localExtendsSorts = getArraySuperSorts(elemSort, objectSort,
							cloneableSort);
	    ArraySort aS = new ArraySortImpl(elemSort, localExtendsSorts,
					     objectSort, cloneableSort);
	    aSH.put(elemSort,aS);
	    return aS;
	}
    }

    private static SetOfSort getArraySuperSorts(Sort elem, 
						Sort objectSort,
						Sort cloneableSort){
	int i = 1;
	while(elem instanceof ArraySort){
	    elem = ((ArraySort) elem).elementSort();
	    i++;
	}
	SetOfSort superSorts = elem.extendsSorts();
	if(elem instanceof PrimitiveSort || elem == objectSort){
	    i--;
	    superSorts = superSorts.add(cloneableSort);
	}
	if(i>0){
	    IteratorOfSort it = superSorts.iterator();
	    superSorts = SetAsListOfSort.EMPTY_SET;
	    while(it.hasNext()){
		Sort s = it.next();
		superSorts = 
		    superSorts.add(getArraySortForDim(s, i, 
						      objectSort, 
						      cloneableSort));
	    }
	}
	return superSorts;
    }


    /**
     * Lookup the symbol of kind "p_name", which is a sort
     * independent identifier for this symbol
     * @return Symbol with (kind) name "p_name"
     * ("ret.getKind().equals(p_name)"), null if no such object exists
     */
    public SortDependingSymbol lookupSymbol  ( Name      p_name ) {
	return definedSymbols.get ( p_name );
    }


    protected void addSymbols ( ListOfSortDependingSymbol p,
				Namespace                 p_func_ns ) {
	IteratorOfSortDependingSymbol it = p.iterator ();
	SortDependingSymbol           s;

	while ( it.hasNext () ) {
	    s              = it.next ();
	    definedSymbols = definedSymbols.put ( s.getKind (), s );
	    p_func_ns.add ( (Named)s );
	}
    }

    /**
     * Create the symbols defined by this sort, insert them into the
     * namespace "p_func_ns"
     * @param p_func_ns Namespace to which functions and predicates
     * should be added to
     */
    public void createSymbols ( Namespace     p_func_ns, Namespace sort_ns ) {
	if ( !symbolsCreated ) {
	    ListOfSortDependingSymbol list =
		SLListOfSortDependingSymbol.EMPTY_LIST;
	    Sort booleanSort=(Sort)sort_ns.lookup(new Name("boolean"));
	    list = list.prepend ( new InstanceofSymbol (this, booleanSort));
	    addSymbols ( list, p_func_ns );

	    symbolsCreated = true;
	}
    }


}
