// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
//
//

package de.uka.ilkd.key.ocl.gf;


import java.io.*;
import java.net.URL;

import de.uka.ilkd.key.util.KeYResourceManager;
import java.util.Arrays;
import org.apache.log4j.Logger;

/** @author Kristofer Johannisson and Hans-Joachim Daniels */

/** handle the copying of GF OCL grammars from the resource directory
 * to a temporary directory (so that GF can access them) */
class TempGrammarFiles {
    
    final static Logger logger = Logger.getLogger(TempGrammarFiles.class.getName());
    final static String grammars2 = "grammars2";
    //for debugging, one might want to keep the generated grammar files
    static boolean deleteOnExit = true;
    File oclgf2Dir = null;
    /// a directory in the tmp directory with a unique name
    File tempG2ParentDir = null;
    /// a directory in grammars2TempDir with name grammars2
    File tempGrammars2Dir = null;
    final static String[] grammar2URLs = {
         "prelude", "abstract", "english", "german", "ExtraEng", "ExtraGer", "."
    };
    final static String grammarFileListName = "grammars2list.txt"; 

    /** copy GF OCL grammars from internal resource directory to
	a temporary directory, where GF can access them */
    public TempGrammarFiles() throws IOException {
	    final URL oclgf2URL = KeYResourceManager.getManager().getResourceFile(TempGrammarFiles.class, grammars2);
	    final URL grammarFileList = KeYResourceManager.getManager().getResourceFile(TempGrammarFiles.class, grammars2 + File.separator + grammarFileListName);
	    final BufferedReader listReader = new BufferedReader(new FileReader(new File(grammarFileList.getFile())));
	    
	    //This is not really safe, so. could do sth. between delete and mkdir. 
	    tempG2ParentDir = File.createTempFile("grammars2_", null);
	    tempG2ParentDir.delete();
	    tempG2ParentDir.mkdir();
	    tempGrammars2Dir = new File (tempG2ParentDir.getPath() + File.separator + grammars2);
	    tempGrammars2Dir.mkdirs();
	    
	    int timer = 0;
	    for (String filename = listReader.readLine(); filename != null; filename = listReader.readLine()) {
	            if (File.separator == "\\") {
	                    filename = filename.replaceAll("/", "\\\\");
	            } else {
	                    //for weird operating systems this actually does sth.
	                    filename = filename.replaceAll("/", File.separator);
	            }
	            logger.debug("filename: " + filename);
	            File currentFile = new File(filename);
	            URL oldFileURL = KeYResourceManager.getManager().getResourceFile(TempGrammarFiles.class, grammars2 + File.separator + filename);
	            File oldFile = new File (oldFileURL.getPath());
	            File newFile = new File (tempGrammars2Dir.getAbsolutePath() + File.separator + filename);
	            if (logger.isDebugEnabled()) {
	                    logger.debug("Copy: " + oldFile.getAbsolutePath() + " --> " + newFile.getAbsolutePath());
	            }
	            if (!newFile.getParentFile().exists()) {
	                 newFile.getParentFile().mkdirs();   
	            }
	            newFile.createNewFile();
	            copyFile(oldFile, newFile);
	            //artificially set a modification time, so that GF knows that the gfc files
	            //are newer than the gf files.
	            //200 is a bit more than the current number of files.
	            //but having files modified in the future does not seem to hurt
	            //so left as a magic number
	            newFile.setLastModified(newFile.lastModified() - 200000 + (timer++ * 1000));
	    }
    }

    /**
     * returns the location of the OCL-GF2 grammar files in the temp directory.
     * @return the location of the OCL-GF2 grammar files in the temp directory.
     */
    public String path2grammars2() {
            return tempGrammars2Dir.getAbsolutePath();
    }
    
    /**
     * creates a file with the given ocl constraint as the content and returns the handle.
     * Every call generates a new File.
     * @param ocl the constraint to be written
     * @return a File object representing the temporary file
     * @throws IOException
     */
    public static File createOCLtmp(String ocl) throws IOException{
            File result = File.createTempFile("ocl_", null);
            PrintWriter toW = new PrintWriter(new BufferedWriter( new FileWriter(result)));
            toW.println(ocl);
            toW.close();
            perhapsDeleteOnExit(result);
            return result;
    }



    /**
     * If the flag deleteOnExit is set, the same named function will
     * be called on candidate
     * @param candidate the File that gets deleted on exit if set so
     */
    private static void perhapsDeleteOnExit(File candidate) {
            if (deleteOnExit) {
                    try {
                            candidate.deleteOnExit();
                    } catch (SecurityException e) {
                            logger.error(e.getLocalizedMessage());
                            e.printStackTrace();
                    }
            }
    }
    
    /** replace the contents of file to with the contents of file from */
    private static void simpleCopy(File from, File to) throws IOException {
            BufferedReader fromR = new BufferedReader(new FileReader(from));
            PrintWriter toW = new PrintWriter(new BufferedWriter( new FileWriter(to)));
            
            String s = new String();
            while((s=fromR.readLine()) != null) {
                    // System.out.println("** "+s);
                    toW.println(s);
            }
            toW.close();
            fromR.close();
            perhapsDeleteOnExit(to);
    }
    
    /**
     * copies the content of src into dest
     * Should dest already exist, will it get deleted first
     * @param src
     * @param dest
     * @throws IOException
     */
    private static void copyFile(File src, File dest) throws IOException {
            if(dest.exists()) {
                    dest.delete();
            }
            byte[] buffer = new byte[0xffff];
            int read = 0;
            InputStream in = null;
            OutputStream out = null;
            try {
                    in = new FileInputStream(src);
                    out = new FileOutputStream(dest);
                    while(true) {
                            read = in.read(buffer);
                            if (read == -1) {
                                    //-1 means EOF
                                    break;
                            }
                            out.write(buffer, 0, read);
                    }
            } finally {
                    //close the streams
                    if (in != null) {
                            try {
                                    in.close();
                            }
                            finally {
                                    if (out != null) {
                                            out.close();
                                    }
                            }
                    }
            }
    }

    public static void cleanupFromUMLTypes(String grammarsDir) {
            String[] endings = {"Eng.gf", "Eng.gfc", "Ger.gf", "Ger.gfc", "OCL.gf", "OCL.gfc", ".gf", ".gfc"};
            for (int i = 0; i < endings.length; i++) {
                    String filename = grammarsDir + File.separator + GFinterface.modelModulName + endings[i];
                    File file = new File(filename);
                    file.deleteOnExit();
                    if (logger.isDebugEnabled()) {
                            logger.debug("scheduled for deletion: " + filename);
                    }
            }
            File file = new File(grammarsDir);
            file.deleteOnExit();
            file = file.getParentFile();
            file.deleteOnExit();
    }
}
