// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.pp;

import java.util.HashMap;

import de.uka.ilkd.key.logic.ldt.NumberLDT;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.util.Service;


/** Stores the mapping from operators to {@link Notation}s.  Each 
 * {@link Notation} represents the concrete syntax for some 
 * {@link de.uka.ilkd.key.logic.op.Operator}.  The {@link LogicPrinter}
 * asks the NotationInfo to find out which Notation to use for a given term.
 * 
 * The Notation associated with an operator might change.  New Notations can
 * be added.
 */
public class NotationInfo {

    /** Factory method: creates a new NotationInfo instance. The
     * actual implementation depends on system properties or service
     * entries. */
    public static final NotationInfo createInstance() {
	return (NotationInfo) Service.find(NotationInfo.class.getName(),
					   NotationInfo.class.getName());
    }
    
    
    /** This maps operators and classes of operators to {@link
     * Notation}s.  The idea is that we first look wether the operator has
     * a Notation registered.  Otherwise, we see if there is one for the
     * <em>class</em> of the operator.
     */
    private HashMap tbl;

    /**
     * Maps terms to abbreviations and reverse.
     */
    private AbbrevMap scm;


    /** Create a new NotationInfo. Do not call this constructor
     * directly. Use the factory method {@link createInstance}
     * instead. */
    public NotationInfo() {
    	createDefaultNotationTable();
    }

    /** Set all notations back to default. */
    public void setBackToDefault() {
    	createDefaultNotationTable();
    }
    
    /** Register the standard set of notations.  This means no
     * abbreviations, and a set of Notations for the built-in operators
     * which corresponds to the parser syntax. 
     */
    private void createDefaultNotationTable() {
		tbl=new HashMap();
		createDefaultOpNotation();
		createDefaultTermSymbolNotation();
		scm = new AbbrevMap();
    }

    /**
     * Registers notations for the built-in operators.  The priorities
     * and associativities correspond to the parser syntax.  
     */
    private void createDefaultOpNotation() {
	tbl.put(Op.TRUE ,new Notation.Constant("true",80));
	tbl.put(Op.FALSE,new Notation.Constant("false",80));
	tbl.put(Op.NOT,new Notation.Prefix("!" ,60,60));
	tbl.put(Op.AND,new Notation.Infix("&"  ,50,60,50));
	tbl.put(Op.OR, new Notation.Infix("|"  ,40,50,40));
	tbl.put(Op.IMP,new Notation.Infix("->" ,30,40,30));
	tbl.put(Op.EQV,new Notation.Infix("<->",20,30,20));
	tbl.put(Op.LIST,new Notation.ModifierList(",",70,80,70));
    	tbl.put(Op.ALL,new Notation.Quantifier("all",60,60));
	tbl.put(Op.EX, new Notation.Quantifier("ex", 60,60));
	tbl.put(Op.DIA,new Notation.Modality("<",">",60,60));
	tbl.put(Op.BOX,new Notation.Modality("[","]",60,60));
	tbl.put(Op.TOUT,new Notation.Modality("[[","]]",60,60));
	tbl.put(Op.DIATRC,new Notation.Modality("diamond_trc{","}",60,60));
	tbl.put(Op.BOXTRC,new Notation.Modality("box_trc{","}",60,60));
	tbl.put(Op.TOUTTRC,new Notation.Modality("throughout_trc{","}",60,60));
	tbl.put(Op.DIATRA,new Notation.Modality("diamond_tra{","}",60,60));
	tbl.put(Op.BOXTRA,new Notation.Modality("box_tra{","}",60,60));
	tbl.put(Op.TOUTTRA,new Notation.Modality("throughout_tra{","}",60,60));
	tbl.put(Op.IF_ELSE, new Notation.IfElse());

	tbl.put(Op.COMPUTE_SPEC_OP,new Notation.Prefix("^" ,60,60));

	createNumLitNotation(NumberLDT.getNumberSymbol());
	tbl.put(Op.SUBST,new Notation.Subst());
	tbl.put(Op.WARY_SUBST_EX,new Notation.Subst());
	tbl.put(Op.WARY_SUBST_ALL,new Notation.Subst());
    }    

    /** 
     * Register notations for standard classes of operators.  This
     * includes Function operators, all kinds of variables, etc.
     */
    private void createDefaultTermSymbolNotation() {
		tbl.put(Function.class, new Notation.Function());
		tbl.put(LogicVariable.class, new Notation.Variable());
		//tbl.put(SchemaVariable.class, new Notation.Variable());
		tbl.put(Metavariable.class, new Notation.Metavariable());
		tbl.put(ProgramVariable.class, new Notation.Variable());
		tbl.put(ProgramMethod.class, new Notation.ProgramMethod());
		tbl.put(Equality.class, new Notation.Infix("=", 70, 80, 80)); 
		tbl.put(UpdateOperator.class, new Notation.Update());
		tbl.put(AnonymousUpdate.class, new Notation.Update());
		tbl.put(ShadowAttributeOp.class, new Notation.ShadowAttribute(120,80));
		tbl.put(AttributeOp.class, new Notation.Attribute(120, 80));
		tbl.put(ShadowArrayOp.class, new Notation.ShadowArrayNot("[","]",120));
		tbl.put(ArrayOp.class, new Notation.ArrayNot("[","]",120));
		tbl.put(ConjCond.class, new Notation.ConjCond());
		tbl.put(ModalOperatorSV.class, new Notation.ModalSV(60, 60));
		tbl.put(SortedSchemaVariable.class, new Notation.SortedSchemaVariable());
    }

    public AbbrevMap getAbbrevMap(){
	return scm;
    }

    public void setAbbrevMap(AbbrevMap am){
	scm = am;
    }

    /** Get the Notation for a given Operator.  
     * If no notation is registered, a Function notation is returned.
     */
    public Notation getNotation(Operator op) {
	createNumLitNotation(NumberLDT.getNumberSymbol());
	createCharLitNotation(NumberLDT.getCharSymbol());
	//For OCL Simplification
	if (tbl.containsKey(op.name().toString())) {
	    return (Notation) tbl.get(op.name().toString());
	}
	//
	else if (tbl.containsKey(op)) {
	    return (Notation) tbl.get(op);
	} else if (tbl.containsKey(op.getClass())) {
	    return (Notation) tbl.get(op.getClass());
	} else if (op instanceof SortedSchemaVariable){
		return (Notation) tbl.get(SortedSchemaVariable.class);
	} else {
	    return new Notation.Function();
	}
    }

    /** Registers an infix notation for a given operator
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createInfixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Infix(token,90,100,100));
    }

    /** Registers a prefix notation for a given operator 
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createPrefixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Prefix(token, 110, 100));
    }

    /** Registers a number literal notation for a given operator.
     * This is done for the `Z' operator which marks number literals.
     * A term <code>Z(3(2(#)))</code> gets printed simply as
     * <code>23</code>.
     * @param op the operator */
    public void createNumLitNotation(Operator op) {
	tbl.put(op, new Notation.NumLiteral());
    }


    /** Registers a character literal notation for a given operator.
     * This is done for the `C' operator which marks character literals.
     * A term <code>C(3(2(#)))</code> gets printed simply as
     * the character corresponding to the unicode value 23 (really 23
     * and not 32, see integer literals)
     * @param op the operator */
    public void createCharLitNotation(Operator op) {
	tbl.put(op, new Notation.CharLiteral());
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for iterate().
     */
    public void createOCLIterateNotation(String token) {
	tbl.put("$"+token, new Notation.OCLIterate());
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for forAll(), exists(), etc.
     */
    public void createOCLCollOpBoundVarNotation(String token) {
	tbl.put("$"+token, new Notation.OCLCollOpBoundVar(token));
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for size(), includes(), etc.
     */
    public void createOCLCollOpNotation(String token) {
	tbl.put("$"+token, new Notation.OCLCollOp(token));
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for supertypes(), oclIsNew(), etc.
     */
    public void createOCLDotOpNotation(String token) {
	tbl.put("$"+token, new Notation.OCLDotOp(token));
    }

    /** 
     * Used for OCL Simplification.
     */
    public void createOCLWrapperNotation(String token) {
	tbl.put("$"+token, new Notation.OCLWrapper());
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for Set{...}, Bag{...}, and Sequence{...}.
     */
    public void createOCLCollectionNotation(String internalName, String token) {
	tbl.put("$"+internalName, new Notation.OCLCollection(token));
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for self, true, false, Set{}, etc.
     */
    public void createConstantNotation(String internalName, String token) {
	tbl.put("$"+internalName, new Notation.Constant(token, 100));
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for if-then-else-endif.
     */
    public void createOCLIfNotation(String token) {
	tbl.put("$"+token, new Notation.OCLIf());
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for "not" and "-".
     */
    public void createOCLPrefixNotation(String internalName, String token, int prio, int ass) {
	tbl.put("$"+internalName, new Notation.Prefix(token, prio, ass));
    }

    /** 
     * Used for OCL Simplification.
     * Syntax for "and", "or", "+", "<", etc.
     */
    public void createOCLInfixNotation(String internalName, String token, int prio, 
				       int assLeft, int assRight) {
	tbl.put("$"+internalName, new Notation.Infix(token, prio, assLeft, assRight));
    }

    /** 
     * Used for OCL Simplification.
     */
    public void createOCLInvariantNotation(String token) {
	tbl.put("$"+token, new Notation.OCLInvariant());
    }

    /** 
     * Used for OCL Simplification.
     */
    public void createOCLListOfInvariantsNotation(String token) {
	tbl.put("$"+token, new Notation.OCLListOfInvariants());
    }
}
