// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.ConstrainedFormula;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.FormulaChangeInfo;
import de.uka.ilkd.key.logic.IteratorOfConstrainedFormula;
import de.uka.ilkd.key.logic.ListOfConstrainedFormula;
import de.uka.ilkd.key.logic.ListOfFormulaChangeInfo;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.PosInTerm;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.logic.SequentChangeInfo;
import de.uka.ilkd.key.rule.BuiltInRule;
import de.uka.ilkd.key.rule.BuiltInRuleApp;
import de.uka.ilkd.key.rule.IteratorOfBuiltInRule;
import de.uka.ilkd.key.rule.ListOfRuleApp;
import de.uka.ilkd.key.rule.SLListOfRuleApp;
import de.uka.ilkd.key.rule.UpdateSimplificationRule;

public class BuiltInRuleAppIndex implements java.io.Serializable {

    private BuiltInRuleIndex index;

    private NewRuleListener  newRuleListener =
        NullNewRuleListener.INSTANCE;
    
    
    public BuiltInRuleAppIndex(BuiltInRuleIndex index) {
	this.index = index;
    }
    
    
    public BuiltInRuleAppIndex(BuiltInRuleIndex index,
			       NewRuleListener  p_newRuleListener) {
	this.index           = index;
	this.newRuleListener = p_newRuleListener;
    }
    
    
    /** 
     * returns a list of built-in rules application applicable
     * for the given goal and position
     */
    public ListOfRuleApp getBuiltInRule(Goal            goal, 
					PosInOccurrence pos, 
					Constraint      userConstraint) {

	ListOfRuleApp result = SLListOfRuleApp.EMPTY_LIST;
	IteratorOfBuiltInRule it = index.rules().iterator();

	while (it.hasNext()) {
	    BuiltInRule bir = it.next();
	    result = bir.isApplicable(goal, pos, userConstraint) ?
		result.prepend(new BuiltInRuleApp(bir, pos, userConstraint)) :
		result;
	}

	return result;
    }


    /** 
     * returns a copy of this index
     */
    public BuiltInRuleAppIndex copy() {
	return new BuiltInRuleAppIndex(index.copy());
    }

    public void setNewRuleListener ( NewRuleListener p_newRuleListener ) {
    	newRuleListener = p_newRuleListener;
    }

    public BuiltInRuleIndex builtInRuleIndex() {
	return index;
    }

    private NewRuleListener getNewRulePropagator () {
    	return newRuleListener;
    }

    public void scanApplicableRules (Goal       goal,
				     Constraint userConstraint) {
	scanSimplificationRule ( goal, userConstraint, getNewRulePropagator () );
    }

    
    private void scanSimplificationRule ( Goal       goal,
					  Constraint userConstraint, 
					  NewRuleListener listener ) {
	BuiltInRule bir = findSimplificationRule ();
	if ( bir != null ) {
	    scanSimplificationRule ( bir, goal, false, userConstraint, listener );
	    scanSimplificationRule ( bir, goal, true,  userConstraint, listener );
	}
    }

    private BuiltInRule findSimplificationRule () {
        return UpdateSimplificationRule.INSTANCE;
        /*
	IteratorOfBuiltInRule it = index.rules ().iterator ();
	while ( it.hasNext () ) {
	    BuiltInRule bir = it.next ();
	    if ( bir instanceof UpdateSimplificationRule )
		return bir;
	}
	return null;
	*/
    }

    private void scanSimplificationRule ( BuiltInRule rule,
					  Goal        goal,
					  boolean     antec,
					  Constraint  userConstraint, 
					  NewRuleListener listener ) {
	final Node                   node = goal.node ();
	final Sequent                seq  = node.sequent ();

	IteratorOfConstrainedFormula it = 
	    ( antec ? seq.antecedent () : seq.succedent () ).iterator ();

	while ( it.hasNext () ) {
	    final ConstrainedFormula cfma = it.next ();
	    scanSimplificationRule ( rule, goal, antec, userConstraint, cfma, listener );
	}
    }


    private void scanSimplificationRule ( BuiltInRule rule, 
                                          Goal goal, 
                                          boolean antec, 
                                          Constraint userConstraint, 
                                          ConstrainedFormula cfma, 
                                          NewRuleListener listener ) {
        final Node                   node = goal.node ();
        final Sequent                seq  = node.sequent ();

        final PosInOccurrence    pos = new PosInOccurrence 
		( cfma, PosInTerm.TOP_LEVEL, seq, antec );

	    if (rule.isApplicable ( goal, pos, userConstraint ) ) {
		BuiltInRuleApp app = new BuiltInRuleApp ( rule,
							  pos,
							  userConstraint );
		listener.ruleAdded ( app, pos );
	    }
    }

    public void reportRuleApps ( NewRuleListener l,
                                 Goal goal,
                                 Constraint userConstraint ) {
        scanSimplificationRule( goal, userConstraint, l );
    }
    
    /** 
     * called if a formula has been replaced
     * @param sci SequentChangeInfo describing the change of the sequent 
     */  
    public void sequentChanged ( Goal goal, SequentChangeInfo sci ) {
        final Proof proof = goal.node().proof();
        final Constraint userConstraint = proof.getUserConstraint().getConstraint();
        
        scanAddedFormulas ( goal, true, sci, userConstraint );
        scanAddedFormulas ( goal, false, sci, userConstraint );
        
        scanModifiedFormulas ( goal, true, sci, userConstraint );
        scanModifiedFormulas ( goal, false, sci, userConstraint );
    }
    
    private void scanAddedFormulas ( Goal goal, boolean antec, SequentChangeInfo sci, final Constraint userConstraint ) {
        ListOfConstrainedFormula cfmas = sci.addedFormulas( antec );
        final BuiltInRule rule = findSimplificationRule();
        final NewRuleListener listener = getNewRulePropagator();
        
        while ( !cfmas.isEmpty() ) {
            final ConstrainedFormula cfma = cfmas.head();
            cfmas = cfmas.tail();
            
            scanSimplificationRule( rule, goal, antec, userConstraint, cfma, listener );
        }
    }


    private void scanModifiedFormulas ( Goal goal, boolean antec, SequentChangeInfo sci, final Constraint userConstraint ) {
        final BuiltInRule rule = findSimplificationRule();
        final NewRuleListener listener = getNewRulePropagator();
        ListOfFormulaChangeInfo fcis = sci.modifiedFormulas( antec );
        
        while ( !fcis.isEmpty() ) {
            final FormulaChangeInfo fci = fcis.head();
            fcis = fcis.tail();
            
            final ConstrainedFormula cfma = fci.getNewFormula();
            scanSimplificationRule( rule, goal, antec, userConstraint, cfma, listener );
        }
    }

}
