// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import java.io.*;
import de.uka.ilkd.key.pp.AbbrevMap;
import de.uka.ilkd.key.pp.PresentationFeatures;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.gui.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.KeYExceptionHandler;
import de.uka.ilkd.key.util.ExceptionHandlerException;
import java.util.*;

public class ProblemLoader implements Runnable {

    File file;
    Main main;
    KeYMediator mediator;
   
    Proof proof = null;
    IteratorOfNode children = null;

    Node currNode = null;
    KeYExceptionHandler exceptionHandler = null;
    Goal currGoal = null;
    String currTacletName = null;
    int currFormula = 0;
    PosInTerm currPosInTerm = PosInTerm.TOP_LEVEL;
    Stack stack = new Stack();
    LinkedList loadedInsts = null;
    ListOfIfFormulaInstantiation ifSeqFormulaList = 
	SLListOfIfFormulaInstantiation.EMPTY_LIST;

    private SwingWorker worker;

    public ProblemLoader(File file, Main main) {
	this.main = main;
	mediator  = main.mediator();
	this.file = file;
	this.exceptionHandler = mediator.getExceptionHandler();
    }

    public void run() {
        /* Invoking start() on the SwingWorker causes a new Thread
         * to be created that will call construct(), and then
         * finished().  Note that finished() is called even if
         * the worker is interrupted because we catch the
         * InterruptedException in doWork().
         */
        worker = new SwingWorker() {
		public Object construct() {
		    Object res = doWork();
		    return res;
		}
		public void finished() {
		    mediator.startInterface(true);
		    Main main = Main.getInstance();
		    String msg = (String) get();
		    if(!"".equals(msg)) {
		        if(Main.batchMode){
			    System.exit(-1);
			} else {
			    ExceptionDialog ed = 
				new ExceptionDialog
				(mediator.mainFrame(),
				        exceptionHandler.getExceptions());
			    exceptionHandler.clear();
			}
		    } else {
			PresentationFeatures.
			    initialize(init.initConfig().
				       createNamespacesForActivatedChoices().
				       functions(), 
				       mediator.getNotationInfo(),
				       mediator.getSelectedProof());
		    }
		    if (Main.batchMode) {
			//System.out.println("Proof: " +proof.openGoals());
			if(proof.openGoals().size()==0) {
			    System.out.println("proof.openGoals.size="+proof.openGoals().size());
			    System.exit(0);
			}
		        mediator.startAutoMode();
		    }
		    //workaround for the 'old' proof format with no builtin rules:
		    //if the proof is loaded go through every open goal and try to
		    //apply decision procedures
		    if (Main.compatMode) {
		        main.applySimplificationOnGoals();
		    }
		}
	    };
        mediator.stopInterface(true);
        worker.start();
    }


    ProblemInitializer init;

    protected ProofOblInput createProblemFile(File file) {
        final String filename = file.getName();
        if(filename.endsWith(".java")){
	    return new KeYJMLInput(filename, file, 
				   main.getProgressMonitor());
	}else if(filename.endsWith(".key") || 
		 filename.endsWith(".proof")){
	    return new KeYUserProblemFile(filename, file, 
					  main.getProgressMonitor());
	}else if(file.isDirectory()){
	    return new JavaInput(file.getPath(), file, 
					main.getProgressMonitor());
	}
	
	throw new IllegalArgumentException
		("Unsupported file extension \'"+
		        filename.substring(filename.lastIndexOf('.'))+
		        "\' of read-in file " + filename +
		 ". Allowed extensions are: .key, .proof, .java or "+
		 "complete directories."); 
    }

   protected Object doWork() {
       String status = "";
       ProofOblInput problemFile = null;
       try{
	   try{
	       problemFile = createProblemFile(file);
	       init = new ProblemInitializer(main);
	       main.setStatusLine("Loading problem");       
	       init.setUpProof(problemFile);
	       proof = mediator.getSelectedProof();
	       mediator.stopInterface(true); // first stop (above) is not enough 
	       // as there is no problem at that time
	       main.setStatusLine("Loading proof");
	       currNode = proof.root(); // initialize loader
	       children = currNode.childrenIterator(); // --"--
	       init.tryReadProof(this);
	       main.setStandardStatusLine();
	   } catch (ExceptionHandlerException e) {
	       throw e;
	   } catch (Throwable thr) {
	       exceptionHandler.reportException(thr);
	   }
       } catch(ExceptionHandlerException ex){
	       main.setStatusLine("Failed to load problem/proof");
	       status =  ex.toString();
       }
       finally {
	   if (problemFile!=null && problemFile instanceof KeYUserProblemFile){
	       ((KeYUserProblemFile) problemFile).close();
	   }
       }
       return status;
   }

   
    public void loadPreferences(String preferences)
    {
        final ProofSettings proofSettings = main.proofSettings();    
        proofSettings.loadSettingsFromString(preferences);
    }

    // note: Expressions without parameters only emit the endExpr signal
    public void beginExpr(char id, String s) {
	//System.out.println("start "+id+"="+s);
	switch (id) {
	case 'b' : 
            stack.push(children);
            if (children.hasNext()) currNode = children.next();
            break;

	case 'r' : 
            if (currNode == null) currNode = children.next();
            // otherwise we already fetched the node at branch point

            // since the 'old' proof format had no implicit mode for the upd. simplifier,
            // this has been inserted to perform brute force simplification
            if (Main.compatMode == true) {
				currGoal = proof.getGoal(currNode);
				currGoal.applyUpdateSimplifier();
				while (!currNode.leaf()) {
					currNode = currNode.childrenIterator().next();
				}
			}


            currGoal      = proof.getGoal(currNode);
	    //System.err.println((currNode == null)+"<-N G->"+ (currGoal == null));

	    //if (currGoal == null) System.err.println(currNode);
           
            // XXX: since there is no implicit mode for the upd. simplifier,
            // this has been inserted to perform brute force simplification
            // until built-in rules can be loaded from a proof text
            /*currGoal.applyUpdateSimplifier();
            while (!currNode.leaf()) {
                currNode = currNode.childrenIterator().next();
            }*/
//            currGoal      = proof.getGoal(currNode);
// removed since the goal is reused for the 1st (here and only) node



            currTacletName= s;
            // set default state
            currFormula   = 0;
            currPosInTerm = PosInTerm.TOP_LEVEL; 
            loadedInsts   = null;
            ifSeqFormulaList = SLListOfIfFormulaInstantiation.EMPTY_LIST;
            break;

	case 'f' :
            currFormula   = Integer.parseInt(s);
            break;

	case 't' :
            currPosInTerm = PosInTerm.parseReverseString(s);
            break;
            
	case 'i' :
            if (loadedInsts == null) loadedInsts = new LinkedList();
            loadedInsts.add(s);
            break;
            
	case 'h' :
	    //             Debug.fail("Detected use of heuristics!");
	    break;
            
	case 'q' : // ifseqformula
	    //            Namespace varNS = mediator.var_ns();
	    //            varNS = currGoal.getVariableNamespace(varNS);
	    //            Term f = parseTerm(s, varNS);
            Sequent seq = currGoal.sequent();
            ifSeqFormulaList = ifSeqFormulaList.append(
	        new IfFormulaInstSeq(seq, Integer.parseInt(s)));
            break;
	case 'u' : //UserLog
            if(proof.userLog==null)
                proof.userLog = new Vector();
            proof.userLog.add(s);
            break;
	case 'v' : //Version log
            if(proof.keyVersionLog==null)
                proof.keyVersionLog = new Vector();
            proof.keyVersionLog.add(s);
            break;
	case 's' : //ProofSettings
            //System.out.println("------------------------\n" + s + "------------\n");
	    //necessary for downward compatibility of the proof format
	    loadPreferences(s);
        break;
	case 'n' : //BuiltIn rules
            if (currNode == null) currNode = children.next();
            currGoal      = proof.getGoal(currNode);
            
            currTacletName = s;
            // set default state
            currFormula   = 0;
            currPosInTerm = PosInTerm.TOP_LEVEL; 
            break;
	}
    }


    public void endExpr(char id) {
	//System.out.println("end "+id);
	switch (id) {
	case 'b' : 
            children = (IteratorOfNode) stack.pop();
            break;
	case 'a' :
	    // do nothing; attribute interactive was modified here
	    break;
	case 'r' : 
            try{
                // TODO: not working with update simplification!
//               currGoal.applyUpdateSimplifier(); goal/node/children anpassen?
               currGoal.apply(constructApp());
               children = currNode.childrenIterator();
               currNode = null;
            } catch(AppConstructionException e) {
		//               System.err.println(e.getMessage());
		throw new RuntimeException(e.getMessage());
            }
            break;
        case 'n' :
        	try {
        		currGoal.apply(constructBuiltinApp());
        		children = currNode.childrenIterator();
        		currNode = null;
        	} catch (BuiltInConstructionException e) {
        		throw new RuntimeException(e.getMessage());
        	}
            break;
	}
        
    }
  
    /**
	 * Constructs rule application for UpdateSimplification from current parser
	 * information
	 * 
	 * @return current rule application for updateSimplification
	 */
	private BuiltInRuleApp constructBuiltinApp() throws BuiltInConstructionException {
		BuiltInRuleApp ourApp = null;
		//PosInSequent posInSeq = null;
		PosInOccurrence pos = null;

		if (currFormula != 0) // otherwise we have no pos
		{
			//posInSeq = PosInSequent.createCfmaPos(PosInOccurrence
			//		.findInSequent(currGoal.sequent(), currFormula, currPosInTerm));
			pos = PosInOccurrence.findInSequent(currGoal.sequent(),
					currFormula, currPosInTerm);
		} else {
			//posInSeq = PosInSequent.createSequentPos();
		}
		
		final Constraint userConstraint = mediator.getUserConstraint()
				.getConstraint();

		final SetOfRuleApp ruleApps = 
		    mediator.getBuiltInRuleApplications(currTacletName, pos);

		if (ruleApps.size() != 1) {
		    if (ruleApps.size() < 1) {
		        throw new BuiltInConstructionException
		        (currTacletName + 
		                " is missing. Most probably the binary "+
		                "for this built-in rule ist not in your path or " +
		        "you do not have the permission to execute it.");
		    } else {
		        throw new BuiltInConstructionException
		        (currTacletName + ": found " + ruleApps.size() +
		                " applications. Don't know what to do !\n" + 
		                "@ " + pos);
		    }
		}
		ourApp = (BuiltInRuleApp) ruleApps.iterator().next();
		return ourApp;
	}
   
    private TacletApp constructApp() throws AppConstructionException {

	TacletApp ourApp = null;
	PosInOccurrence pos = null;
	if (currFormula != 0) // otherwise we have no pos
	    pos = PosInOccurrence.findInSequent(currGoal.sequent(),
						currFormula,
						currPosInTerm);

	//System.err.print("Want to apply "+currTacletName+" at "+currGoal);
	SetOfTacletApp applics;
	try {
	    applics=mediator.getTacletApplications(currGoal, currTacletName, pos);
	} catch(IndexOutOfBoundsException e) { // pos. is not a valid subterm
	    System.err.println(currTacletName+":"+pos);
	    e.printStackTrace();
	    throw new AppConstructionException(e.toString());
	}
	if (applics.size() != 1) {
	    throw new AppConstructionException(currTacletName+": found "+
		applics.size()+" applications. Don't know what to do!\n"+
		"@ "+pos);
	}
	ourApp = applics.iterator().next();
	ourApp = constructInsts(ourApp);

	ourApp = ourApp.setIfFormulaInstantiations(ifSeqFormulaList,
	    mediator.getServices(), 
            mediator.getUserConstraint().getConstraint ());

	ourApp = ourApp.createDependingFunctions ( mediator.func_ns () );

	return ourApp;
    }


    
    /** 1st pass: only VariableSV */
    public static TacletApp parseSV1(TacletApp app, SchemaVariable sv,
                                     String value) {
        LogicVariable lv=new LogicVariable(new Name(value),
				           app.getRealSort(sv));
        Term instance = TermFactory.DEFAULT.createVariableTerm(lv);
        return app.addCheckedInstantiation(sv, instance, true);
    }
    


    /** 2nd pass: all other SV */
    public static TacletApp parseSV2(TacletApp app, SchemaVariable sv,
                                     String value, Goal targetGoal) {
        TacletApp result;
        Proof p = targetGoal.node().proof();
        if (sv.isVariableSV()) {
	    // ignore -- already done
            result = app;
        } else if (sv.isProgramSV()) {
	    ProgramElement pe = 
	        TacletInstantiationsTableModel.getProgramElement(
		    app, value, sv, p.getServices());
	    result = app.addInstantiation(sv, pe, true);
        } else if ( sv.isDependingSV() ) {
	    result = app.createDependingConstant ( value, sv, true );
        } else {
	    Namespace varNS = p.getNamespaces().variables();
	    varNS = app.extendVarNamespaceForSV(varNS, sv);
	    Term instance = parseTerm(value, p, varNS, 
	            targetGoal.getVariableNamespace(varNS));
	    result = app.addCheckedInstantiation(sv,instance, true); 
        }
        return result;
    }
    
    
   
   
    private TacletApp constructInsts(TacletApp app) {
	if (loadedInsts == null) return app;
	SetOfSchemaVariable uninsts = app.uninstantiatedVars();

	// first pass: add variables
	Iterator it = loadedInsts.iterator();
	while (it.hasNext()) {
	    String s = (String) it.next();
	    int eq = s.indexOf('=');
	    String varname = s.substring(0, eq);
	    String value = s.substring(eq+1, s.length());
	    SchemaVariable sv = lookupName(uninsts, varname);
         
	    if (sv.isVariableSV()) {
                app = parseSV1(app, sv, value);
	    }
	}

	// second pass: add everything else
	it = loadedInsts.iterator();
	while (it.hasNext()) {
	    String s = (String) it.next();
	    int eq = s.indexOf('=');
	    String varname = s.substring(0, eq);
	    String value = s.substring(eq+1, s.length());
         
	    SchemaVariable sv = lookupName(uninsts, varname);
            app = parseSV2(app, sv, value, currGoal);
         
	}
	return app;
    }
   
   
    public static Term parseTerm(String value, Proof proof, 
            Namespace varNS, Namespace progVar_ns) {
	try {
            return TermParserFactory.createInstance().
		parse(new StringReader(value), null,
		      proof.getServices(), 
		      varNS, 
		      proof.getNamespaces().functions(), 
		      proof.getNamespaces().sorts(),
		      progVar_ns,
		      new AbbrevMap());
	} catch(ParserException e) {
	    throw new RuntimeException("While parsing "+value+"\n"+
				       varNS+"\n"+e);         
	}
    }
   
    public static Term parseTerm(String value, Proof proof) {
	return parseTerm(value, proof, proof.getNamespaces().variables(),
	        proof.getNamespaces().programVariables());
    }
   
   
    private SchemaVariable lookupName(SetOfSchemaVariable set, String name) {
	IteratorOfSchemaVariable it = set.iterator();
	while (it.hasNext()) {
	    SchemaVariable v = it.next();
	    if (v.name().toString().equals(name)) return v;
	}
	return null; // handle this better!
    }
   
    private class AppConstructionException extends Exception {

	AppConstructionException(String s) {
	    super(s);
	}
    }
    
    private class BuiltInConstructionException extends Exception {

    	BuiltInConstructionException(String s) {
    	    super(s);
    }
    }
   
}
