// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

/**
 * Replaces program variables.
 */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.ArrayOfProgramElement;
import de.uka.ilkd.key.java.visitor.ProgVarReplaceVisitor;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import java.util.ArrayList;
import java.util.Map;


public class ProgVarReplacer {

    /**
     * map specifying the replacements to be done
     */
    private final Map map;


    /**
     * creates a ProgVarReplacer that replaces program variables as specified
     * by the map parameter
     */
    public ProgVarReplacer(Map m) {
        map = m;
    }


    /**
     * merges "next" into "base"
     * precondition:
     *   "next" is the result of replacing in "base" the formula at position
     *   "idx" by calling Semisequent.replace()
     *   (this implies that "next" contains exactly one removed and one added
     *   formula)
     */
    private void mergeSemiCIs(SemisequentChangeInfo base,
                              SemisequentChangeInfo next,
                              int idx) {
        Debug.assertTrue(next.modifiedFormulas().isEmpty());

        IteratorOfConstrainedFormula remIt = next.removedFormulas().iterator();
        Debug.assertTrue(remIt.hasNext());
        ConstrainedFormula remCf = remIt.next();
        Debug.assertFalse(remIt.hasNext());
        base.removedFormula(idx, remCf);

        IteratorOfConstrainedFormula addIt = next.addedFormulas().iterator();
        Debug.assertTrue(addIt.hasNext());
        ConstrainedFormula addCf = addIt.next();
        Debug.assertFalse(addIt.hasNext());
        base.addedFormula(idx, addCf);

        base.setFormulaList(next.getFormulaList());
        base.setSemisequent(next.semisequent());
    }


    /**
     * replaces in a goal
     */
    public void replace(Goal goal) {
	//globals
    	SetOfProgramVariable set = replace(goal.getGlobalProgVars());
	goal.setGlobalProgVars(set);

	//taclet apps
	replace(goal.ruleAppIndex().tacletIndex());

	//sequent
	SequentChangeInfo sci = replace(goal.sequent());
	goal.setSequent(sci);
    }


    /**
     * replaces in a set
     */
    public SetOfProgramVariable replace(SetOfProgramVariable vars) {
    	SetOfProgramVariable result = vars;

    	IteratorOfProgramVariable it = vars.iterator();
	while(it.hasNext()) {
	    ProgramVariable var = it.next();
	    ProgramVariable newVar = (ProgramVariable)map.get(var);
	    if(newVar != null) {
	    	result = result.remove(var);
	    	result = result.add(newVar);
	    }
	}

	return result;
    }


    /**
     * replaces in the partially instantiated apps of a taclet index
     */
    public void replace(TacletIndex tacletIndex) {
	ListOfNoPosTacletApp noPosTacletApps
		= tacletIndex.getPartialInstantiatedApps();
	SetOfNoPosTacletApp appsToBeRemoved, appsToBeAdded;
	appsToBeRemoved = SetAsListOfNoPosTacletApp.EMPTY_SET;
	appsToBeAdded   = SetAsListOfNoPosTacletApp.EMPTY_SET;

	IteratorOfNoPosTacletApp it = noPosTacletApps.iterator();
	while(it.hasNext()) {
	    NoPosTacletApp noPosTacletApp = it.next();
	    SVInstantiations insts = noPosTacletApp.instantiations();

	    SVInstantiations newInsts = replace(insts);

	    if(newInsts != insts) {
	    	NoPosTacletApp newNoPosTacletApp
		 	= NoPosTacletApp.createNoPosTacletApp(
				noPosTacletApp.taclet(),
		    		newInsts,
				noPosTacletApp.constraint(),
				noPosTacletApp.newMetavariables(),
				noPosTacletApp.ifFormulaInstantiations());
		appsToBeRemoved = appsToBeRemoved.add(noPosTacletApp);
		appsToBeAdded   = appsToBeAdded.add(newNoPosTacletApp);
	    }
	}

	tacletIndex.removeTaclets(appsToBeRemoved);
	tacletIndex.addTaclets(appsToBeAdded);
    }


    /**
     * replaces in an SVInstantiations
     */
    public SVInstantiations replace(SVInstantiations insts) {
   	SVInstantiations result = insts;

    	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it;
	it = insts.pairIterator();
	while(it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry e = it.next();
	    SchemaVariable sv     = e.key();
	    InstantiationEntry ie = e.value();
	    Object inst = ie.getInstantiation();

	    if(ie instanceof ContextInstantiationEntry) {
		Statement s = (Statement) inst;
		Statement newS = replace(s);
		if(newS != s) {
		    ContextInstantiationEntry cie = (ContextInstantiationEntry) ie;
		    result = result.replace(cie.prefix(),
					    cie.suffix(),
					    cie.activeStatementContext(),
					    newS);
		}
	    } else if(ie instanceof OperatorInstantiation) {
	    	/*nothing to be done (currently)*/
	    } else if(ie instanceof ProgramInstantiation) {
		Statement s = (Statement) inst;
		Statement newS = replace(s);
		if(newS != s) {
		    result = result.replace(sv, newS);
		}
	    } else if(ie instanceof ProgramListInstantiation) {
		ArrayOfProgramElement a = (ArrayOfProgramElement) inst;
		int size = a.size();
		Statement[] array = new Statement[size];

		boolean changedSomething = false;
		for(int i = 0; i < size; i++) {
		    Statement s = (Statement) a.getProgramElement(i);
		    array[i] = replace(s);
		    if(array[i] != s) {
		    	changedSomething = true;
		    }
		}

		if(changedSomething) {
		    ArrayOfProgramElement newA = new ArrayOfProgramElement(array);
		    result = result.replace(sv, newA);
		}
	    } else if(ie instanceof TermInstantiation) {
		Term t = (Term) inst;
		Term newT = replace(t);
		if(newT != t) {
		    result = result.replace(sv, newT);
		}
	    } else {
		Debug.fail("unexpected subtype of InstantiationEntry");
	    }
	}

	return result;
    }


    /**
     * replaces in a sequent
     */
    public SequentChangeInfo replace(Sequent s) {
        SemisequentChangeInfo anteCI = replace(s.antecedent());
        SemisequentChangeInfo succCI = replace(s.succedent());

        Semisequent newAntecedent = anteCI.semisequent();
        Semisequent newSuccedent  = succCI.semisequent();

        Sequent newSequent = Sequent.createSequent(newAntecedent,
                                                   newSuccedent);

        SequentChangeInfo result = SequentChangeInfo.createSequentChangeInfo
                                              (anteCI, succCI, newSequent, s);
        return result;
    }


    /**
     * replaces in a semisequent
     */
    public SemisequentChangeInfo replace(Semisequent s) {
    	SemisequentChangeInfo result = new SemisequentChangeInfo();
        result.setFormulaList(s.toList());
        result.setSemisequent(s);

        for(int i = 0; i < s.size(); i++) {
            ConstrainedFormula oldcf = s.get(i);
            ConstrainedFormula newcf = replace(oldcf);

            if(newcf != oldcf) {
                SemisequentChangeInfo semiCI
                                      = result.semisequent().replace(i, newcf);
                mergeSemiCIs(result, semiCI, i);
            }
        }

        return result;
    }


    /**
     * replaces in a constrained formula
     */
    public ConstrainedFormula replace(ConstrainedFormula cf) {
        ConstrainedFormula result = cf;

	Term formula = cf.formula();
	Term newFormula = replace(formula);

	if(newFormula != formula) {
            result = new ConstrainedFormula(newFormula, cf.constraint());
	}

        return result;
    }


    /**
     * replaces in a term
     */
    public Term replace(Term t) {
        Term result = t;

        if(t.op() instanceof ProgramVariable) {
            ProgramVariable newVar = (ProgramVariable)map.get(t.op());
            if(newVar != null) {
                result = TermFactory.DEFAULT.createVariableTerm(newVar);
            }
        } else {
            Term newSubTerms[] = new Term[t.arity()];
            ArrayList vars = new ArrayList();

	    boolean changedSubTerm = false;
            for(int i = 0; i < t.arity(); i++) {
	    	Term subTerm = t.sub(i);
                newSubTerms[i] = replace(subTerm);

		if(newSubTerms[i] != subTerm) {
		    changedSubTerm = true;
		}

                ArrayOfQuantifiableVariable vbh = t.varsBoundHere(i);
                for(int j = 0; j < vbh.size(); j++) {
                    vars.add(vbh.getQuantifiableVariable(j));
                }
            }

            QuantifiableVariable v1[] = new QuantifiableVariable[0];
            v1 = (QuantifiableVariable[])(vars.toArray(v1));

            JavaBlock jb = t.javaBlock();
	    JavaBlock newJb = jb;
            if (!jb.isEmpty()) {
	    	Statement s = (Statement)jb.program();
		Statement newS = replace(s);
		if(newS != s) {
  		    newJb = JavaBlock.createJavaBlock((StatementBlock)newS);
		}
            }

	    Operator op = t.op();
            Operator newOp = op;
	    if(op instanceof UpdateOperator) {
     		newOp = replace((UpdateOperator)op);
	    }

	    if(newOp != op || changedSubTerm || newJb != jb) {
            	result = TermFactory.DEFAULT.createTerm(newOp,
	    						newSubTerms,
							v1,
							newJb,
							null);
	    }
        }

        return result;
    }


    /**
     * replaces the locations inside an update
     */
    private UpdateOperator replace(UpdateOperator updateOp) {
        UpdateOperator result = updateOp;
        final Location[] locs = new Location[updateOp.locationCount()];
        boolean toReplace = false;
        for(int i = 0, size = updateOp.locationCount(); i<size; i++) {
            Location loc = updateOp.location(i);
            if (loc instanceof ProgramVariable &&
                    !((ProgramVariable)loc).isMember()) {
                ProgramVariable newVar = (ProgramVariable)map.get(loc);
                if(newVar != null) {
                    locs[i] = newVar;
                    toReplace = true;
                } else {
                    locs[i] = loc;
                }
            } else {
                locs[i] = loc;
            }
        }
        if (toReplace) {
            result = UpdateOperator.createUpdateOp(locs);
        }
        return result;
    }


    /**
     * replaces in a statement
     */
    public Statement replace(Statement s) {
        ProgVarReplaceVisitor pvrv = new ProgVarReplaceVisitor(s, map, false);
	pvrv.start();
	return (Statement) pvrv.result();
    }
}
