// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.proof.mgt.MethodContract;
import de.uka.ilkd.key.proof.mgt.Contract;


/** represents a proof obligation originating from an OCL pre/post constraint
 * as input for the prover.
 */
public class OCLPostPO extends OCLProofOblInput {

    private String poPreText;
    private String poPostText;
    private Modality modality;
    private String location;
    private ReprModelMethod meth;
    private Term proofObl;

    /** creates a new representation of a proof obligation from an OCL 
     * constraint with the given name, originating from a representation 
     * of a model method and having the given Strings as text of the pre
     * condition and of the post condition. 
     */
    public OCLPostPO(String name,
		     ReprModelMethod meth,
		     String poPreText, 
		     String postText, 
		     Modality modality,
		     String location) {
	super(name+" of "+meth.getName());
	this.location = location;
	this.poPostText=postText;
	this.poPreText=poPreText;
	this.meth = meth;
	this.modality = modality;
    }


    /**
     * Set the result of reading the OCL input.
     */
    protected final void setProofObligation(Term theResultingProofObligation) {
	this.proofObl = theResultingProofObligation;
    }

    public Term getPOTerm() {
	return proofObl;
    }

    public String poPreText() {
	return poPreText;
    }

    public String poPostText() {
	return poPostText;
    }

    /** reads the proof obligation, i.e., invokes the OCL-DL-translation on
     * the OCL constraint of this OCL proof obligation.
     */
    public void read(ModStrategy mod) {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	de.uka.ilkd.key.ProverFacade fac 
	    = initConfig.createFacadeForTransl();
	if (!mod.modifyProgramVariables()) {	    
	    fac.doNotAddProgramVariables();
	}
        setProofObligation(generateMethodPO(fac,
					    methRepr(), "prepost", 
					    poPreText(), poPostText(), 
					    modality, 
					    getJavaPath()));
    }


    public ProofList getPO() {
        if (proofObl == null) throw 
           new IllegalStateException("No proofObl term");
        String s = createProofHeader(methRepr().getContainingClass()
				     .getRootDirectory());
	return new ProofList(
           new Proof(name(), proofObl, s, 
		     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex(),
                     initConfig.getServices()), name());
    }

    private ReprModelMethod methRepr() {
	return meth;
    }

    public ReprModelClass getModelClass() {
	return methRepr().getContainingClass();
    }

    public ReprModelMethod getModelMethod() {
	return methRepr();
    }

    public Object getObjectOfContract() {
	return getModelMethod();
    }

    public boolean fitsToContract(Contract ct) {
	if (!(ct instanceof MethodContract)) {
	    return false;
	}
	MethodContract mct = (MethodContract)ct;
	return mct.getModelMethod().equals(getModelMethod())
	    && mct.getPreText().equals(poPreText())
	    && mct.getModality().equals(modality)
	    && mct.getPostText().equals(poPostText());
    }



//     public boolean equals(Object cmp) {
// 	if (!(cmp instanceof OCLPostPO)) {
// 	    return false;
// 	}
// 	return super.equals(cmp) 
// 	    && (this.methRepr.equals(((OCLPostPO)cmp).getModelMethod()));
//     }

}
