// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.logic.SetOfChoice;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import java.util.LinkedList;

/** 
 *  A RewriteTaclet represents a taclet, whose find can be matched against any
 * term in the sequent no matter where it occurs. The only constraint to be
 * fulfilled is that the term matches the structure described by the term of the
 * find-part.
 */
public class RewriteTaclet extends FindTaclet{

    /** If true: The updates above the find position and the updates
     * in front of the if formulas have to be equivalent; there may
     * not be any program modalities above the find position */
    protected boolean sameUpdatePrefix;

    /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters that represents a rewrite rule.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequent, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     * @param find the find term of the Taclet
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     * @param replaceWith a boolean indicating if one of the goal descriptions
     * is a replacewith (required for location check)
     */
    public RewriteTaclet(Name name, TacletApplPart applPart,  
			 ListOfTacletGoalTemplate  goalTemplates, 
			 ListOfRuleSet           heuristics,
			 Constraint                constraint,
			 TacletAttributes          attrs,
			 Term                      find, 
			 MapFromSchemaVariableToTacletPrefix prefixMap,
			 boolean                   p_sameUpdatePrefix,
			 SetOfChoice choices){
	super(name, applPart, goalTemplates, heuristics, constraint,
	      attrs, find, prefixMap, choices);
	sameUpdatePrefix = p_sameUpdatePrefix;
	
	cacheMatchInfo();
    }	

      
    /** 
     * this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected boolean ignoreTopLevelUpdates() {
	return false;
    }
    
    /**
     * @return true iff: The updates above the find position and the
     * updates in front of the if formulas have to be equivalent;
     * there may not be any program modalities above the find position
     */
    public boolean getSameUpdatePrefix () {
	return sameUpdatePrefix;
    }


    /**
     * the top level operator has to be a simultaneous update. This method 
     * checks if the assignment pairs of the update contain free logic
     * variables and gives a veto if positive
     * @param t the Term to check
     * @return false if vetoing 
     */
    private boolean veto(Term t) {
	for (int i = 0; i<t.arity() - 1; i++) {
	    if (t.sub(i).freeVars() != SetAsListOfQuantifiableVariable.EMPTY_SET) {
		return true;
	    }
	}
	return false;
    }

    /**
     * For taclets with <code>getSameUpdatePrefix ()</code>, collect
     * the updates above <code>p_pos</code> and add them to the update
     * context of the instantiations object <code>p_mc</code>.
     * @return the new instantiations with the additional updates, or
     * <code>null</code>, if program modalities appear above
     * <code>p_pos</code>
     */
    public MatchConditions checkUpdatePrefix
	( PosInOccurrence p_pos,
	  MatchConditions p_mc,
	  Services        p_services,
	  Constraint      p_userConstraint ) {
	if ( !getSameUpdatePrefix () ) 
	    return p_mc;

	SVInstantiations svi = p_mc.getInstantiations ();
	if ( p_pos.posInTerm () != null ) {
	    PIOPathIterator it = p_pos.iterator ();
	    Operator        op;

	    while ( it.next () != -1 ) {
		final Term t = it.getSubTerm ();
		op = t.op ();

		if ( op instanceof UpdateOperator &&
		     it.getChild () == ((UpdateOperator)op).arity () - 1) {		    
		    if ( veto(t) ) {
			return null;
		    } else {
			svi = svi.addUpdate ( it.getSubTerm () );
		    }
		    
		}
		else if ( op instanceof Modality ||
			  op instanceof ModalOperatorSV)
		    return null;
	    }
	}

	return p_mc.setInstantiations ( svi );
    }

    /**
     * does the work for applyReplacewith (wraps recursion) 
     */
    private Term replace(Term term, Term with, IntIterator it,
			 Services services, SVInstantiations svInst) {	
	if (it.hasNext()) {	    
	    int sub=it.next();
	    Term[] subs=new Term[term.arity()];
	    for (int i=0;i<term.arity();i++) {
		if (i!=sub) {
		    subs[i]=term.sub(i);
		} else {
		    subs[i]=replace(term.sub(i), with, it, services, svInst);
		}
	    }

	    boolean containsBoundVars = false;
	    LinkedList[] vars = new LinkedList[term.arity()];
	    for (int i=0; i<term.arity(); i++) {
		vars[i] = new LinkedList();
		for (int j=0; j<term.varsBoundHere(i).size(); j++) {
		    containsBoundVars = true;
		    vars[i].add(term.varsBoundHere(i).getQuantifiableVariable(j));
		}
	    }
	    ArrayOfQuantifiableVariable[] origvars = null;
	    if (containsBoundVars) {
		origvars = new ArrayOfQuantifiableVariable[term.arity()];
		for (int i=0; i<term.arity(); i++) {
		    origvars[i] = new ArrayOfQuantifiableVariable(vars[i]);
		}
	    }
	    //

	    //	    if(term.op() instanceof Modality && with.op() instanceof ModalOperatorSV) {
	    //		// This seems never be the case, I probably misunderstood this part. /WOJ
	    //		System.out.println("If you see this message contact Wojciech Mostowski and provide the steps to reproduce this message (there is nothing wrong!)");
	    //		ModalOperatorSV modalsv = (ModalOperatorSV)with.op();
	    //		Modality o = (Modality) svInst.getInstantiation(modalsv);
	    //		if (o == null){
	    //		    Debug.fail("No instantiation for " + term.op() + " found!");
	    //		}
	    //		else if (o != (Modality)term.op()) {
	    //		    Debug.fail("Something wrong: "+o+" should be the same as "+ (Modality)term.op());
	    //		}
	    //		return TermFactory.DEFAULT.createTerm(o, subs, vars,
	    //						      term.javaBlock(), 
	    //						      term.ruleVar());
	    //	    }else{


		return TermFactory.DEFAULT.createTerm(term.op(), 
						      subs, 
						      origvars,
						      term.javaBlock(), 
						      term.ruleVar());
		//	    }    
	} 
	return svInst == SVInstantiations.EMPTY_SVINSTANTIATIONS ?
	    with : syntacticalReplace(with, services, svInst);
    }


    /**
     * Check if p_pos contains an explicit metavariable instantiation,
     * and creates in this case a new simple term by replacing the
     * metavariable with the instantiation
     */
    private PosInOccurrence flatten ( PosInOccurrence p_pos,
				      Services        p_services) {
	if ( p_pos.termBelowMetavariable () != null ) {
	    Term flatTerm = replace ( p_pos.constrainedFormula ().formula (),
				      p_pos.termBelowMetavariable (),
				      p_pos.posInTerm ().iterator (),
				      p_services,
				      SVInstantiations.EMPTY_SVINSTANTIATIONS );

	    PosInOccurrence pos = new PosInOccurrence
		( new ConstrainedFormula ( flatTerm,
					   p_pos.constrainedFormula ().constraint () ),
		  p_pos.posInTerm (),
		  null,
		  p_pos.isInAntec() );

	    IntIterator it = p_pos.posInTermBelowMetavariable ().iterator ();
	    while ( it.hasNext () )
		pos = pos.down ( it.next () );

	    return pos;
	}

	return p_pos;
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param services the Services encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyReplacewith(TacletGoalTemplate gt, 
				    Goal               goal,
				    PosInOccurrence    posOfFind,
				    Services           services,
				    MatchConditions    matchCond) {
	if ( gt instanceof RewriteTacletGoalTemplate ) {
	    PosInOccurrence flatPos = flatten ( posOfFind, services );

	    Term term = flatPos.constrainedFormula().formula();
	    IntIterator it = flatPos.posInTerm().iterator();
	    Term rwTemplate=((RewriteTacletGoalTemplate)gt).replaceWith();
	    Term result = replace(term, rwTemplate, it, services, matchCond.getInstantiations ());
	    Constraint c = matchCond.getConstraint ();
	    if ( createCopies ( goal, posOfFind, matchCond ) )
                goal.addFormula ( new ConstrainedFormula ( result, c ),
                                  posOfFind );
            else
                goal.changeFormula ( new ConstrainedFormula ( result, c ),
                                     posOfFind );
	} else {
	    // Then there was no replacewith...
	    // This is strange in a RewriteTaclet, but who knows...
	}
    }

    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent
     * @param services the Services encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyAdd(Sequent         add, 
			    Goal            goal,
			    PosInOccurrence posOfFind,
			    Services        services,
			    MatchConditions matchCond) {
	if (posOfFind.isInAntec()) {
	    addToAntec(add.antecedent(), goal, posOfFind, services, matchCond);
	    addToSucc(add.succedent(), goal, null, services, matchCond);
	} else {
	    addToAntec(add.antecedent(), goal, null, services, matchCond);	
	    addToSucc(add.succedent(), goal, posOfFind, services, matchCond);
	}
    }
    
    protected Taclet setName(String s) {
	RewriteTacletBuilder b = new RewriteTacletBuilder();
	b.setFind(find());
	b.setSameUpdatePrefix ( getSameUpdatePrefix () );
	return super.setName(s, b);
    }


    StringBuffer toStringFind(StringBuffer sb) {
	StringBuffer res = super.toStringFind ( sb );
	if ( getSameUpdatePrefix () )
	    res.append ( "sameUpdateLevel\n" );
	return res;
    }

}
