// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
/*
 * Created on 26.11.2004
 */
package de.uka.ilkd.key.rule.updatesimplifier;

import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.util.Debug;

/**
 * The default implementation of an assignment pair.
 * There exists a lazy one to which may spare resources
 * {@see AsssignmentPairLazy} 
 * @author bubel
 */
public class AssignmentPairImpl implements AssignmentPair {

    /**
     * @param accessOp
     * @param locSubs
     * @param value
     */
  
    private final Location accessOp;
    private final Term[] locSubs;
    private final Term value;
    private int cachedLocationHashCode;
    
    
    /**
     * creates the assignment <code>accessOp(locSubs):=value</code> 
     * @param accessOp the Location descriptor 
     * @param locSubs the subterms of the location
     * @param value the value the location is updated to
     */
    public AssignmentPairImpl(final Location accessOp, final Term[] locSubs,
            final Term value) {
        this.accessOp = accessOp;
        this.locSubs = locSubs;
        this.value = value;
    }
        
    /**
     * 
     * returns the left side of the assignment pair as a term
     * 
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#locationAsTerm()
     */
    public Term locationAsTerm() {       
        return TermFactory.DEFAULT.createTerm(accessOp, locSubs,
                new ArrayOfQuantifiableVariable(), JavaBlock.EMPTY_JAVABLOCK, 
                null);
    }

    /**
     * returns the location descriptor
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#location()
     */
    public Location location() {
        return accessOp;
    }

    /**
     * returns the sub terms of the location
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#locationSubs()
     */
    public Term[] locationSubs() {     
        return locSubs;
    }

    /**
     * returns the value the location is updated to
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#value()
     */
    public Term value() {
        return value;
    }

    /**
     * true if the locations of both assignment pairs are equal
     * 
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#equalLocations(de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair)
     */
    public boolean equalLocations(AssignmentPair p) {
        if (p.location() != accessOp) {
            return false;
        }
        Term[] cmpLocSubs = p.locationSubs();
        Debug.assertTrue(locSubs.length == cmpLocSubs.length);
        for (int i = 0; i < locSubs.length; i++) {
            if (!locSubs[i].equals(cmpLocSubs[i])) {
                return false;
            }
        }
        return true;
    }

    /**     
     * hashcode implementation
     * @see de.uka.ilkd.key.rule.updatesimplifier.AssignmentPair#locationHashCode()
     */
    public int locationHashCode() {      
        if (cachedLocationHashCode == 0) {
            cachedLocationHashCode = accessOp.hashCode();
            for (int i = 0; i<locSubs.length; i++) {
                cachedLocationHashCode += 17*locSubs[i].hashCode();
            }        
            if (cachedLocationHashCode == 0) {
                cachedLocationHashCode = 1;
            }
        }        
        return cachedLocationHashCode;
    }

    /** customized hashcode */
    public int hashCode() {
        return cachedLocationHashCode * 31 + value.hashCode();
    }
    
    /** customized equals */
    public boolean equals(Object o) {
        if (!(o instanceof AssignmentPair)) {
            return false;
        }
        final AssignmentPair cmp = (AssignmentPair)o;
        return equalLocations(cmp) && cmp.value().equals(value);
    }

    public String toString() {
        return locationAsTerm() + ":=" + value();    
    }
}
